/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#ifndef __CIDEV_H__
#define __CIDEV_H__

#ifdef __linux__
#ifndef __KERNEL__
#include "global.h"
#endif  /*  __KERNEL__  */
#endif  /*  __linux__  */

#ifdef __FreeBSD__
#ifndef _KERNEL
#include "global.h"
#endif  /*  _KERNEL  */
#endif  /*  __FreeBSD__ */

#define CIDEV_MAGIC         (0x47465341)
#define CIDEV_ENTRY_MAGIC   (CIDEV_MAGIC + 1)
#define CIDEV_BLOCKSIZE     (512)
#define CIDEV_VERSION       (0x47030100)

#define CIDEV_KEYWORD_SIZE  (64)
#define CIDEV_KEYWORD_SLEN  (CIDEV_KEYWORD_SIZE - 1)

/* cidev_marker_t is first 8 bytes of every block */
#define CIDEV_STRING_SIZE   (CIDEV_BLOCKSIZE - CIDEV_KEYWORD_SIZE - 8)
#define CIDEV_STRING_SLEN   (CIDEV_STRING_SIZE - 1)


/* 
 * in core structure used to store cidev information
 */

struct cidev_entry
{
    unsigned char keyword[CIDEV_KEYWORD_SIZE];
    unsigned char string[CIDEV_STRING_SIZE];
    int id;
    int inuse;
    struct cidev_entry *next;
};
typedef struct cidev_entry cidev_entry_t;


/* 
 * on disk structure forms boundary of first/last blocks of cidev:
 */

struct cidev_header
{
    uint32 magic;
    uint32 inuse;       /* set TRUE at creation, cleared by erase */
    uint32 version;
    char name[256];     /* name of cidev, /dev/whatever */
};
typedef struct cidev_header cidev_header_t;


/* 
 * on disk structure of all blocks between headers:
 * [cidev_marker] [CIDEV_KEYWORD_SIZE] [CIDEV_STRING_SIZE]
 * 0            7 8                 71 72              511 
 */

struct cidev_marker
{
    uint32 magic;
    uint32 inuse;
};
typedef struct cidev_marker cidev_marker_t;


#ifdef __KERNEL__

typedef void ci_bdev_t;

/* Prototypes for kernel cidev functions */
ci_bdev_t *cidev_open(char *name);
unsigned char *cidev_name(ci_bdev_t *dev);
void cidev_close(ci_bdev_t *dev);
void cidev_set_blocksize(ci_bdev_t *dev, int size);
int cidev_inuse(ci_bdev_t *dev, int *bool);
int cidev_read_header(ci_bdev_t *dev, cidev_header_t *head);
int cidev_entry_count(ci_bdev_t *dev, int *count);
int cidev_read_entries(ci_bdev_t *dev, cidev_entry_t *list, int *count);
int cidev_read_entry_num(ci_bdev_t *dev, int entry_num, cidev_entry_t *entry);
int cidev_read_entry(ci_bdev_t *dev, int entry_id, cidev_entry_t *entry);
int cidev_read_entry_keyword(ci_bdev_t *dev, char *kw, int n, cidev_entry_t *entry);
int cidev_keyword_count(ci_bdev_t *dev, char *kw, int *count);
char *get_string_val(char *src, char *key, int *len);

#else

/* Prototypes for functions in libcidev */
int cidev_inuse(int cfd, int *bool);
int cidev_read_header(int cfd, cidev_header_t *head);
int cidev_create(int cfd, char *name, cidev_entry_t *list);
int cidev_erase(int cfd);
int cidev_entry_count(int cfd, int *count);
int cidev_read_entries(int cfd, cidev_entry_t *list, int *count);
int cidev_read_entry_num(int cfd, int entry_num, cidev_entry_t *entry);
int cidev_read_entry(int cfd, int entry_id, cidev_entry_t *entry);
int cidev_write_entry(int cfd, int entry_id, cidev_entry_t *entry);
int cidev_remove_entry(int cfd, int entry_id);
int cidev_add_entry(int cfd, cidev_entry_t *entry);
int cidev_read_entry_keyword(int cfd, char *kw, int n, cidev_entry_t *entry);
int cidev_keyword_count(int cfd, char *kw, int *count);
char *get_string_val(char *src, char *key, int *len);

#endif /* KERNEL */

/* Prototypes for ondisk routines */
void cidev_header_in(cidev_header_t *ch, char *buf);
void cidev_header_out(cidev_header_t *ch, char *buf);
void cidev_marker_in(cidev_marker_t *cm, char *buf);
void cidev_marker_out(cidev_marker_t *cm, char *buf);

#endif
