/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/**
 * gfs_eattr.c - user space extended attribute tool. This replaces seteattr
 * and geteattr as a tool to get and set extended attributes on a gfs
 * file.
 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <errno.h>
#include <dirent.h>
#include <ctype.h>

#include "libgfs_ea.h"
#include "gfs_release.cf"
#include "copyright.cf"

int newline = 0;

#define printe(fmt, args...) fprintf(stderr, "gfs_eattr error: " fmt, ##args)
#define print(fmt, args...) printf("gfs_eattr: " fmt, ##args)

static int get_fd(char *filename){
  int fd;
  fd = open(filename, O_RDONLY);
  if (fd < 0){
    printe("couldn't open %s : %s\n", filename, strerror(errno));
    exit(1);
  }
  return fd;
}

int get_list(int fd){
  char *list;
  char *listp;
  unsigned int list_len;

  if (gfs_listeattrsize(fd, &list_len) < 0){
    printe("couldn't get the length of the extended attribute name list : %s\n", gfs_eaerror(errno));
    return 1;
  }
  if (list_len == 0)
    return 0;
  list = (char *)malloc(list_len);
  if (!list){
    printe("couldn't allocate memory for the extended attribute name list : %s\n", gfs_eaerror(errno));
    return 1;
  }
  if (gfs_listeattr(fd, list, &list_len) < 0){
    printe("couldn't get the list of extended attribute names : %s\n", gfs_eaerror(errno));
    return 1;
  }
  listp = list;
  while (listp < list + list_len){
    printf("%s\n", listp);
    listp = listp + strlen(listp) + 1;
  }
  return 0;
}
 
int get_data(int fd, char *name){
  unsigned int data_len;
  char *data;

  if (gfs_geteattrsize(fd, name, &data_len) < 0){
    printe("cannot get extended attribute data size for %s : %s\n",
           name, gfs_eaerror(errno));
    return 1;
  }
  if (newline)
    data = (char *)malloc(data_len + 1);
  else
    data = (char *)malloc(data_len);
  if (!data){
    print("cannot allocate memory for extended attribute data\n");
    return 1;
  }
  if (gfs_geteattr(fd, name, data, &data_len) < 0){
    printe("cannot get extented attribute data for %s : %s\n",
           name, gfs_eaerror(errno));
    return 1;
  }
  if (newline)
    data[data_len++] = '\n';
  if(write(STDOUT_FILENO, data, data_len) != data_len){
    printe("error writing out data: %s\n", strerror(errno));
    return 1;
  }
  return 0;
}

int usage(FILE *stream){
  fprintf(stream, "Usage:\n\n");

  fprintf(stream, "gfs_eattr [options] -f <file>\n\n");
  
  fprintf(stream, "Options:\n");
  fprintf(stream,
"  -c <name> [-oi]  create a new extended attribute for the file\n"
"  -g <name> [-n]   get the value of the extended attribute\n"
"  -h               print this usage information\n"
"  -i <file>        file containing attribute value, or '-' for stdin\n"
"  -l               get a list of extended attribute names. (default)\n"
"  -n               print a newline after the value\n"
"  -o <value>       attribute value (read from stdin is not specified)\n"
"  -p <name> [-oi]  replace the value of an extended attribute for the file\n"
"  -r <name>        remove the extended attribute from the file\n"
"  -s <name> [-oi]  set the value of an extended attribute for the file\n"
"  -V               Version information\n\n");
  return 1;
}


unsigned int get_from_file(char *buf, int fd){
  int bytes;
  char test;
  unsigned int total = 0;
  while ( (bytes = read(fd, buf + total, GFS_EA_DATA_MAX - total)) != 0){
    if (bytes < 0 && errno != -EINTR){
      printe("couldn't read from stdin: %s\n", strerror(errno));
      exit(1);
    }
    total += bytes;
  }
  if (total == GFS_EA_DATA_MAX && read(fd, &test, 1) != 0){
      printe("input too large for ea\n");
      exit(1);
  }
  return total;
}

int main(int argc, char *argv[]){
  unsigned int data_len;
  int fd = -1;
  int cmd = 0;
  char *msg;
  int in_fd = -1;
  char *name = NULL;
  char *data = NULL;
  char data_buf[GFS_EA_DATA_MAX];
  int action = 0;
  int opt;
  while((opt = getopt(argc, argv, "f:o:g:c:p:s:r:i:lnhV")) != -1){
    switch(opt){
    case 'h':
      usage(stdout);
      return 0;
    case 'V':
      printf("%s %s (built %s %s)\n", argv[0],
             GFS_RELEASE_NAME, __DATE__, __TIME__);
      printf("%s\n", GFS_COPYRIGHT);
      return 0;
    case 'n':
      newline = 1;
      continue;
    case 'o':
      if (in_fd != -1)
        return usage(stderr);
      data = optarg;
      continue;
    case 'i':
      if (data)
        return usage(stderr);
      if(!strcmp(optarg, "-"))
        in_fd = STDIN_FILENO;
      else
        in_fd = get_fd(optarg);
      continue;
    case 'f':
      fd = get_fd(optarg);
      continue;
    case 'l':
      if (action != 0)
        return usage(stderr);
      action = opt;
      continue;
    case ':':
    case '?':
      fprintf(stderr, "Please use '-h' for usage.\n");
      return 1;
      break;
    default:
      if (action != 0) {
        fprintf(stderr, "Please use '-h' for usage.\n");
        return 1;
      }
      action = opt;
      name = optarg;
      continue;
    }
  }
  if (argc != optind || fd < 0)
    return usage(stderr);
  switch(action){
  case 'g':
    if (!name || data || in_fd != -1) 
      return usage(stderr);
    return get_data(fd, name);
  case 'c': case 'p': case 's':
    if (action == 'c'){
      cmd = GFS_EA_CREATE;
      msg = "create";
    }
    else if (action == 'p'){
      cmd = GFS_EA_REPLACE;
      msg = "replace the value for";
    }
    else if (action == 's'){
      cmd = GFS_EA_SET;
      msg = "set the value for";
    }
    else
      return usage(stderr);
    if (!name)
      return usage(stderr);
    if (!data){
      if (in_fd == -1)
        in_fd = STDIN_FILENO;
      data_len = get_from_file(data_buf, in_fd);
      data = data_buf;
    }
    else
      data_len = strlen(data);
    if (gfs_seteattr(fd, name, data, data_len, cmd) < 0){
      printe("cannot %s %s : %s\n", msg, name, gfs_eaerror(errno));
      return 1;
    }
    return 0;
  case 'r':
    if(!name || data || in_fd != -1)
      return usage(stderr);
    if (gfs_removeeattr(fd, name) < 0){
      printe("cannot remove %s : %s\n", name, gfs_eaerror(errno));
      return 1;
    }
    return 0;
  case 0: case 'l':
    if (name || data || in_fd != -1)
      return usage(stderr);
    return get_list(fd);
  case ':':
  case '?':
    fprintf(stderr, "Please use '-h' for usage.\n");
    return 1;
  default:
    fprintf(stderr, "Bad programmer! You forgot to catch the %c flag\n", opt);
    return 1;
  }
}











