/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/utsname.h>
#include <linux/version.h>
#include <linux/pagemap.h>
#include <linux/iobuf.h>
#include <linux/xattr.h>
#include <linux/posix_acl.h>

#include "acl.h"
#include "bmap.h"
#include "linux_dcache.h"
#include "dio.h"
#include "dir.h"
#include "glock.h"
#include "glops.h"
#include "file.h"
#include "linux_file.h"
#include "inode.h"
#include "linux_inode.h"
#include "linux_page.h"
#include "quota.h"
#include "rgrp.h"
#include "trans.h"
#include "eattr.h"

struct inode_operations gfs_file_iops;
struct inode_operations gfs_dir_iops;
struct inode_operations gfs_symlink_iops;
struct inode_operations gfs_dev_iops;
struct address_space_operations gfs_file_aops;

/**
 * inode_attr_in - Copy attributes from the dinode into the VFS inode
 * @ip: The GFS inode
 *
 */

static void
inode_attr_in(struct gfs_inode *ip, struct inode *ino)
{
	unsigned int mode;

	ino->i_ino = ip->i_num.no_formal_ino;

	switch (ip->i_di.di_type) {
	case GFS_FILE_REG:
		mode = S_IFREG;
		ino->i_rdev = NODEV;
		break;
	case GFS_FILE_DIR:
		mode = S_IFDIR;
		ino->i_rdev = NODEV;
		break;
	case GFS_FILE_LNK:
		mode = S_IFLNK;
		ino->i_rdev = NODEV;
		break;
	case GFS_FILE_BLK:
		mode = S_IFBLK;
		ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
		break;
	case GFS_FILE_CHR:
		mode = S_IFCHR;
		ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
		break;
	case GFS_FILE_FIFO:
		mode = S_IFIFO;
		ino->i_rdev = NODEV;
		break;
	case GFS_FILE_SOCK:
		mode = S_IFSOCK;
		ino->i_rdev = NODEV;
		break;
	default:
		GFS_ASSERT_INODE(FALSE, ip,
				 printk("type = %u\n", ip->i_di.di_type);
		    );
		break;
	};

	ino->i_mode = mode | (ip->i_di.di_mode & S_IALLUGO);
	ino->i_nlink =
	    (ip->i_di.di_nlink <
	     (uint32_t) ((nlink_t) - 1)) ? ip->i_di.di_nlink : ((nlink_t) - 1);
	ino->i_uid = ip->i_di.di_uid;
	ino->i_gid = ip->i_di.di_gid;
	ino->i_size = ip->i_di.di_size;
	ino->i_atime = ip->i_di.di_atime;
	ino->i_mtime = ip->i_di.di_mtime;
	ino->i_ctime = ip->i_di.di_ctime;
	ino->i_blksize = PAGE_SIZE;
	ino->i_blocks =
	    ip->i_di.di_blocks << (ip->i_sbd->sd_sb.sb_bsize_shift -
				   GFS_BASIC_BLOCK_SHIFT);
	ino->i_generation = ip->i_di.di_header.mh_incarn;
	ino->i_version = ++event;

}

/**
 * gfs_inode_attr_in - Copy attributes from the dinode into the VFS inode
 * @ip: The GFS inode
 *
 */

void
gfs_inode_attr_in(struct gfs_inode *ip)
{
	struct inode *ino;

	ino = gfs_iget(ip, NO_ST_CREATE);
	if (ino) {
		inode_attr_in(ip, ino);
		iput(ino);
	}

}

/**
 * gfs_inode_attr_out - Copy attributes from VFS inode into the dinode
 * @ip: The GFS inode
 *
 * Only copy out the attributes that we want the VFS layer
 * to be able to modify.
 */

void
gfs_inode_attr_out(struct gfs_inode *ip)
{
	struct inode *ino;

	ino = gfs_iget(ip, NO_ST_CREATE);
	if (ino) {
		ip->i_di.di_mode = ino->i_mode & S_IALLUGO;
		ip->i_di.di_uid = ino->i_uid;
		ip->i_di.di_gid = ino->i_gid;
		ip->i_di.di_atime = ino->i_atime;
		ip->i_di.di_mtime = ino->i_mtime;
		ip->i_di.di_ctime = ino->i_ctime;
		iput(ino);
	}

}

/**
 * gfs_iget - Get/Create a struct inode for a struct gfs_inode
 * @ip: the struct gfs_inode to get the struct inode for
 *
 * Returns: An inode
 */

struct inode *
gfs_iget(struct gfs_inode *ip, int create)
{
	struct inode *inode = NULL, *tmp;

	spin_lock(&ip->i_arch.i_lock);
	if (ip->i_vnode)
		inode = igrab(ip->i_vnode);
	spin_unlock(&ip->i_arch.i_lock);

	if (inode || !create)
		goto out;

	RETRY_MALLOC(tmp = new_inode(ip->i_sbd->sd_vfs), tmp);

	inode_attr_in(ip, tmp);

	if (ip->i_di.di_type == GFS_FILE_REG) {
		tmp->i_op = &gfs_file_iops;
		tmp->i_fop = (struct file_operations *)&gfs_file_fops;
		tmp->i_mapping->a_ops = &gfs_file_aops;
	} else if (ip->i_di.di_type == GFS_FILE_DIR) {
		tmp->i_op = &gfs_dir_iops;
		tmp->i_fop = &gfs_dir_fops;
	} else if (ip->i_di.di_type == GFS_FILE_LNK) {
		tmp->i_op = &gfs_symlink_iops;
	} else {
		tmp->i_op = &gfs_dev_iops;
		init_special_inode(tmp, tmp->i_mode, tmp->i_rdev);
	}

	tmp->u.generic_ip = NULL;

	for (;;) {
		spin_lock(&ip->i_arch.i_lock);
		if (!ip->i_vnode)
			break;
		inode = igrab(ip->i_vnode);
		spin_unlock(&ip->i_arch.i_lock);

		if (inode) {
			iput(tmp);
			goto out;
		}
		yield();
	}

	inode = tmp;

	gfs_hold_istruct(ip);

	ip->i_vnode = inode;
	inode->u.generic_ip = ip;

	insert_inode_hash(inode);

	spin_unlock(&ip->i_arch.i_lock);

      out:

	return inode;
}

/**
 * gfs_setattr - Change attributes on an inode
 * @dentry: The dentry which is changing
 * @attr: The structure describing the change
 *
 * The VFS layer wants to change one or more of an inodes attributes.  Write
 * that change out to disk.
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_setattr(struct dentry *dentry, struct iattr *attr)
{
	struct inode *inode = dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al;
	struct gfs_holder i_gh;
	struct buffer_head *dibh;
	uint32_t ouid, ogid, nuid, ngid;
	int error = 0;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_EXCLUSIVE, 0, &i_gh);
	if (error)
		goto fail;

	error = inode_change_ok(inode, attr);
	if (error)
		goto fail_gunlock;

	if (attr->ia_valid & ATTR_SIZE) {
		error = vmtruncate(inode, attr->ia_size);
		if (error)
			goto fail_gunlock;

		error =
		    gfs_truncatei(ip, attr->ia_size, gfs_truncator_page,
				  gfs_get_user_cred());
		if (error)
			goto fail_gunlock;
	}

	else if (attr->ia_valid & (ATTR_UID | ATTR_GID)) {
		ouid = ip->i_di.di_uid;
		ogid = ip->i_di.di_gid;
		nuid = attr->ia_uid;
		ngid = attr->ia_gid;

		if (!(attr->ia_valid & ATTR_UID) || ouid == nuid)
			ouid = nuid = NO_QUOTA_CHANGE;
		if (!(attr->ia_valid & ATTR_GID) || ogid == ngid)
			ogid = ngid = NO_QUOTA_CHANGE;

		al = gfs_get_alloc(ip);

		error =
		    gfs_get_quota_locks(ip, nuid, ngid, gfs_get_user_cred());
		if (error)
			goto fail_alloc;

		if (ouid != NO_QUOTA_CHANGE || ogid != NO_QUOTA_CHANGE) {
			error = gfs_check_quota(ip, nuid, ngid);
			if (error)
				goto fail_gunlock_q;
		}

		/*  Trans may require:
		   one dinode block and one quota change block  */

		error =
		    gfs_trans_begin(sdp, TRANS_SETATTR, 1 + GFS_MAX_EA_ACL_BLKS,
				    1);
		if (error)
			goto fail_gunlock_q;

		error = gfs_get_inode_buffer(ip, &dibh);
		if (error)
			goto fail_end_trans;

		if (ouid != NO_QUOTA_CHANGE || ogid != NO_QUOTA_CHANGE) {
			gfs_trans_add_quota(sdp, -ip->i_di.di_blocks, ouid,
					    ogid);
			gfs_trans_add_quota(sdp, ip->i_di.di_blocks, nuid,
					    ngid);
		}

		inode_setattr(inode, attr);
		gfs_inode_attr_out(ip);

		if (attr->ia_valid & ATTR_MODE)
			error = gfs_acl_setattr(inode);

		gfs_trans_add_bh(sdp, ip->i_gl, dibh);
		gfs_dinode_out(&ip->i_di, (dibh)->b_data);
		gfs_drelse(dibh);

		gfs_trans_end(sdp);

		gfs_put_quota_locks(ip);
		gfs_put_alloc(ip);
	}

	else {
		/*  Trans may require:
		   one dinode block.  */

		error =
		    gfs_trans_begin(sdp, TRANS_SETATTR, 1 + GFS_MAX_EA_ACL_BLKS,
				    0);
		if (error)
			goto fail_gunlock;

		error = gfs_get_inode_buffer(ip, &dibh);
		if (!error) {
			inode_setattr(inode, attr);
			gfs_inode_attr_out(ip);

			if (attr->ia_valid & ATTR_MODE)
				error = gfs_acl_setattr(inode);

			gfs_trans_add_bh(sdp, ip->i_gl, dibh);
			gfs_dinode_out(&ip->i_di, (dibh)->b_data);
			gfs_drelse(dibh);
		}

		gfs_trans_end(sdp);
	}

	gfs_glock_dq_uninit(&i_gh);

	mark_inode_dirty(inode);

	return error;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock_q:
	gfs_put_quota_locks(ip);

      fail_alloc:
	gfs_put_alloc(ip);

      fail_gunlock:
	gfs_glock_dq_uninit(&i_gh);

      fail:
	return error;
}

/**
 * lookup_cdpn_sub_at - Maybe lookup a Context Dependent Pathname
 * @sdp: the filesystem
 * @dentry: the original dentry to lookup
 * @new_dentry: the new dentry, if this was a substitutable path.
 *
 */

static void
lookup_cdpn_sub_at(struct gfs_sbd *sdp, struct dentry *dentry,
		   struct dentry **new_dentry)
{
	char buf[2 * __NEW_UTS_LEN + 2];

	if (gfs_filecmp(&dentry->d_name, "@hostname", 9))
		*new_dentry =
		    lookup_one_len(system_utsname.nodename, dentry->d_parent,
				   strlen(system_utsname.nodename));
	else if (gfs_filecmp(&dentry->d_name, "@mach", 5))
		*new_dentry =
		    lookup_one_len(system_utsname.machine, dentry->d_parent,
				   strlen(system_utsname.machine));
	else if (gfs_filecmp(&dentry->d_name, "@os", 3))
		*new_dentry =
		    lookup_one_len(system_utsname.sysname, dentry->d_parent,
				   strlen(system_utsname.sysname));
	else if (gfs_filecmp(&dentry->d_name, "@uid", 4))
		*new_dentry =
		    lookup_one_len(buf, dentry->d_parent,
				   sprintf(buf, "%u", current->fsuid));
	else if (gfs_filecmp(&dentry->d_name, "@gid", 4))
		*new_dentry =
		    lookup_one_len(buf, dentry->d_parent,
				   sprintf(buf, "%u", current->fsgid));
	else if (gfs_filecmp(&dentry->d_name, "@sys", 4))
		*new_dentry = lookup_one_len(buf, dentry->d_parent,
					     sprintf(buf, "%s_%s",
						     system_utsname.machine,
						     system_utsname.sysname));

	return;
}

/**
 * lookup_cdpn_sub_brace - Maybe lookup a Context Dependent Pathname
 * @sdp: the filesystem
 * @dentry: the original dentry to lookup
 * @new_dentry: the new dentry, if this was a substitutable path.
 *
 */

static void
lookup_cdpn_sub_brace(struct gfs_sbd *sdp, struct dentry *dentry,
		      struct dentry **new_dentry)
{
	char buf[2 * __NEW_UTS_LEN + 2];

	if (gfs_filecmp(&dentry->d_name, "{hostname}", 10))
		*new_dentry =
		    lookup_one_len(system_utsname.nodename, dentry->d_parent,
				   strlen(system_utsname.nodename));
	else if (gfs_filecmp(&dentry->d_name, "{mach}", 6))
		*new_dentry =
		    lookup_one_len(system_utsname.machine, dentry->d_parent,
				   strlen(system_utsname.machine));
	else if (gfs_filecmp(&dentry->d_name, "{os}", 4))
		*new_dentry =
		    lookup_one_len(system_utsname.sysname, dentry->d_parent,
				   strlen(system_utsname.sysname));
	else if (gfs_filecmp(&dentry->d_name, "{uid}", 5))
		*new_dentry =
		    lookup_one_len(buf, dentry->d_parent,
				   sprintf(buf, "%u", current->fsuid));
	else if (gfs_filecmp(&dentry->d_name, "{gid}", 5))
		*new_dentry =
		    lookup_one_len(buf, dentry->d_parent,
				   sprintf(buf, "%u", current->fsgid));
	else if (gfs_filecmp(&dentry->d_name, "{sys}", 5))
		*new_dentry = lookup_one_len(buf, dentry->d_parent,
					     sprintf(buf, "%s_%s",
						     system_utsname.machine,
						     system_utsname.sysname));

	return;
}

/**
 * gfs_lookup - Look up a filename in a directory and return its inode
 * @dir: The directory inode
 * @dentry: The dentry of the new inode
 *
 * Called by the VFS layer. Lock dir and call gfs_lookupi()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static struct dentry *
gfs_lookup(struct inode *dir, struct dentry *dentry)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip, *ip;
	struct gfs_holder d_gh, i_gh;
	struct inode *inode = NULL;
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	/*  Do Context Dependent Path Name expansion  */

	if (*dentry->d_name.name == '@' && dentry->d_name.len > 1) {
		struct dentry *new_dentry = NULL;

		lookup_cdpn_sub_at(dip->i_sbd, dentry, &new_dentry);
		if (new_dentry) {
			return new_dentry;
		}
	} else if (*dentry->d_name.name == '{' && dentry->d_name.len > 1) {
		struct dentry *new_dentry = NULL;

		lookup_cdpn_sub_brace(dip->i_sbd, dentry, &new_dentry);
		if (new_dentry) {
			return new_dentry;
		}
	}

	dentry->d_op = &gfs_dops;

	gfs_init_holder(dip->i_gl, 0, 0, &d_gh);

	error = gfs_lookupi(&d_gh, &dentry->d_name, gfs_get_user_cred(), &i_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		return ERR_PTR(error);
	}

	if (i_gh.gh_gl) {
		ip = (struct gfs_inode *) i_gh.gh_gl->gl_object;

		gfs_glock_dq_uninit(&d_gh);
		gfs_glock_dq_uninit(&i_gh);

		inode = gfs_iget(ip, ST_CREATE);
		gfs_put_istruct(ip);
	} else
		gfs_uninit_holder(&d_gh);

	d_add(dentry, inode);

	return NULL;
}

/**
 * gfs_create - Create a file
 * @dir: The directory in which to create the file
 * @dentry: The dentry of the new file
 * @mode: The mode of the new file
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip, *ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_holder d_gh, i_gh;
	struct inode *inode;
	int new;
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	gfs_init_holder(dip->i_gl, 0, 0, &d_gh);

	error =
	    gfs_createi(&d_gh, &dentry->d_name, GFS_FILE_REG, mode & S_IALLUGO,
			gfs_get_user_cred(), &new, &i_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		goto fail;
	}

	GFS_ASSERT_SBD(i_gh.gh_gl, sdp,);
	ip = (struct gfs_inode *) i_gh.gh_gl->gl_object;

	if (new) {
		gfs_trans_end(sdp);

		gfs_inplace_relse(dip);
		gfs_put_quota_locks(dip);
		gfs_put_alloc(dip);

		ip->i_arch.i_creat_task = current;
		ip->i_arch.i_creat_pid = current->pid;
	}

	gfs_glock_dq_uninit(&d_gh);
	gfs_glock_dq_uninit(&i_gh);

	inode = gfs_iget(ip, ST_CREATE);
	gfs_put_istruct(ip);

	d_instantiate(dentry, inode);
	if (new)
		mark_inode_dirty(inode);

	return 0;

      fail:
	return error;
}

/**
 * gfs_link - Link to a file
 * @old_dentry: The inode to link
 * @dir: Add link to this directory
 * @dentry: The name of the link
 *
 * Link the inode in "old_dentry" into the directory "dir" with the
 * name in "dentry".
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct inode *inode = old_dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_alloc *al = NULL;
	struct gfs_holder ghs[2];
	int alloc_required;
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	if (ip->i_di.di_type == GFS_FILE_DIR) {
		error = -EPERM;
		goto fail;
	}

	gfs_init_holder(dip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[0]);
	gfs_init_holder(ip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[1]);

	error = gfs_glock_nq_m(2, ghs);
	if (error)
		goto fail_uninit;

	error = permission(dir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	error = gfs_dir_search(dip, &dentry->d_name, NULL, NULL);
	switch (error) {
	case -ENOENT:
		break;
	case 0:
		error = -EEXIST;
	default:
		goto fail_gunlock;
	}

	if (!dip->i_di.di_nlink) {
		error = -EINVAL;
		goto fail_gunlock;
	}
	if (dip->i_di.di_entries == (uint32_t) - 1) {
		error = -EFBIG;
		goto fail_gunlock;
	}
	if (!ip->i_di.di_nlink) {
		error = -EINVAL;
		goto fail_gunlock;
	}
	if (ip->i_di.di_nlink == (uint32_t) - 1) {
		error = -EMLINK;
		goto fail_gunlock;
	}

	error =
	    gfs_diradd_alloc_required(dip, &dentry->d_name, &alloc_required);
	if (error)
		goto fail_gunlock;

	if (alloc_required) {
		al = gfs_get_alloc(dip);

		error =
		    gfs_get_quota_locks(dip, NO_QUOTA_CHANGE, NO_QUOTA_CHANGE,
					gfs_get_user_cred());
		if (error)
			goto fail_alloc;

		al->al_metablks = sdp->sd_max_dirres;
		al->al_trans = TRANS_LINK;

		error = gfs_inplace_reserv(dip);
		if (error)
			goto fail_gunlock_q;

		error =
		    gfs_check_quota(dip, dip->i_di.di_uid, dip->i_di.di_gid);
		if (error)
			goto fail_ipres;

		/*  Trans may require:
		   two dinode blocks, directory modifications to add an entry,
		   RG bitmap blocks to allocate from, and quota change  */

		error =
		    gfs_trans_begin(sdp, TRANS_LINK,
				    2 + sdp->sd_max_dirres +
				    al->al_rgd->rd_ri.ri_length, 1);
		if (error)
			goto fail_ipres;
	} else {
		/*  Trans may require:
		   Two dinode blocks and a leaf block.  */

		error = gfs_trans_begin(sdp, TRANS_LINK, 3, 0);
		if (error)
			goto fail_ipres;
	}

	error = gfs_dir_add(dip, &dentry->d_name, &ip->i_num, ip->i_di.di_type);
	if (error)
		goto fail_end_trans;

	error = gfs_change_nlink(ip, +1);
	if (error)
		goto fail_end_trans;

	gfs_trans_end(sdp);

	if (alloc_required) {
		GFS_ASSERT_INODE(al->al_metaalloced, dip,);
		gfs_inplace_relse(dip);
		gfs_put_quota_locks(dip);
		gfs_put_alloc(dip);
	}

	gfs_glock_dq_m(2, ghs);

	gfs_uninit_holder(&ghs[0]);
	gfs_uninit_holder(&ghs[1]);

	atomic_inc(&inode->i_count);

	d_instantiate(dentry, inode);
	mark_inode_dirty(inode);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_ipres:
	if (alloc_required)
		gfs_inplace_relse(dip);

      fail_gunlock_q:
	if (alloc_required)
		gfs_put_quota_locks(dip);

      fail_alloc:
	if (alloc_required)
		gfs_put_alloc(dip);

      fail_gunlock:
	gfs_glock_dq_m(2, ghs);

      fail_uninit:
	gfs_uninit_holder(&ghs[0]);
	gfs_uninit_holder(&ghs[1]);

      fail:
	return error;
}

/**
 * gfs_unlink - Unlink a file
 * @dir: The inode of the directory containing the file to unlink
 * @dentry: The file itself
 *
 * Unlink a file.  Call gfs_unlinki()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_unlink(struct inode *dir, struct dentry *dentry)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_inode *ip =
	    (struct gfs_inode *) dentry->d_inode->u.generic_ip;
	struct gfs_holder ghs[2];
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	gfs_limit_unlinked(sdp);

	gfs_init_holder(dip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[0]);
	gfs_init_holder(ip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[1]);

	error = gfs_glock_nq_m(2, ghs);
	if (error)
		goto fail;

	error = permission(dir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	if ((dip->i_di.di_mode & S_ISVTX) &&
	    dip->i_di.di_uid != current->fsuid &&
	    ip->i_di.di_uid != current->fsuid && !capable(CAP_FOWNER)) {
		error = -EPERM;
		goto fail_gunlock;
	}

	error = gfs_revalidate(dip, &dentry->d_name, ip);
	if (error)
		goto fail_gunlock;

	/*  Trans may require:
	   Two dinode blocks and one modified directory leaf block
	   and one unlinked tag.  */

	error = gfs_trans_begin(sdp, TRANS_UNLINK, 3, 1);
	if (error)
		goto fail_gunlock;

	error = gfs_unlinki(dip, &dentry->d_name, ip);
	if (error)
		goto fail_end_trans;

	gfs_trans_end(sdp);

	gfs_glock_dq_m(2, ghs);

	gfs_uninit_holder(&ghs[0]);
	gfs_uninit_holder(&ghs[1]);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock:
	gfs_glock_dq_m(2, ghs);

      fail:
	gfs_uninit_holder(&ghs[0]);
	gfs_uninit_holder(&ghs[1]);

	return error;
}

/**
 * gfs_symlink - Create a symlink
 * @dir: The directory to create the symlink in
 * @dentry: The dentry to put the symlink in
 * @symname: The thing which the link points to
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip, *ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_holder d_gh, i_gh;
	struct inode *inode;
	struct buffer_head *dibh;
	int size;
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	/*  Must be stuffed with a null terminator for gfs_follow_link()  */

	size = strlen(symname);
	if (size > sdp->sd_sb.sb_bsize - sizeof (struct gfs_dinode) - 1) {
		error = -ENAMETOOLONG;
		goto fail;
	}

	gfs_init_holder(dip->i_gl, 0, 0, &d_gh);

	error =
	    gfs_createi(&d_gh, &dentry->d_name, GFS_FILE_LNK, 0777,
			gfs_get_user_cred(), NULL, &i_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		goto fail;
	}

	GFS_ASSERT_SBD(i_gh.gh_gl, sdp,);
	ip = (struct gfs_inode *) i_gh.gh_gl->gl_object;

	ip->i_di.di_size = size;

	error = gfs_get_inode_buffer(ip, &dibh);
	GFS_ASSERT_INODE(!error, ip,);

	gfs_dinode_out(&ip->i_di, (dibh)->b_data);
	memcpy((dibh)->b_data + sizeof (struct gfs_dinode), symname, size);

	gfs_drelse(dibh);

	gfs_trans_end(sdp);

	gfs_inplace_relse(dip);
	gfs_put_quota_locks(dip);
	gfs_put_alloc(dip);

	gfs_glock_dq_uninit(&d_gh);
	gfs_glock_dq_uninit(&i_gh);

	inode = gfs_iget(ip, ST_CREATE);
	gfs_put_istruct(ip);

	d_instantiate(dentry, inode);
	mark_inode_dirty(inode);

	return 0;

      fail:
	return error;
}

/**
 * gfs_mkdir - Make a directory
 * @dir: The parent directory of the new one
 * @dentry: The dentry of the new directory
 * @mode: The mode of the new directory
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip, *ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_holder d_gh, i_gh;
	struct inode *inode;
	struct buffer_head *dibh;
	struct gfs_dinode *di;
	struct gfs_dirent *dent;
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	gfs_init_holder(dip->i_gl, 0, 0, &d_gh);

	error =
	    gfs_createi(&d_gh, &dentry->d_name, GFS_FILE_DIR, mode & S_IALLUGO,
			gfs_get_user_cred(), NULL, &i_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		goto fail;
	}

	GFS_ASSERT_SBD(i_gh.gh_gl, sdp,);
	ip = (struct gfs_inode *) i_gh.gh_gl->gl_object;

	ip->i_di.di_nlink = 2;
	ip->i_di.di_size = sdp->sd_sb.sb_bsize - sizeof (struct gfs_dinode);
	ip->i_di.di_flags |= GFS_DIF_JDATA;
	ip->i_di.di_payload_format = GFS_FORMAT_DE;
	ip->i_di.di_entries = 2;

	error = gfs_get_inode_buffer(ip, &dibh);
	GFS_ASSERT_INODE(!error, ip,);

	di = (struct gfs_dinode *) (dibh)->b_data;

	error = gfs_dirent_alloc(ip, dibh, 1, &dent);
	GFS_ASSERT_INODE(!error, ip,);	/*  This should never fail  */

	dent->de_inum = di->di_num;	/*  already GFS endian  */
	dent->de_hash = gfs_dir_hash(".", 1);
	dent->de_hash = cpu_to_gfs32(dent->de_hash);
	dent->de_type = cpu_to_gfs16(GFS_FILE_DIR);
	memcpy((char *) (dent + 1), ".", 1);
	di->di_entries = cpu_to_gfs32(1);

	error = gfs_dirent_alloc(ip, dibh, 2, &dent);
	GFS_ASSERT_INODE(!error, ip,);	/*  This should never fail  */

	gfs_inum_out(&dip->i_num, (char *) &dent->de_inum);
	dent->de_hash = gfs_dir_hash("..", 2);
	dent->de_hash = cpu_to_gfs32(dent->de_hash);
	dent->de_type = cpu_to_gfs16(GFS_FILE_DIR);
	memcpy((char *) (dent + 1), "..", 2);

	gfs_dinode_out(&ip->i_di, (char *) di);

	gfs_drelse(dibh);

	error = gfs_change_nlink(dip, +1);
	GFS_ASSERT_INODE(!error, dip,);	/*  dip already pinned  */

	gfs_trans_end(sdp);

	gfs_inplace_relse(dip);
	gfs_put_quota_locks(dip);
	gfs_put_alloc(dip);

	gfs_glock_dq_uninit(&d_gh);
	gfs_glock_dq_uninit(&i_gh);

	inode = gfs_iget(ip, ST_CREATE);
	gfs_put_istruct(ip);

	d_instantiate(dentry, inode);
	mark_inode_dirty(inode);

	return 0;

      fail:
	return error;
}

/**
 * gfs_rmdir - Remove a directory
 * @dir: The parent directory of the directory to be removed
 * @dentry: The dentry of the directory to remove
 *
 * Remove a directory. Call gfs_rmdiri()
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_rmdir(struct inode *dir, struct dentry *dentry)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_inode *ip =
	    (struct gfs_inode *) dentry->d_inode->u.generic_ip;
	struct gfs_holder ghs[2];
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	gfs_limit_unlinked(sdp);

	gfs_init_holder(dip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[0]);
	gfs_init_holder(ip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[1]);

	error = gfs_glock_nq_m(2, ghs);
	if (error)
		goto fail;

	error = permission(dir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	if ((dip->i_di.di_mode & S_ISVTX) &&
	    dip->i_di.di_uid != current->fsuid &&
	    ip->i_di.di_uid != current->fsuid && !capable(CAP_FOWNER)) {
		error = -EPERM;
		goto fail_gunlock;
	}

	error = gfs_revalidate(dip, &dentry->d_name, ip);
	if (error)
		goto fail_gunlock;

	GFS_ASSERT_INODE(ip->i_di.di_entries >= 2, ip,
			 gfs_dinode_print(&ip->i_di, TRUE);
	    );
	if (ip->i_di.di_entries > 2) {
		error = -ENOTEMPTY;
		goto fail_gunlock;
	}

	/*  Trans may require:
	   Two dinode blocks, one directory leaf block containing the
	   entry to be rmdired, two leaf blocks containing . and .. of
	   the directory being rmdired, and one unlinked tag  */

	error = gfs_trans_begin(sdp, TRANS_RMDIR, 5, 1);
	if (error)
		goto fail_gunlock;

	error = gfs_rmdiri(dip, &dentry->d_name, ip);
	if (error)
		goto fail_end_trans;

	gfs_trans_end(sdp);

	gfs_glock_dq_m(2, ghs);

	gfs_uninit_holder(&ghs[0]);
	gfs_uninit_holder(&ghs[1]);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_gunlock:
	gfs_glock_dq_m(2, ghs);

      fail:
	gfs_uninit_holder(&ghs[0]);
	gfs_uninit_holder(&ghs[1]);

	return error;
}

/**
 * gfs_mknod - Make a special file
 * @dir: The directory in which the special file will reside
 * @dentry: The dentry of the special file
 * @mode: The mode of the special file
 * @rdev: The device specification of the special file
 *
 */

static int
gfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int rdev)
{
	struct gfs_inode *dip = (struct gfs_inode *) dir->u.generic_ip, *ip;
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_holder d_gh, i_gh;
	struct inode *inode;
	struct buffer_head *dibh;
	kdev_t dev;
	uint16_t type = 0;
	uint32_t major = 0, minor = 0;
	int error;

	GFS_ASSERT_INODE(dentry->d_parent->d_inode == dir, dip,);

	switch (mode & S_IFMT) {
	case S_IFBLK:
		type = GFS_FILE_BLK;
		dev = to_kdev_t(rdev);
		major = MAJOR(dev);
		minor = MINOR(dev);
		break;
	case S_IFCHR:
		type = GFS_FILE_CHR;
		dev = to_kdev_t(rdev);
		major = MAJOR(dev);
		minor = MINOR(dev);
		break;
	case S_IFIFO:
		type = GFS_FILE_FIFO;
		break;
	case S_IFSOCK:
		type = GFS_FILE_SOCK;
		break;
	default:
		GFS_ASSERT_SBD(FALSE, sdp, printk("mode = %d\n", mode);
		    );
		break;
	};

	gfs_init_holder(dip->i_gl, 0, 0, &d_gh);

	error =
	    gfs_createi(&d_gh, &dentry->d_name, type, mode & S_IALLUGO,
			gfs_get_user_cred(), NULL, &i_gh);
	if (error) {
		gfs_uninit_holder(&d_gh);
		goto fail;
	}

	GFS_ASSERT_SBD(i_gh.gh_gl, sdp,);
	ip = (struct gfs_inode *) i_gh.gh_gl->gl_object;

	ip->i_di.di_major = major;
	ip->i_di.di_minor = minor;

	error = gfs_get_inode_buffer(ip, &dibh);
	GFS_ASSERT_INODE(!error, ip,);

	gfs_dinode_out(&ip->i_di, (dibh)->b_data);

	gfs_drelse(dibh);

	gfs_trans_end(sdp);

	gfs_inplace_relse(dip);
	gfs_put_quota_locks(dip);
	gfs_put_alloc(dip);

	gfs_glock_dq_uninit(&d_gh);
	gfs_glock_dq_uninit(&i_gh);

	inode = gfs_iget(ip, ST_CREATE);
	gfs_put_istruct(ip);

	d_instantiate(dentry, inode);
	mark_inode_dirty(inode);

	return 0;

      fail:
	return error;
}

/**
 * gfs_rename - Rename a file
 * @odir: Parent directory of old file name
 * @odentry: The old dentry of the file
 * @ndir: Parent directory of new file name
 * @ndentry: The new dentry of the file
 *
 * Returns: 0 on success, -EXXXX on failure
 */

static int
gfs_rename(struct inode *odir, struct dentry *odentry,
	   struct inode *ndir, struct dentry *ndentry)
{
	struct gfs_inode *odip = (struct gfs_inode *) odir->u.generic_ip;
	struct gfs_inode *ndip = (struct gfs_inode *) ndir->u.generic_ip;
	struct gfs_inode *ip =
	    (struct gfs_inode *) odentry->d_inode->u.generic_ip;
	struct gfs_inode *nip = NULL;
	struct gfs_sbd *sdp = odip->i_sbd;
	struct qstr name;
	struct gfs_alloc *al;
	struct gfs_holder ghs[4], r_gh;
	unsigned int num_gh;
	int dir_rename = FALSE;
	int alloc_required;
	unsigned int x;
	int error;

	GFS_ASSERT_INODE(odentry->d_parent->d_inode == odir, odip,);
	GFS_ASSERT_INODE(ndentry->d_parent->d_inode == ndir, ndip,);

	gfs_limit_unlinked(sdp);

	if (ndentry->d_inode) {
		nip = (struct gfs_inode *) ndentry->d_inode->u.generic_ip;

		if (ip == nip) {
			error = 0;
			goto fail;
		}
	}

	/*  Make sure we aren't trying to move a dirctory into it's subdir  */

	if (ip->i_di.di_type == GFS_FILE_DIR && odip != ndip) {
		dir_rename = TRUE;

		error =
		    gfs_glock_nq_init(sdp->sd_rename_gl, LM_ST_EXCLUSIVE, 0,
				      &r_gh);
		if (error)
			goto fail;

		error = gfs_ok_to_move(ip, ndip);
		if (error)
			goto fail_renlock;
	}

	gfs_init_holder(odip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[0]);
	gfs_init_holder(ndip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[1]);
	num_gh = 2;

	if (nip)
		gfs_init_holder(nip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[num_gh++]);

	if (dir_rename)
		gfs_init_holder(ip->i_gl, LM_ST_EXCLUSIVE, 0, &ghs[num_gh++]);

	error = gfs_glock_nq_m(num_gh, ghs);
	if (error)
		goto fail_uninit;

	/*  Check out the old directory  */

	error = permission(odir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	if ((odip->i_di.di_mode & S_ISVTX) &&
	    odip->i_di.di_uid != current->fsuid &&
	    ip->i_di.di_uid != current->fsuid && !capable(CAP_FOWNER)) {
		error = -EPERM;
		goto fail_gunlock;
	}

	error = gfs_revalidate(odip, &odentry->d_name, ip);
	if (error)
		goto fail_gunlock;

	/*  Check out the new directory  */

	error = permission(ndir, MAY_WRITE | MAY_EXEC);
	if (error)
		goto fail_gunlock;

	if (nip) {
		if ((ndip->i_di.di_mode & S_ISVTX) &&
		    ndip->i_di.di_uid != current->fsuid &&
		    nip->i_di.di_uid != current->fsuid &&
		    !capable(CAP_FOWNER)) {
			error = -EPERM;
			goto fail_gunlock;
		}

		error = gfs_revalidate(ndip, &ndentry->d_name, nip);
		if (error)
			goto fail_gunlock;

		if (nip->i_di.di_type == GFS_FILE_DIR) {
			GFS_ASSERT_INODE(nip->i_di.di_entries >= 2, ip,
					 gfs_dinode_print(&nip->i_di, TRUE);
			    );
			if (nip->i_di.di_entries > 2) {
				error = -ENOTEMPTY;
				goto fail_gunlock;
			}
		}
	} else {
		error = gfs_dir_search(ndip, &ndentry->d_name, NULL, NULL);
		switch (error) {
		case -ENOENT:
			error = 0;
			break;
		case 0:
			error = -EEXIST;
		default:
			goto fail_gunlock;
		};

		if (odip != ndip) {
			if (!ndip->i_di.di_nlink) {
				error = -EINVAL;
				goto fail_gunlock;
			}
			if (ndip->i_di.di_entries == (uint32_t) - 1) {
				error = -EFBIG;
				goto fail_gunlock;
			}
			if (ip->i_di.di_type == GFS_FILE_DIR
			    && ndip->i_di.di_nlink == (uint32_t) - 1) {
				error = -EMLINK;
				goto fail_gunlock;
			}
		}
	}

	error =
	    gfs_diradd_alloc_required(ndip, &ndentry->d_name, &alloc_required);
	if (error)
		goto fail_gunlock;

	if (alloc_required) {
		al = gfs_get_alloc(ndip);

		error =
		    gfs_get_quota_locks(ndip, NO_QUOTA_CHANGE, NO_QUOTA_CHANGE,
					gfs_get_user_cred());
		if (error)
			goto fail_alloc;

		al->al_metablks = sdp->sd_max_dirres;
		al->al_trans = TRANS_RENAME;

		error = gfs_inplace_reserv(ndip);
		if (error)
			goto fail_gunlock_q;

		error =
		    gfs_check_quota(ndip, ndip->i_di.di_uid, ndip->i_di.di_gid);
		if (error)
			goto fail_ipres;

		/*  Trans may require:
		   Dinodes for the srcdir, srcino, dstdir, dstino.  Blocks for
		   adding the entry to dstdir.  RG bitmaps for that allocation.
		   One leaf block in the srcdir for removal of the entry.
		   One leaf block for changing .. in srcino (if it's a directory).
		   Two leaf blocks for removing . and .. from dstino (if it exists
		   and it's a directory), one unlinked tag, and one quota block.  */

		error =
		    gfs_trans_begin(sdp, TRANS_RENAME,
				    8 + sdp->sd_max_dirres +
				    al->al_rgd->rd_ri.ri_length, 2);
		if (error)
			goto fail_ipres;
	} else {
		/*  Trans may require:
		   Dinodes for the srcdir, srcino, dstdir, dstino.  One block for
		   adding the entry to dstdir.
		   One leaf block in the srcdir for removal of the entry.
		   One leaf block for changing .. in srcino (if it's a directory).
		   Two leaf blocks for removing . and .. from dstino (if it exists
		   and it's a directory), and one unlinked tag.  */

		error = gfs_trans_begin(sdp, TRANS_RENAME, 9, 1);
		if (error)
			goto fail_ipres;
	}

	/*  Remove the target file, if it exists  */

	if (nip) {
		if (nip->i_di.di_type == GFS_FILE_DIR)
			error = gfs_rmdiri(ndip, &ndentry->d_name, nip);
		else
			error = gfs_unlinki(ndip, &ndentry->d_name, nip);

		if (error)
			goto fail_end_trans;
	}

	if (dir_rename) {
		error = gfs_change_nlink(ndip, +1);
		if (error)
			goto fail_end_trans;
		error = gfs_change_nlink(odip, -1);
		if (error)
			goto fail_end_trans;

		name.len = 2;
		name.name = "..";

		error = gfs_dir_mvino(ip, &name, &ndip->i_num, GFS_FILE_DIR);
		if (error)
			goto fail_end_trans;
	}

	error = gfs_dir_del(odip, &odentry->d_name);
	if (error)
		goto fail_end_trans;

	error =
	    gfs_dir_add(ndip, &ndentry->d_name, &ip->i_num, ip->i_di.di_type);
	if (error)
		goto fail_end_trans;

	if (dir_rename)
		gfs_trans_add_gl(sdp, sdp->sd_rename_gl);

	gfs_trans_end(sdp);

	if (alloc_required) {
		/*  Don't check al->al_metaalloced and friends.  */
		gfs_inplace_relse(ndip);
		gfs_put_quota_locks(ndip);
		gfs_put_alloc(ndip);
	}

	gfs_glock_dq_m(num_gh, ghs);

	for (x = 0; x < num_gh; x++)
		gfs_uninit_holder(&ghs[x]);

	if (dir_rename)
		gfs_glock_dq_uninit(&r_gh);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_ipres:
	if (alloc_required)
		gfs_inplace_relse(ndip);

      fail_gunlock_q:
	if (alloc_required)
		gfs_put_quota_locks(ndip);

      fail_alloc:
	if (alloc_required)
		gfs_put_alloc(ndip);

      fail_gunlock:
	gfs_glock_dq_m(num_gh, ghs);

      fail_uninit:
	for (x = 0; x < num_gh; x++)
		gfs_uninit_holder(&ghs[x]);

      fail_renlock:
	if (dir_rename)
		gfs_glock_dq_uninit(&r_gh);

      fail:
	return error;
}

/**
 * gfs_irevalidate - Check inode hasn't changed
 * @dentry: The dentry
 *
 * Revalidate the inode. i.e. grab the Glock and sync the incore dinode
 * with the ondisk inode
 *
 * Returns: 0 on success, or error
 */

static int
gfs_irevalidate(struct dentry *dentry)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) dentry->d_inode->u.generic_ip;
	struct gfs_holder i_gh;
	int error;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
	if (!error)
		gfs_glock_dq_uninit(&i_gh);

	return error;
}

/**
 * gfs_readlink - Read the value of a symlink
 * @dentry: the symlink
 * @buf: the buffer to read the symlink data into
 * @size: the size of the buffer
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_readlink(struct dentry *dentry, char *user_buf, int user_size)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) dentry->d_inode->u.generic_ip;
	char array[GFS_FAST_NAME_SIZE], *buf = array;
	unsigned int len = GFS_FAST_NAME_SIZE;
	int error;

	error = gfs_readlinki(ip, &buf, &len);
	if (error)
		goto fail;

	GFS_ASSERT_INODE(len, ip,);

	if (user_size > len - 1)
		user_size = len - 1;

	error = (copy_to_user(user_buf, buf, user_size)) ? -EFAULT : 0;
	if (error)
		goto fail_gfree;

	if (buf != array)
		kfree(buf);

	return user_size;

      fail_gfree:
	if (buf != array)
		kfree(buf);

      fail:
	return error;
}

/**
 * gfs_follow_link - Follow a symbolic link
 * @dentry: The dentry of the link
 * @nd: Data that we pass to vfs_follow_link()
 *
 * This can handle symlinks of any size. It is optimised for symlinks
 * under GFS_FAST_NAME_SIZE.
 *
 * Returns: 0 on success or error code
 */

static int
gfs_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) dentry->d_inode->u.generic_ip;
	char array[GFS_FAST_NAME_SIZE], *buf = array;
	unsigned int len = GFS_FAST_NAME_SIZE;
	int error;

	error = gfs_readlinki(ip, &buf, &len);
	if (!error) {
		error = vfs_follow_link(nd, buf);
		if (buf != array)
			kfree(buf);
	}

	return error;
}

/**
 * get_block - Fills in a buffer head with details about a block
 * @inode: The inode
 * @lblock: The block number to look up
 * @bh_result: The buffer head to return the result in
 * @create: Non-zero if we may add block to the file
 *
 */

static int
get_block(struct inode *inode, long lblock, struct buffer_head *bh_result,
	  int create)
{
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	int new = create;
	uint64_t dblock;
	int error;

	GFS_ASSERT_INODE(gfs_glock_is_locked_by_me(ip->i_gl)
			 || (!create && gfs_glock_is_held_excl(ip->i_gl)), ip,);

	if (lblock < 0) {
		printk("GFS: fsid=%s: get_block() called with lblock < 0\n",
		       sdp->sd_fsname);
		error = -EINVAL;
		goto out;
	}

	error = gfs_block_map(ip, (uint64_t) lblock, &new, &dblock, NULL);
	if (error)
		goto out;

	GFS_ASSERT_INODE(dblock || !create, ip,);

	if (!dblock)
		goto out;

#if (BITS_PER_LONG < 64)
	if (dblock & 0xFFFFFFFF00000000ll) {
		printk
		    ("GFS: fsid=%s: get_block(): block maps out of \"unsigned long\" range\n",
		     sdp->sd_fsname);
		error = -EINVAL;
		goto out;
	}
#endif

	bh_result->b_dev = inode->i_dev;
	bh_result->b_blocknr = (unsigned long) dblock;
	bh_result->b_state |= 1UL << BH_Mapped;
	if (new)
		bh_result->b_state |= 1UL << BH_New;

      out:

	return error;
}

/**
 * get_block_noalloc - Fills in a buffer head with details about a block.
 * This version is only called by writepage().
 * @inode: The inode
 * @lblock: The block number to look up
 * @bh_result: The buffer head to return the result in
 * @create: Non-zero if we may add block to the file - this field is supplied
 * by the VFS, and we always ignore it. Our block must have been allocated by
 * nopage to avoid deadlocks.
 *
 * Buffers are normally mapped during writepage, but we can't guarantee
 * this when multiple processes are syncing a file, as occurs with 
 * direct IO.
 *
 * Returns:  0 on success, -EXXX on failure
 */

static int
get_block_noalloc(struct inode *inode, long lblock,
		  struct buffer_head *bh_result, int create)
{
	int error;

	error = get_block(inode, lblock, bh_result, 0);
	GFS_ASSERT_INODE(buffer_mapped(bh_result),
			 (struct gfs_inode *) inode->u.generic_ip,);

	return error;
}

/**
 * stuffed_readpage - Fill in a Linux page with stuffed file data
 * @ip: the inode
 * @page: the page
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
stuffed_readpage(struct gfs_inode *ip, struct page *page)
{
	struct buffer_head *dibh;
	void *kaddr;
	int error;

	GFS_ASSERT_INODE(PageLocked(page), ip,);

	error = gfs_get_inode_buffer(ip, &dibh);
	if (!error) {
		kaddr = kmap(page);
		memcpy((char *) kaddr,
		       (dibh)->b_data + sizeof (struct gfs_dinode),
		       ip->i_di.di_size);
		memset((char *) kaddr + ip->i_di.di_size, 0,
		       PAGE_CACHE_SIZE - ip->i_di.di_size);
		kunmap(page);

		gfs_drelse(dibh);

		SetPageUptodate(page);
	}

	return error;
}

/**
 * readi_readpage - readpage that goes through gfs_internal_read()
 * @page: The page to read
 *
 * Returns: 0 on success, error code otherwise
 */

static int
readi_readpage(struct page *page)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) page->mapping->host->u.generic_ip;
	void *kaddr;
	int ret;

	kaddr = kmap(page);

	ret =
	    gfs_internal_read(ip, kaddr,
			      (uint64_t) page->index << PAGE_CACHE_SHIFT,
			      PAGE_CACHE_SIZE);
	if (ret >= 0) {
		if (ret < PAGE_CACHE_SIZE)
			memset(kaddr + ret, 0, PAGE_CACHE_SIZE - ret);
		SetPageUptodate(page);
		ret = 0;
	}

	kunmap(page);

	UnlockPage(page);

	return ret;
}

/**
 * gfs_readpage - readpage with locking
 * @file: The file to read a page for
 * @page: The page to read
 *
 * Returns: 0 on success, error code otherwise
 */

static int
gfs_readpage(struct file *file, struct page *page)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) page->mapping->host->u.generic_ip;
	int error;

	if (!gfs_glock_is_locked_by_me(ip->i_gl)) {
		UnlockPage(page);
		bitch_about(ip->i_sbd, &ip->i_sbd->sd_last_unlocked_aop,
			    "unpatched unlocked readpage request");
		return -ENOSYS;
	}

	if (!gfs_is_jdata(ip)) {
		if (gfs_is_stuffed(ip) && !page->index) {
			error = stuffed_readpage(ip, page);
			UnlockPage(page);
		} else
			error = block_read_full_page(page, get_block);
	} else
		error = readi_readpage(page);

	return error;
}

/**
 * gfs_writepage - Write complete page
 * @page: Page to write
 *
 */

static int
gfs_writepage(struct page *page)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) page->mapping->host->u.generic_ip;
	int error;

	GFS_ASSERT_INODE(gfs_glock_is_held_excl(ip->i_gl)
			 && !gfs_is_stuffed(ip), ip,);

	error = block_write_full_page(page, get_block_noalloc);

	gfs_flush_meta_cache(ip);

	return error;
}

/**
 * gfs_prepare_write - Prepare to write to a file
 * @file: The file to write to
 * @page: The page which is to be prepared for writing
 * @from: From (byte range within page)
 * @to: To (byte range within page)
 *
 */

static int
gfs_prepare_write(struct file *file, struct page *page, unsigned from,
		  unsigned to)
{
	struct gfs_inode *ip =
	    (struct gfs_inode *) page->mapping->host->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	uint64_t file_size = ((uint64_t) page->index << PAGE_CACHE_SHIFT) + to;
	int error = 0;

	if (!gfs_glock_is_locked_by_me(ip->i_gl)) {
		bitch_about(sdp, &sdp->sd_last_unlocked_aop,
			    "unpatched unlocked prepare_write request");
		error = -ENOSYS;
		goto out;
	}

	error = gfs_check_quota(ip, ip->i_di.di_uid, ip->i_di.di_gid);
	if (error)
		goto out;

	if (gfs_is_stuffed(ip)) {
		if (file_size >
		    sdp->sd_sb.sb_bsize - sizeof (struct gfs_dinode)) {
			error =
			    gfs_unstuff_dinode(ip, gfs_unstuffer_page, page);
			if (!error)
				error =
				    block_prepare_write(page, from, to,
							get_block);
		} else if (!Page_Uptodate(page))
			error = stuffed_readpage(ip, page);
	} else
		error = block_prepare_write(page, from, to, get_block);

      out:

	return error;
}

/**
 * gfs_commit_write - Commit write to a file
 * @file: The file to write to
 * @page: The page containing the data
 * @from: From (byte range within page)
 * @to: To (byte range within page)
 *
 */

static int
gfs_commit_write(struct file *file, struct page *page, unsigned from,
		 unsigned to)
{
	struct inode *inode = page->mapping->host;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	struct buffer_head *dibh;
	uint64_t file_size = ((uint64_t) page->index << PAGE_CACHE_SHIFT) + to;
	void *kaddr;
	int error = 0;

	if (gfs_is_stuffed(ip)) {
		GFS_ASSERT_INODE(PageLocked(page), ip,);

		error = gfs_get_inode_buffer(ip, &dibh);
		if (error)
			goto fail;

		gfs_trans_add_bh(sdp, ip->i_gl, dibh);

		kaddr = kmap(page);
		memcpy((dibh)->b_data + sizeof (struct gfs_dinode) + from,
		       (char *) kaddr + from, to - from);
		kunmap(page);

		gfs_drelse(dibh);

		SetPageUptodate(page);

		if (inode->i_size < file_size)
			inode->i_size = file_size;
	} else {
		error = generic_commit_write(file, page, from, to);
		if (error)
			goto fail;
	}

	return 0;

      fail:
	ClearPageUptodate(page);

	return error;
}

/**
 * gfs_bmap - Block map function
 * @mapping: Address space info
 * @lblock: The block to map
 *
 * Returns:  The disk address for the block or 0 on hole or error
 */

static int
gfs_bmap(struct address_space *mapping, long lblock)
{
	struct gfs_inode *ip = (struct gfs_inode *) mapping->host->u.generic_ip;
	struct gfs_holder i_gh;
	int dblock = 0;
	int error;

	error = gfs_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY, &i_gh);
	if (error)
		goto out;

	if (!gfs_is_stuffed(ip))
		dblock = generic_block_bmap(mapping, lblock, get_block);

	gfs_glock_dq_uninit(&i_gh);

      out:

	return dblock;
}

/**
 * functionname - summary
 * @param1: description
 * @param2: description
 * @param3: description
 *
 * Function description
 *
 * Returns: what is returned
 */

int
gfs_access_permission(struct inode *inode, int mask)
{
	int error = 0;
	uid_t user = current->fsuid;
	umode_t mode = inode->i_mode;
	struct posix_acl *acl;

	if (mask & MAY_WRITE) {
		if (IS_RDONLY(inode)
		    && (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode))) {
			error = -EROFS;
			goto out;
		}
		if (IS_IMMUTABLE(inode)) {
			error = -EACCES;
			goto out;
		}
	}

	if (capable(CAP_DAC_OVERRIDE))
		if (!(mask & MAY_EXEC) || (mode & S_IXUGO))
			goto out;

	if (capable(CAP_DAC_READ_SEARCH) &&
	    (mask == MAY_READ || (!(mask & MAY_WRITE) && S_ISDIR(mode))))
		goto out;

	if (user == inode->i_uid) {
		if ((mask & (mode >> 6)) != mask)
			error = -EACCES;
		goto out;
	}

	if ((mask & (mode >> 3)) == mask) {
		error = gfs_getacl(inode, TRUE, &acl);
		if (acl) {
			error = posix_acl_permission(inode, acl, mask);
			goto out;
		} else if (error && error != -ENODATA)
			goto out;
		error = 0;
		if (in_group_p(inode->i_gid)) {
			error = 0;
			goto out;
		}
	} else if (in_group_p(inode->i_gid)) {
		error = -EACCES;
		goto out;
	}

	if ((mask & mode) == mask)
		goto out;
	error = -EACCES;

      out:
	return error;
}

/**
 * gfs_direct_IO - kludge to tell the VFS we support Direct I/O
 * @rw:
 * @file:
 * @iobuf:
 * @blocknr:
 * @blocksize:
 *
 * This should never be called.  It panics if it is.
 *
 * Returns: -ENOSYS;
 */

static int
gfs_direct_IO(int rw, struct file *file, struct kiobuf *iobuf,
	      unsigned long blocknr, int blocksize)
{
	GFS_ASSERT_SBD(FALSE,
		       (struct gfs_sbd *) file->f_dentry->d_inode->i_sb->u.
		       generic_sbp,);
	return -ENOSYS;
}

/**
 * gfs_permission - check permissions on a file
 * @ip: the file
 * @mask: the type of access we want
 * @cred: the structure defining our credentials
 *
 * Returns: 0 for Permission Granted, -EXXX on failure
 */

int
gfs_permission(struct gfs_inode *ip, int mask, int *cred)
{
	struct inode *inode;
	int error = 0;

	if (mask & GFS_MAY_EXCEED) {
		if (!capable(CAP_SYS_RESOURCE))
			error = -EACCES;
	} else if (mask & GFS_MAY_ADMIN) {
		if (!capable(CAP_SYS_ADMIN))
			error = -EACCES;
	} else if (mask & GFS_IS_FOWNER) {
		if (((cred) ? current->fsuid : 0) != ip->i_di.di_uid
		    && !capable(CAP_FOWNER))
			error = -EACCES;
	} else if (mask & GFS_MAY_WRITE_USER_EA) {
		inode = gfs_iget(ip, NO_ST_CREATE);
		if (inode) {
			if (!S_ISREG(inode->i_mode)
			    && (!S_ISDIR(inode->i_mode)
				|| inode->i_mode & S_ISVTX))
				error = -EACCES;
			else
				error = permission(inode, MAY_WRITE);
			iput(inode);
		}
	} else if (mask & (GFS_MAY_READ | GFS_MAY_WRITE | GFS_MAY_EXEC)) {
		inode = gfs_iget(ip, NO_ST_CREATE);
		if (inode) {
			error =
			    permission(inode,
				       mask & (GFS_MAY_READ | GFS_MAY_WRITE |
					       GFS_MAY_EXEC));
			iput(inode);
		}
	}

	return error;
}

/**
 * gfs_try_toss_vnode - See if we can toss a vnode from memory
 * @ip: the inode
 *
 * Returns:  TRUE if the vnode was tossed
 */

void
gfs_try_toss_vnode(struct gfs_inode *ip)
{
	struct inode *inode;

	inode = gfs_iget(ip, NO_ST_CREATE);
	if (!inode)
		return;

	d_prune_aliases(inode);

	if (ip->i_di.di_type == GFS_FILE_DIR) {
		struct list_head *head = &inode->i_dentry;
		struct dentry *d = NULL;

		spin_lock(&dcache_lock);
		if (list_empty(head))
			spin_unlock(&dcache_lock);
		else {
			d = list_entry(head->next, struct dentry, d_alias);
			dget_locked(d);
			spin_unlock(&dcache_lock);

			if (have_submounts(d))
				dput(d);
			else {
				shrink_dcache_parent(d);
				dput(d);
				d_prune_aliases(inode);
			}
		}
	}

	inode->i_nlink = 0;
	iput(inode);

	return;
}

/**
 * get_eatype - get the type of the ea, and trucate the type from the name
 * @namep: ea name, possibly with type appended
 *
 * Returns: GFS_EATYPE_XXX
 */

int
get_eatype(const char *name, char **truncated_name)
{
	int type;

	if (strncmp(name, "system.", 7) == 0) {
		type = GFS_EATYPE_SYS;
		*truncated_name = strchr(name, '.') + 1;
	} else if (strncmp(name, "user.", 5) == 0) {
		type = GFS_EATYPE_USR;
		*truncated_name = strchr(name, '.') + 1;
	} else {
		type = GFS_EATYPE_UNUSED;
		*truncated_name = NULL;
	}

	return type;
}

/**
 * gfs_setxattr - Set (or create or replace) an inode's extended attribute
 * @dentry: inode's dentry
 * @name: name of the extended attribute
 * @data: the value of the extended attribute
 * @size: the size of data
 * @flags: used to specify create or replace actions
 *
 * Returns:  0 on success, -EXXX on error
 */

int
gfs_setxattr(struct dentry *dentry, const char *name, const void *data,
	     size_t size, int flags)
{
	struct gfs_easet_io req;
	int err = 0;
	struct inode *inode = dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	char *truncated_name;

	req.es_type = get_eatype(name, &truncated_name);
	if (req.es_type == GFS_EATYPE_UNUSED)
		err = -EOPNOTSUPP;
	else {
		req.es_data = (void *) data;
		req.es_name = truncated_name;
		req.es_data_len = size;
		req.es_name_len = strlen(truncated_name);
		req.es_cmd =
		    (flags & XATTR_CREATE) ? GFS_EACMD_CREATE
		    : ((flags & XATTR_REPLACE) ? GFS_EACMD_REPLACE :
		       GFS_EACMD_SET);
		err = gfs_set_eattr(sdp, ip, &req, gfs_get_user_cred());
	}

	return err;
}

/**
 * gfs_getxattr -
 * @dentry:
 * @name:
 * @data:
 * @size:
 *
 * Returns:  0 on success, -EXXX on error
 */

ssize_t
gfs_getxattr(struct dentry * dentry, const char *name, void *data, size_t size)
{
	struct gfs_eaget_io req;
	int err = 0;
	struct inode *inode = dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	char *truncated_name;

	req.eg_type = get_eatype(name, &truncated_name);
	if (req.eg_type == GFS_EATYPE_UNUSED)
		err = -EOPNOTSUPP;
	else {
		req.eg_name = truncated_name;
		req.eg_name_len = strlen(truncated_name);
		req.eg_data = data;
		req.eg_data_len = size;
		req.eg_len = NULL;
		err =
		    gfs_get_eattr(sdp, ip, &req, gfs_ea_memcpy,
				  gfs_get_user_cred());
	}

	return err;
}

/**
 * gfs_listxattr - 
 * @dentry:
 * @buffer:
 * @size:
 *
 * Returns:  0 on success, -EXXX on error
 */

ssize_t
gfs_listxattr(struct dentry * dentry, char *buffer, size_t size)
{
	struct gfs_eaget_io req;
	int err = 0;
	struct inode *inode = dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;

	req.eg_type = 0;
	req.eg_name = NULL;
	req.eg_name_len = 0;
	req.eg_data = buffer;
	req.eg_data_len = size;
	req.eg_len = NULL;

	err = gfs_get_eattr(sdp, ip, &req, gfs_ea_memcpy, gfs_get_user_cred());

	return err;
}

/**
 * gfs_removexattr -
 * @dentry:
 * @name:
 *
 * Returns:  0 on success, -EXXX on error
 */

int
gfs_removexattr(struct dentry *dentry, const char *name)
{
	struct gfs_easet_io req;
	int err = 0;
	struct inode *inode = dentry->d_inode;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;
	struct gfs_sbd *sdp = ip->i_sbd;
	char *truncated_name;

	req.es_type = get_eatype(name, &truncated_name);
	if (req.es_type == GFS_EATYPE_UNUSED)
		err = -EOPNOTSUPP;
	else {
		req.es_name = truncated_name;
		req.es_data = NULL;
		req.es_data_len = 0;
		req.es_name_len = strlen(truncated_name);
		req.es_cmd = GFS_EACMD_REMOVE;
		err = gfs_set_eattr(sdp, ip, &req, gfs_get_user_cred());
	}

	return err;
}

struct inode_operations gfs_file_iops = {
      revalidate:gfs_irevalidate,
      setattr:gfs_setattr,
      setxattr:gfs_setxattr,
      getxattr:gfs_getxattr,
      listxattr:gfs_listxattr,
      removexattr:gfs_removexattr,
      permission:gfs_access_permission,
};

struct address_space_operations gfs_file_aops = {
      writepage:gfs_writepage,
      readpage:gfs_readpage,
      sync_page:block_sync_page,
      prepare_write:gfs_prepare_write,
      commit_write:gfs_commit_write,
      bmap:gfs_bmap,
      direct_IO:gfs_direct_IO,
};

struct inode_operations gfs_dev_iops = {
      revalidate:gfs_irevalidate,
      setattr:gfs_setattr,
      setxattr:gfs_setxattr,
      getxattr:gfs_getxattr,
      listxattr:gfs_listxattr,
      removexattr:gfs_removexattr,
      permission:gfs_access_permission,
};

struct inode_operations gfs_dir_iops = {
      create:gfs_create,
      lookup:gfs_lookup,
      link:gfs_link,
      unlink:gfs_unlink,
      symlink:gfs_symlink,
      mkdir:gfs_mkdir,
      rmdir:gfs_rmdir,
      mknod:gfs_mknod,
      rename:gfs_rename,
      revalidate:gfs_irevalidate,
      setattr:gfs_setattr,
      setxattr:gfs_setxattr,
      getxattr:gfs_getxattr,
      listxattr:gfs_listxattr,
      removexattr:gfs_removexattr,
      permission:gfs_access_permission,
};

struct inode_operations gfs_symlink_iops = {
      readlink:gfs_readlink,
      follow_link:gfs_follow_link,
      revalidate:gfs_irevalidate,
      setattr:gfs_setattr,
      setxattr:gfs_setxattr,
      getxattr:gfs_getxattr,
      listxattr:gfs_listxattr,
      removexattr:gfs_removexattr,
      permission:gfs_access_permission,
};
