/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/init.h>
#include <linux/proc_fs.h>

#include "linux_dcache.h"
#include "dio.h"
#include "glock.h"
#include "linux_mount.h"
#include "linux_super.h"

static DECLARE_FSTYPE_DEV(gfs_fs_type, "gfs", gfs_read_super);

struct proc_dir_entry *gfs_proc_entry = NULL;

volatile int gfs_in_panic = FALSE;
kmem_cache_t *gfs_glock_cachep = NULL;
kmem_cache_t *gfs_inode_cachep = NULL;
kmem_cache_t *gfs_bufdata_cachep = NULL;
kmem_cache_t *gfs_mh_cache_cachep = NULL;
int gfs_glock_in_aspace = FALSE;
char *gfs_mount_args = NULL;
struct semaphore gfs_mount_args_lock;
uint32_t gfs_random_number;

/**
 * init_gfs_fs - Register GFS as a filesystem
 *
 * Returns: 0 on success, error code on failure
 */

int __init
init_gfs_fs(void)
{
	int error = 0;

	init_MUTEX(&gfs_mount_args_lock);

	gfs_proc_entry =
	    create_proc_read_entry("fs/gfs", S_IFREG | 0200, NULL, NULL, NULL);
	if (!gfs_proc_entry) {
		printk("GFS: can't register /proc/fs/gfs\n");
		error = -EINVAL;
		goto fail;
	}
	gfs_proc_entry->write_proc = gfs_proc_write;

	/*  Initialize the random number generator  */

	gfs_random_number = xtime.tv_usec;

	/*  Can we stuff "struct gfs_glock"s in "struct inode"s?  */

	if ((((unsigned long) sizeof (struct inode)) -
	     ((unsigned long) &(((struct inode *) NULL)->u))) >=
	    sizeof (struct gfs_glock))
		gfs_glock_in_aspace = TRUE;

	gfs_glock_cachep =
	    kmem_cache_create("gfs_glock", sizeof (struct gfs_glock), 0, 0,
			      NULL, NULL);
	if (!gfs_glock_cachep)
		goto fail2;

	gfs_inode_cachep =
	    kmem_cache_create("gfs_inode", sizeof (struct gfs_inode), 0, 0,
			      NULL, NULL);
	if (!gfs_inode_cachep)
		goto fail2;

	gfs_bufdata_cachep =
	    kmem_cache_create("gfs_bufdata", sizeof (struct gfs_bufdata), 0, 0,
			      NULL, NULL);
	if (!gfs_bufdata_cachep)
		goto fail2;

	gfs_mh_cache_cachep =
	    kmem_cache_create("gfs_mh_cache", sizeof (struct gfs_mh_cache), 0,
			      0, NULL, NULL);
	if (!gfs_mh_cache_cachep)
		goto fail2;

	error = register_filesystem(&gfs_fs_type);
	if (error)
		goto fail2;

	printk("GFS %s (built %s %s) installed\n",
	       GFS_RELEASE_NAME, __DATE__, __TIME__);

	return 0;

      fail2:
	if (gfs_mh_cache_cachep)
		kmem_cache_destroy(gfs_mh_cache_cachep);

	if (gfs_bufdata_cachep)
		kmem_cache_destroy(gfs_bufdata_cachep);

	if (gfs_inode_cachep)
		kmem_cache_destroy(gfs_inode_cachep);

	if (gfs_glock_cachep)
		kmem_cache_destroy(gfs_glock_cachep);

	down(&gfs_mount_args_lock);
	if (gfs_mount_args) {
		kfree(gfs_mount_args);
		gfs_mount_args = NULL;
	}
	up(&gfs_mount_args_lock);
	remove_proc_entry("fs/gfs", NULL);

      fail:
	return error;
}

/**
 * exit_gfs_fs - Unregister the file system
 *
 */

void __exit
exit_gfs_fs(void)
{
	unregister_filesystem(&gfs_fs_type);

	kmem_cache_destroy(gfs_mh_cache_cachep);
	kmem_cache_destroy(gfs_bufdata_cachep);
	kmem_cache_destroy(gfs_inode_cachep);
	kmem_cache_destroy(gfs_glock_cachep);

	down(&gfs_mount_args_lock);
	if (gfs_mount_args) {
		kfree(gfs_mount_args);
		gfs_mount_args = NULL;
	}
	up(&gfs_mount_args_lock);
	remove_proc_entry("fs/gfs", NULL);
}

MODULE_DESCRIPTION("Global File System " GFS_RELEASE_NAME);
MODULE_AUTHOR("Red Hat, Inc.");
MODULE_LICENSE("GPL");

module_init(init_gfs_fs);
module_exit(exit_gfs_fs);

/*  Symbols  */

EXPORT_NO_SYMBOLS;
