/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"

#include <linux/sched.h>
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#include <linux/blkdev.h>
#include <linux/seq_file.h>

#include "bmap.h"
#include "daemon.h"
#include "linux_dcache.h"
#include "dio.h"
#include "glock.h"
#include "glops.h"
#include "inode.h"
#include "linux_inode.h"
#include "locking.h"
#include "log.h"
#include "linux_mount.h"
#include "linux_page.h"
#include "quota.h"
#include "recovery.h"
#include "rgrp.h"
#include "super.h"
#include "linux_super.h"
#include "trans.h"
#include "unlinked.h"

static int
gfs_scand_bounce(void *data)
{
	gfs_scand((struct gfs_sbd *) data);
	return 0;
}
static int
gfs_glockd_bounce(void *data)
{
	gfs_glockd((struct gfs_sbd *) data);
	return 0;
}
static int
gfs_recoverd_bounce(void *data)
{
	gfs_recoverd((struct gfs_sbd *) data);
	return 0;
}
static int
gfs_logd_bounce(void *data)
{
	gfs_logd((struct gfs_sbd *) data);
	return 0;
}
static int
gfs_quotad_bounce(void *data)
{
	gfs_quotad((struct gfs_sbd *) data);
	return 0;
}
static int
gfs_inoded_bounce(void *data)
{
	gfs_inoded((struct gfs_sbd *) data);
	return 0;
}

/**
 * gfs_read_super - Read in superblock
 * @sb: The VFS superblock
 * @data: Mount options
 * @silent: Don't complain if its not a GFS filesystem
 *
 * Returns: The VFS superblock, or NULL on error
 */

struct super_block *
gfs_read_super(struct super_block *sb, void *data, int silent)
{
	struct gfs_sbd *sdp;
	struct gfs_holder mount_gh, sb_gh, ji_gh;
	struct inode *inode;
	int super = TRUE, jindex = TRUE;
	unsigned int x;
	int error;

	sdp = vmalloc(sizeof (struct gfs_sbd));
	if (!sdp)
		goto fail;

	memset(sdp, 0, sizeof (struct gfs_sbd));

	sb->u.generic_sbp = sdp;
	sdp->sd_vfs = sb;

	gfs_sync_dev(sdp);
	invalidate_buffers(sb->s_dev);

	/*  Init rgrp variables  */

	INIT_LIST_HEAD(&sdp->sd_rglist);
	init_MUTEX(&sdp->sd_rindex_lock);
	INIT_LIST_HEAD(&sdp->sd_rg_mru_list);
	spin_lock_init(&sdp->sd_rg_mru_lock);
	INIT_LIST_HEAD(&sdp->sd_rg_recent);
	spin_lock_init(&sdp->sd_rg_recent_lock);
	spin_lock_init(&sdp->sd_rg_forward_lock);

	for (x = 0; x < GFS_GHASH_SIZE; x++) {
		sdp->sd_ghash[x].hb_lock = RW_LOCK_UNLOCKED;
		INIT_LIST_HEAD(&sdp->sd_ghash[x].hb_list);
	}

	INIT_LIST_HEAD(&sdp->sd_reclaim_list);
	spin_lock_init(&sdp->sd_reclaim_lock);
	init_waitqueue_head(&sdp->sd_reclaim_wchan);

	for (x = 0; x < GFS_MCHASH_SIZE; x++)
		INIT_LIST_HEAD(&sdp->sd_mh_cache[x]);
	spin_lock_init(&sdp->sd_mh_cache_lock);

	init_MUTEX(&sdp->sd_freeze_lock);

	init_MUTEX(&sdp->sd_thread_lock);
	init_completion(&sdp->sd_thread_completion);

	spin_lock_init(&sdp->sd_log_seg_lock);
	INIT_LIST_HEAD(&sdp->sd_log_seg_list);
	init_waitqueue_head(&sdp->sd_log_seg_wait);
	INIT_LIST_HEAD(&sdp->sd_log_ail);
	INIT_LIST_HEAD(&sdp->sd_log_incore);
	init_MUTEX(&sdp->sd_log_lock);
	INIT_LIST_HEAD(&sdp->sd_unlinked_list);
	spin_lock_init(&sdp->sd_unlinked_lock);
	INIT_LIST_HEAD(&sdp->sd_quota_list);
	spin_lock_init(&sdp->sd_quota_lock);

	INIT_LIST_HEAD(&sdp->sd_dirty_j);
	spin_lock_init(&sdp->sd_dirty_j_lock);

	spin_lock_init(&sdp->sd_plock_lock);
	INIT_LIST_HEAD(&sdp->sd_plock_list);

	spin_lock_init(&sdp->sd_arch.sd_ail_lock);
	INIT_LIST_HEAD(&sdp->sd_arch.sd_recovery_bufs);

	gfs_init_tune_data(sdp);

	if (!gfs_make_args((char *) data, &sdp->sd_args)) {
		printk("GFS: can't parse mount arguments\n");
		error = -EINVAL;
		goto fail_vfree;
	}

	/*  Copy out mount flags  */

	if (sb->s_flags & (MS_NOATIME | MS_NODIRATIME))
		set_bit(SDF_NOATIME, &sdp->sd_flags);
	if (sb->s_flags & MS_SYNCHRONOUS)
		set_bit(SDF_SYNC_MOUNT, &sdp->sd_flags);
	if (sb->s_flags & MS_RDONLY) {
		set_bit(SDF_ROFS, &sdp->sd_flags);
		set_bit(SDF_NOATIME, &sdp->sd_flags);
	}

	/*  Hack to support sendfile  */

	sb->s_flags |= MS_SENDFILE_FOP;

	/*  Setup up Virtual Super Block  */

	sb->s_magic = GFS_MAGIC;
	sb->s_op = &gfs_sops;
	sb->s_flags |= MS_NOATIME | MS_NODIRATIME;
	sb->s_maxbytes = MAX_LFS_FILESIZE;

	if (sdp->sd_args.ar_posixacls)
		sb->s_flags |= MS_POSIXACL;

	/*  Set up the buffer cache and fill in some fake values
	   to allow us to read in the superblock.  */

	sdp->sd_device_bsize = get_hardsect_size(sb->s_dev);

	sdp->sd_sb.sb_bsize = sdp->sd_device_bsize;
	if (sdp->sd_sb.sb_bsize < GFS_BASIC_BLOCK)
		sdp->sd_sb.sb_bsize = GFS_BASIC_BLOCK;
	sdp->sd_sb.sb_bsize_shift = ffs(sdp->sd_sb.sb_bsize) - 1;
	sdp->sd_fsb2bb_shift =
	    sdp->sd_sb.sb_bsize_shift - GFS_BASIC_BLOCK_SHIFT;
	sdp->sd_fsb2bb = 1 << sdp->sd_fsb2bb_shift;

	GFS_ASSERT_SBD(sizeof (struct gfs_sb) <= sdp->sd_sb.sb_bsize, sdp,);

	set_blocksize(sb->s_dev, sdp->sd_sb.sb_bsize);
	sb->s_blocksize = sdp->sd_sb.sb_bsize;
	sb->s_blocksize_bits = sdp->sd_sb.sb_bsize_shift;

	error = gfs_mount_lockproto(sdp, silent);
	if (error)
		goto fail_vfree;

	printk("GFS: fsid=%s: Joined cluster. Now mounting FS...\n",
	       sdp->sd_fsname);

	if ((sdp->sd_lockstruct.ls_flags & LM_LSFLAG_LOCAL)
	    && !sdp->sd_args.ar_ignore_local_fs) {

		/*  Force local [p|f]locks  */
		sdp->sd_args.ar_localflocks = TRUE;

		/*  Force local read ahead and caching  */
		sdp->sd_args.ar_localcaching = TRUE;
	}

	/*  Start up the scand thread  */

	error = kernel_thread(gfs_scand_bounce, sdp, 0);
	if (error < 0) {
		printk("GFS: fsid=%s: can't start scand thread: %d\n",
		       sdp->sd_fsname, error);
		goto fail_lockproto;
	}
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Start up the glockd thread  */

	for (sdp->sd_glockd_num = 0;
	     sdp->sd_glockd_num < sdp->sd_args.ar_num_glockd;
	     sdp->sd_glockd_num++) {
		error = kernel_thread(gfs_glockd_bounce, sdp, 0);
		if (error < 0) {
			printk
			    ("GFS: fsid=%s: can't start glockd thread: %d\n",
			     sdp->sd_fsname, error);
			goto fail_glockd;
		}
		wait_for_completion(&sdp->sd_thread_completion);
	}

	error = gfs_glock_nq_num(sdp,
				 GFS_MOUNT_LOCK, &gfs_nondisk_glops,
				 LM_ST_EXCLUSIVE, LM_FLAG_NOEXP | GL_NOCACHE,
				 &mount_gh);
	if (error) {
		printk("GFS: fsid=%s: can't acquire mount glock: %d\n",
		       sdp->sd_fsname, error);
		goto fail_glockd;
	}

	error = gfs_glock_nq_num(sdp,
				 GFS_LIVE_LOCK, &gfs_nondisk_glops,
				 LM_ST_SHARED, LM_FLAG_NOEXP | GL_EXACT,
				 &sdp->sd_live_gh);
	if (error) {
		printk("GFS: fsid=%s: can't acquire live glock: %d\n",
		       sdp->sd_fsname, error);
		goto fail_gunlock_mount;
	}

	sdp->sd_live_gh.gh_owner = NULL;

	error = gfs_glock_nq_num(sdp,
				 GFS_SB_LOCK, &gfs_meta_glops,
				 (sdp->sd_args.
				  ar_upgrade) ? LM_ST_EXCLUSIVE : LM_ST_SHARED,
				 0, &sb_gh);
	if (error) {
		printk("GFS: fsid=%s: can't acquire superblock glock: %d\n",
		       sdp->sd_fsname, error);
		goto fail_gunlock_live;
	}

	error = gfs_read_sb(sdp, sb_gh.gh_gl, silent);
	if (error) {
		printk("GFS: fsid=%s: can't read superblock: %d\n",
		       sdp->sd_fsname, error);
		goto fail_gunlock_sb;
	}

	/*  Get rid of buffers from the original block size  */
	sb_gh.gh_gl->gl_ops->go_inval(sb_gh.gh_gl, DIO_METADATA | DIO_DATA);

	/*  Set up the buffer cache and SB for real  */

	if (sdp->sd_sb.sb_bsize < sdp->sd_device_bsize) {
		printk
		    ("GFS: fsid=%s: FS block size (%u) is too small for device block size (%u)\n",
		     sdp->sd_fsname, sdp->sd_sb.sb_bsize, sdp->sd_device_bsize);
		error = -EINVAL;
		goto fail_gunlock_sb;
	}
	if (sdp->sd_sb.sb_bsize > PAGE_SIZE) {
		printk
		    ("GFS: fsid=%s: FS block size (%u) is too big for machine page size (%u)\n",
		     sdp->sd_fsname, sdp->sd_sb.sb_bsize,
		     (unsigned int) PAGE_SIZE);
		error = -EINVAL;
		goto fail_gunlock_sb;
	}

	set_blocksize(sb->s_dev, sdp->sd_sb.sb_bsize);
	sb->s_blocksize = sdp->sd_sb.sb_bsize;
	sb->s_blocksize_bits = sdp->sd_sb.sb_bsize_shift;

	sb_gh.gh_gl->gl_arch.gl_aspace->i_blkbits = sdp->sd_sb.sb_bsize_shift;

	/*  Read in journal index inode  */

	error = gfs_get_jiinode(sdp);
	if (error) {
		printk("GFS: fsid=%s: can't get journal index inode: %d\n",
		       sdp->sd_fsname, error);
		goto fail_gunlock_sb;
	}

	init_MUTEX(&sdp->sd_jindex_lock);

	/*  Get a handle on the transaction glock  */

	error =
	    gfs_get_glstruct(sdp, GFS_TRANS_LOCK, &gfs_trans_glops, ST_CREATE,
			     &sdp->sd_trans_gl);
	if (error)
		goto fail_ji_free;
	set_bit(GLF_STICKY, &sdp->sd_trans_gl->gl_flags);

	/*  Upgrade version numbers if we need to  */

	if (sdp->sd_args.ar_upgrade) {
		error = gfs_do_upgrade(sdp, sb_gh.gh_gl);
		if (error)
			goto fail_trans_gl;
	}

	/*  Load in the journal index  */

	error = gfs_jindex_hold(sdp, &ji_gh);
	if (error) {
		printk("GFS: fsid=%s: can't read journal index: %d\n",
		       sdp->sd_fsname, error);
		goto fail_trans_gl;
	}

	if (sdp->sd_lockstruct.ls_jid >= sdp->sd_journals) {
		printk("GFS: fsid=%s: can't mount journal #%u\n",
		       sdp->sd_fsname, sdp->sd_lockstruct.ls_jid);
		printk
		    ("GFS: fsid=%s: there are only %u journals (0 - %u)\n",
		     sdp->sd_fsname, sdp->sd_journals, sdp->sd_journals - 1);
		error = -EINVAL;
		goto fail_gunlock_ji;
	}
	sdp->sd_jdesc = sdp->sd_jindex[sdp->sd_lockstruct.ls_jid];
	sdp->sd_log_seg_free = sdp->sd_jdesc.ji_nsegment - 1;

	error = gfs_glock_nq_num(sdp,
				 sdp->sd_jdesc.ji_addr, &gfs_meta_glops,
				 LM_ST_EXCLUSIVE, LM_FLAG_NOEXP,
				 &sdp->sd_journal_gh);
	if (error) {
		printk
		    ("GFS: fsid=%s: can't acquire the journal glock: %d\n",
		     sdp->sd_fsname, error);
		goto fail_gunlock_ji;
	}

	if (sdp->sd_lockstruct.ls_first) {
		for (x = 0; x < sdp->sd_journals; x++) {
			error = gfs_recover_journal(sdp, x, TRUE);
			if (error) {
				printk
				    ("GFS: fsid=%s: error recovering journal %u: %d\n",
				     sdp->sd_fsname, x, error);
				goto fail_gunlock_journal;
			}
		}

		sdp->sd_lockstruct.ls_ops->lm_others_may_mount(sdp->
							       sd_lockstruct.
							       ls_lockspace);
		sdp->sd_lockstruct.ls_first = FALSE;
	} else {
		error =
		    gfs_recover_journal(sdp, sdp->sd_lockstruct.ls_jid, TRUE);
		if (error) {
			printk
			    ("GFS: fsid=%s: error recovering my journal: %d\n",
			     sdp->sd_fsname, error);
			goto fail_gunlock_journal;
		}
	}

	gfs_glock_dq_uninit(&ji_gh);
	jindex = FALSE;

	/*  Disown my Journal glock  */

	sdp->sd_journal_gh.gh_owner = NULL;

	/*  Drop our cache and reread all the things we read before the replay.  */

	error = gfs_read_sb(sdp, sb_gh.gh_gl, FALSE);
	if (error) {
		printk("GFS: fsid=%s: can't read superblock: %d\n",
		       sdp->sd_fsname, error);
		goto fail_gunlock_journal;
	}

	gfs_glock_force_drop(sdp->sd_jiinode->i_gl);

	error = gfs_jindex_hold(sdp, &ji_gh);
	if (error) {
		printk("GFS: fsid=%s: can't read journal index: %d\n",
		       sdp->sd_fsname, error);
		goto fail_gunlock_journal;
	}
	gfs_glock_dq_uninit(&ji_gh);

	/*  Make the FS read/write  */

	if (!test_bit(SDF_ROFS, &sdp->sd_flags)) {
		error = gfs_make_fs_rw(sdp);
		if (error) {
			printk("GFS: fsid=%s: can't make FS RW: %d\n",
			       sdp->sd_fsname, error);
			goto fail_gunlock_journal;
		}
	}

	/*  Start up the recover thread  */

	error = kernel_thread(gfs_recoverd_bounce, sdp, 0);
	if (error < 0) {
		printk("GFS: fsid=%s: can't start recoverd thread: %d\n",
		       sdp->sd_fsname, error);
		goto fail_recover_dump;
	}
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Read in the resource index inode  */

	error = gfs_get_riinode(sdp);
	if (error) {
		printk("GFS: fsid=%s: can't get resource index inode: %d\n",
		       sdp->sd_fsname, error);
		goto fail_recoverd;
	}

	/*  Get the root inode  */

	error = gfs_get_rootinode(sdp);
	if (error) {
		printk("GFS: fsid=%s: can't read in root inode: %d\n",
		       sdp->sd_fsname, error);
		goto fail_ri_free;
	}

	/*  Read in the quota inode  */

	error = gfs_get_qinode(sdp);
	if (error) {
		printk("GFS: fsid=%s: can't get quota file inode: %d\n",
		       sdp->sd_fsname, error);
		goto fail_root_free;
	}

	/*  Read in the license inode  */

	error = gfs_get_linode(sdp);
	if (error) {
		printk("GFS: fsid=%s: can't get license file inode: %d\n",
		       sdp->sd_fsname, error);
		goto fail_qi_free;
	}

	/*  We're through with the superblock lock  */

	gfs_glock_dq_uninit(&sb_gh);
	super = FALSE;

	/*  Get the inode/dentry  */

	inode = gfs_iget(sdp->sd_rooti, ST_CREATE);
	sb->s_root = d_alloc_root(inode);
	if (!sb->s_root) {
		iput(inode);
		printk("GFS: fsid=%s: can't get root dentry\n", sdp->sd_fsname);
		error = -ENOMEM;
		goto fail_li_free;
	}

	/*  Start up the logd thread  */

	sdp->sd_jindex_refresh_time = jiffies;

	error = kernel_thread(gfs_logd_bounce, sdp, 0);
	if (error < 0) {
		printk("GFS: fsid=%s: can't start logd thread: %d\n",
		       sdp->sd_fsname, error);
		goto fail_dput;
	}
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Start up the quotad thread  */

	error = kernel_thread(gfs_quotad_bounce, sdp, 0);
	if (error < 0) {
		printk("GFS: fsid=%s: can't start quotad thread: %d\n",
		       sdp->sd_fsname, error);
		goto fail_logd;
	}
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Start up the inoded thread  */

	error = kernel_thread(gfs_inoded_bounce, sdp, 0);
	if (error < 0) {
		printk("GFS: fsid=%s: can't start inoded thread: %d\n",
		       sdp->sd_fsname, error);
		goto fail_quotad;
	}
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Get a handle on the rename lock  */

	error =
	    gfs_get_glstruct(sdp, GFS_RENAME_LOCK, &gfs_nondisk_glops,
			     ST_CREATE, &sdp->sd_rename_gl);
	if (error)
		goto fail_inoded;

	gfs_glock_dq_uninit(&mount_gh);

	return sb;

      fail_inoded:
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_INODED_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_inoded_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

      fail_quotad:
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_QUOTAD_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_quotad_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

      fail_logd:
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_LOGD_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_logd_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

      fail_dput:
	dput(sb->s_root);

      fail_li_free:
	gfs_put_istruct(sdp->sd_linode);

      fail_qi_free:
	gfs_put_istruct(sdp->sd_qinode);

      fail_root_free:
	gfs_put_istruct(sdp->sd_rooti);

      fail_ri_free:
	gfs_put_istruct(sdp->sd_riinode);
	gfs_clear_rgrpd(sdp);

      fail_recoverd:
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_RECOVERD_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_recoverd_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

      fail_recover_dump:
	clear_bit(SDF_JOURNAL_LIVE, &sdp->sd_flags);
	gfs_unlinked_cleanup(sdp);
	gfs_quota_cleanup(sdp);

      fail_gunlock_journal:
	gfs_glock_dq_uninit(&sdp->sd_journal_gh);

      fail_gunlock_ji:
	if (jindex)
		gfs_glock_dq_uninit(&ji_gh);

      fail_trans_gl:
	gfs_put_glstruct(sdp->sd_trans_gl);

      fail_ji_free:
	gfs_put_istruct(sdp->sd_jiinode);
	gfs_clear_journals(sdp);

      fail_gunlock_sb:
	if (super)
		gfs_glock_dq_uninit(&sb_gh);

      fail_gunlock_live:
	gfs_glock_dq_uninit(&sdp->sd_live_gh);

      fail_gunlock_mount:
	gfs_glock_dq_uninit(&mount_gh);

      fail_glockd:
	clear_bit(SDF_GLOCKD_RUN, &sdp->sd_flags);
	wake_up(&sdp->sd_reclaim_wchan);
	while (sdp->sd_glockd_num--)
		wait_for_completion(&sdp->sd_thread_completion);

	down(&sdp->sd_thread_lock);
	clear_bit(SDF_SCAND_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_scand_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

      fail_lockproto:
	gfs_clear_ghash(sdp, TRUE);
	gfs_unmount_lockproto(sdp);
	gfs_clear_dirty_j(sdp);
	while (gfs_invalidate_inodes(sdp))
		yield();

      fail_vfree:
	vfree(sdp);

      fail:
	sb->u.generic_sbp = NULL;
	sb->s_dev = NODEV;

	return NULL;
}

/**
 * gfs_get_inode - New "get an inode by number" function
 * @sdp: the filesystem
 * @inum: The inode number to get.
 * @inode: the new inode
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_get_inode(struct gfs_sbd *sdp, struct gfs_inum *inum, struct inode **inode)
{
	struct gfs_holder i_gh, ri_gh, rgd_gh;
	struct gfs_rgrpd *rgd;
	struct gfs_dinode *di;
	struct gfs_inode *ip;
	struct buffer_head *bh;
	int error;

	error = gfs_glock_nq_num(sdp,
				 inum->no_formal_ino, &gfs_inode_glops,
				 LM_ST_SHARED, LM_FLAG_ANY | GL_LOCAL_EXCL,
				 &i_gh);
	if (error)
		goto fail;

	error = gfs_get_istruct(i_gh.gh_gl, inum, NO_ST_CREATE, &ip);
	if (error)
		goto fail_gunlock;
	if (ip)
		goto out;

	error =
	    gfs_dread(sdp, inum->no_addr, i_gh.gh_gl, DIO_START | DIO_WAIT, &bh);
	if (error)
		goto fail_gunlock;

	di = (struct gfs_dinode *) (bh)->b_data;

	if (gfs32_to_cpu(di->di_header.mh_magic) != GFS_MAGIC ||
	    gfs32_to_cpu(di->di_header.mh_type) != GFS_METATYPE_DI ||
	    (gfs32_to_cpu(di->di_flags) & GFS_DIF_UNUSED)) {
		gfs_drelse(bh);
		error = -EINVAL;
		goto fail_gunlock;
	}

	gfs_drelse(bh);

	error = gfs_rindex_hold(sdp, &ri_gh);
	if (error)
		goto fail_gunlock;

	rgd = gfs_blk2rgrpd(sdp, inum->no_addr);
	if (!rgd) {
		error = -EINVAL;
		goto fail_rindex;
	}

	error = gfs_glock_nq_init(rgd->rd_gl, LM_ST_SHARED, 0, &rgd_gh);
	if (error)
		goto fail_rindex;

	if (gfs_get_block_type(rgd, inum->no_addr) != GFS_BLKST_USEDMETA) {
		error = -EINVAL;
		goto fail_gunlock_rg;
	}

	gfs_glock_dq_uninit(&rgd_gh);
	gfs_glock_dq_uninit(&ri_gh);

	error = gfs_get_istruct(i_gh.gh_gl, inum, ST_CREATE, &ip);
	if (error)
		goto fail_gunlock;

	atomic_inc(&sdp->sd_fh2dentry_misses);

      out:
	*inode = gfs_iget(ip, ST_CREATE);
	gfs_put_istruct(ip);

	gfs_glock_dq_uninit(&i_gh);

	return 0;

      fail_gunlock_rg:
	gfs_glock_dq_uninit(&rgd_gh);

      fail_rindex:
	gfs_glock_dq_uninit(&ri_gh);

      fail_gunlock:
	i_gh.gh_flags |= GL_NOCACHE;
	gfs_glock_dq_uninit(&i_gh);

      fail:
	return error;
}

/**
 * gfs_fh_to_dentry - Turn an NFS filehandle into a GFS dentry
 * @sb: The super block
 * @fh: The filehandle
 * @len: The length of the filehandle
 * @fhtype: The type of the filehandle
 * @parent: True is the dentry to be returned is for the parent of the object
 *
 * If no dentry can be found, then this routine must create a new "root"
 * dentry and return that with DCACHE_NFSD_DISCONNECTED set in the flags.
 *
 * Returns: dentry of object or dentry of parent.
 */

static struct dentry *
gfs_fh_to_dentry(struct super_block *sb, __u32 * fh, int len, int fhtype,
		 int parent)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;
	struct gfs_inum inum;
	struct list_head *tmp, *head;
	struct inode *inode;
	struct dentry *dentry = NULL;
	__u32 generation = 0;
	int gen_ok = FALSE;
	int error;

	switch (fhtype) {
	case 3:
	case 5:
	case 6:
		break;
	default:
		error = -EINVAL;
		goto fail;
	}

	if (!parent) {
		inum.no_formal_ino = (uint64_t) gfs32_to_cpu(fh[0]) << 32;
		inum.no_formal_ino |= (uint64_t) gfs32_to_cpu(fh[1]);
		generation = gfs32_to_cpu(fh[2]);
		gen_ok = TRUE;
	} else {
		inum.no_formal_ino = (uint64_t) gfs32_to_cpu(fh[3]) << 32;
		inum.no_formal_ino |= (uint64_t) gfs32_to_cpu(fh[4]);
		if (fhtype == 6) {
			generation = gfs32_to_cpu(fh[5]);
			gen_ok = TRUE;
		}
	}

	error = -EINVAL;
	if (inum.no_formal_ino == sdp->sd_jiinode->i_num.no_formal_ino ||
	    inum.no_formal_ino == sdp->sd_riinode->i_num.no_formal_ino ||
	    inum.no_formal_ino == sdp->sd_qinode->i_num.no_formal_ino ||
	    inum.no_formal_ino == sdp->sd_linode->i_num.no_formal_ino)
		goto fail;

	inum.no_addr = inum.no_formal_ino;

	error = gfs_get_inode(sdp, &inum, &inode);
	if (error)
		goto fail;

	if (gen_ok && generation != inode->i_generation) {
		iput(inode);
		error = -ESTALE;
		goto fail;
	}

	spin_lock(&dcache_lock);
	for (head = &inode->i_dentry, tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		dentry = list_entry(tmp, struct dentry, d_alias);
		if (!(dentry->d_flags & DCACHE_NFSD_DISCONNECTED)) {
			dget_locked(dentry);
			dentry->d_vfs_flags |= DCACHE_REFERENCED;
			spin_unlock(&dcache_lock);
			iput(inode);
			goto out;
		}
	}
	spin_unlock(&dcache_lock);

	dentry = d_alloc_root(inode);
	if (!dentry) {
		iput(inode);
		error = -ENOMEM;
		goto fail;
	}
	dentry->d_flags |= DCACHE_NFSD_DISCONNECTED;

      out:

	return dentry;

      fail:
	return ERR_PTR(error);
}

/**
 * gfs_dentry_to_fh - Turn a GFS dentry into an NFS filehandle
 * @dentry: dentry to build a filehandle for
 * @fh: Point to the filehandle
 * @lenp: Available length (passed in), used length (when filehandle is built)
 * @need_parent: Filehandle must include sufficient info to find parent
 *
 * Type 0: Reserved, don't use
 * Type 1,2: Use in the same way as the generic NFS handle code
 * Type 255: Used to indicate an error has occured
 *
 * Returns: Type of filehandle
 */

static int
gfs_dentry_to_fh(struct dentry *dentry, __u32 * fh, int *lenp, int need_parent)
{
	struct inode *inode;
	struct gfs_inode *ip;
	int maxlen = *lenp;
	int type = 255;

	if (maxlen < 3)
		goto out;

	inode = dentry->d_inode;
	ip = (struct gfs_inode *) inode->u.generic_ip;

	fh[0] = cpu_to_gfs32((uint32_t) (ip->i_num.no_formal_ino >> 32));
	fh[1] = cpu_to_gfs32((uint32_t) (ip->i_num.no_formal_ino & 0xFFFFFFFF));
	fh[2] = cpu_to_gfs32(inode->i_generation);
	type = *lenp = 3;

	if (maxlen < 5 || !need_parent)
		goto out;

	inode = dentry->d_parent->d_inode;
	ip = (struct gfs_inode *) inode->u.generic_ip;

	fh[3] = cpu_to_gfs32((uint32_t) (ip->i_num.no_formal_ino >> 32));
	fh[4] = cpu_to_gfs32((uint32_t) (ip->i_num.no_formal_ino & 0xFFFFFFFF));
	type = *lenp = 5;

	if (maxlen < 6)
		goto out;

	fh[5] = cpu_to_gfs32(inode->i_generation);
	type = *lenp = 6;

      out:

	return type;
}

/**
 * gfs_write_inode - Make sure the inode is stable on the disk
 * @inode: The inode
 * @sync: synchronous write flag
 *
 */

static void
gfs_write_inode(struct inode *inode, int sync)
{
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;

	if (ip && sync && !gfs_in_panic)
		gfs_log_flush_glock(ip->i_gl);

}

/**
 * gfs_put_inode - put an inode
 * @inode: The inode
 *
 * If i_nlink is zero, any dirty data for the inode is thrown away.
 * If a process on another machine has the file open, it may need that
 * data.  So, sync it out.
 */

static void
gfs_put_inode(struct inode *inode)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) inode->i_sb->u.generic_sbp;
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;

	if (ip &&
	    !inode->i_nlink && S_ISREG(inode->i_mode) &&
	    !sdp->sd_args.ar_localcaching)
		gfs_sync_page_i(inode, 0);

}

/**
 * gfs_statfs - Gather and return stats about the filesystem
 * @sb: The superblock
 * @statfsbuf: The buffer
 *
 * Returns: 0 on success or error code
 */

static int
gfs_statfs(struct super_block *sb, struct statfs *sfs)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;
	struct gfs_usage usage;
	int error;

	error = gfs_stat_gfs(sdp, &usage, TRUE);
	if (error)
		return error;

	memset(sfs, 0, sizeof (struct statfs));

	sfs->f_type = GFS_MAGIC;
	sfs->f_bsize = usage.gu_block_size;
	sfs->f_blocks = usage.gu_total_blocks;
	sfs->f_bfree =
	    usage.gu_free + usage.gu_free_dinode + usage.gu_free_meta;
	sfs->f_bavail =
	    usage.gu_free + usage.gu_free_dinode + usage.gu_free_meta;
	sfs->f_files =
	    usage.gu_used_dinode + usage.gu_free_dinode + usage.gu_free_meta +
	    usage.gu_free;
	sfs->f_ffree =
	    usage.gu_free_dinode + usage.gu_free_meta + usage.gu_free;
	sfs->f_namelen = GFS_FNAMESIZE;

	return 0;
}

/**
 * gfs_clear_inode - Deallocate an inode when VFS is done with it
 * @inode: The VFS inode
 *
 */

static void
gfs_clear_inode(struct inode *inode)
{
	struct gfs_inode *ip = (struct gfs_inode *) inode->u.generic_ip;

	if (ip) {
		spin_lock(&ip->i_arch.i_lock);
		ip->i_vnode = NULL;
		inode->u.generic_ip = NULL;
		spin_unlock(&ip->i_arch.i_lock);

		gfs_glock_schedule_for_reclaim(ip->i_gl);
		gfs_put_istruct(ip);
	}

}

/**
 * gfs_remount_fs - called when the FS is remounted
 * @sb:  the filesystem
 * @flags:  the remount flags
 * @data:  extra data passed in (not used right now)
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
gfs_remount_fs(struct super_block *sb, int *flags, char *data)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;
	int error = 0;

	if (*flags & (MS_NOATIME | MS_NODIRATIME))
		set_bit(SDF_NOATIME, &sdp->sd_flags);
	else
		clear_bit(SDF_NOATIME, &sdp->sd_flags);

	if (*flags & MS_SYNCHRONOUS)
		set_bit(SDF_SYNC_MOUNT, &sdp->sd_flags);
	else
		clear_bit(SDF_SYNC_MOUNT, &sdp->sd_flags);

	if (*flags & MS_RDONLY) {
		if (!test_bit(SDF_ROFS, &sdp->sd_flags)) {
			error = gfs_make_fs_ro(sdp);
			if (!error)
				set_bit(SDF_NOATIME, &sdp->sd_flags);
		} else
			set_bit(SDF_NOATIME, &sdp->sd_flags);
	} else if (!(*flags & MS_RDONLY)
		   && test_bit(SDF_ROFS, &sdp->sd_flags)) {
		error = gfs_make_fs_rw(sdp);
	}

	/*  Don't let the VFS update atimes.  */
	*flags |= MS_NOATIME | MS_NODIRATIME;

	return error;
}

/**
 * gfs_write_super - disk commit all incore transactions
 * @sb: the filesystem
 *
 * This function is called every time sync(2) is called.
 * After this exits, all dirty buffers and synced.
 */

static void
gfs_write_super(struct super_block *sb)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;

	if (!gfs_in_panic)
		gfs_log_flush(sdp);

}

/**
 * gfs_put_super - Unmount the filesystem
 * @sb: The VFS superblock
 *
 */

static void
gfs_put_super(struct super_block *sb)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;
	int error;

	GFS_ASSERT_SBD(list_empty(&sdp->sd_plock_list), sdp,);

	/*  Unfreeze the filesystem, if we need to  */

	down(&sdp->sd_freeze_lock);
	if (sdp->sd_freeze_count)
		gfs_glock_dq_uninit(&sdp->sd_freeze_gh);
	up(&sdp->sd_freeze_lock);

	/*  Kill off the inode thread  */
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_INODED_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_inoded_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Kill off the quota thread  */
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_QUOTAD_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_quotad_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Kill off the log thread  */
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_LOGD_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_logd_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Kill off the recoverd thread  */
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_RECOVERD_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_recoverd_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

	/*  Kill off the glockd threads  */
	clear_bit(SDF_GLOCKD_RUN, &sdp->sd_flags);
	wake_up(&sdp->sd_reclaim_wchan);
	while (sdp->sd_glockd_num--)
		wait_for_completion(&sdp->sd_thread_completion);

	/*  Kill off the scand thread  */
	down(&sdp->sd_thread_lock);
	clear_bit(SDF_SCAND_RUN, &sdp->sd_flags);
	wake_up_process(sdp->sd_scand_process);
	up(&sdp->sd_thread_lock);
	wait_for_completion(&sdp->sd_thread_completion);

	if (!test_bit(SDF_ROFS, &sdp->sd_flags)) {
		gfs_log_flush(sdp);
		gfs_quota_sync(sdp);
		gfs_quota_sync(sdp);

		error = gfs_make_fs_ro(sdp);
		GFS_ASSERT_SBD(!error, sdp,);
	}

	/*  At this point, we're through modifying the disk  */

	/*  Release stuff  */

	gfs_put_istruct(sdp->sd_riinode);
	gfs_put_istruct(sdp->sd_jiinode);
	gfs_put_istruct(sdp->sd_rooti);
	gfs_put_istruct(sdp->sd_qinode);
	gfs_put_istruct(sdp->sd_linode);

	gfs_put_glstruct(sdp->sd_trans_gl);
	gfs_put_glstruct(sdp->sd_rename_gl);

	gfs_glock_dq_uninit(&sdp->sd_journal_gh);

	gfs_glock_dq_uninit(&sdp->sd_live_gh);

	/*  Get rid of rgrp bitmap structures  */
	gfs_clear_rgrpd(sdp);
	gfs_clear_journals(sdp);

	/*  Take apart glock structures and buffer lists  */
	gfs_clear_ghash(sdp, TRUE);

	/*  Unmount the locking protocol  */
	gfs_unmount_lockproto(sdp);

	/*  At this point, we're through participating in the lockspace  */

	gfs_clear_dirty_j(sdp);

	/*  Get rid of any extra inodes  */
	while (gfs_invalidate_inodes(sdp))
		yield();

	vfree(sdp);

	sb->u.generic_sbp = NULL;
}

/**
 * gfs_show_options - Show mount options for /proc/mounts
 * @s: seq_file structure
 * @mnt: vfsmount
 *
 * Returns: 0 on success or error code
 */

static int
gfs_show_options(struct seq_file *s, struct vfsmount *mnt)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) mnt->mnt_sb->u.generic_sbp;
	struct gfs_args *args = &sdp->sd_args;

	if (args->ar_lockproto[0]) {
		seq_printf(s, ",lockproto=");
		seq_puts(s, args->ar_lockproto);
	}

	if (args->ar_locktable[0]) {
		seq_printf(s, ",locktable=");
		seq_puts(s, args->ar_locktable);
	}

	if (args->ar_hostdata[0]) {
		seq_printf(s, ",hostdata=");
		seq_puts(s, args->ar_hostdata);
	}

	if (args->ar_ignore_local_fs)
		seq_printf(s, ",ignore_local_fs");
	if (args->ar_localflocks)
		seq_printf(s, ",localflocks");
	if (args->ar_localcaching)
		seq_printf(s, ",localcaching");
	if (args->ar_upgrade)
		seq_printf(s, ",upgrade");
	if (args->ar_num_glockd != GFS_GLOCKD_DEFAULT)
		seq_printf(s, ",num_glockd=%u", args->ar_num_glockd);
	if (args->ar_posixacls)
		seq_printf(s, ",acl");

	return 0;
}

/**
 * gfs_write_super_lockfs - prevent further writes to the filesystem
 * @sb: the VFS structure for the filesystem
 *
 */

static void
gfs_write_super_lockfs(struct super_block *sb)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;
	int error;

	for (;;) {
		error = gfs_freeze_fs(sdp);
		if (!error)
			break;

		switch (error) {
		case -EBUSY:
			printk
			    ("GFS: fsid=%s: waiting for recovery before freeze\n",
			     sdp->sd_fsname);
			break;

		default:
			printk("GFS: fsid=%s: error freezing FS: %d\n",
			       sdp->sd_fsname, error);
			break;
		}

		printk("GFS: fsid=%s: retrying...\n", sdp->sd_fsname);

		current->state = TASK_UNINTERRUPTIBLE;
		schedule_timeout(HZ);
	}
}

/**
 * gfs_unlockfs - reallow writes to the filesystem
 * @sb: the VFS structure for the filesystem
 *
 */

static void
gfs_unlockfs(struct super_block *sb)
{
	struct gfs_sbd *sdp = (struct gfs_sbd *) sb->u.generic_sbp;
	gfs_unfreeze_fs(sdp);
}

struct super_operations gfs_sops = {
      write_inode:gfs_write_inode,
      put_inode:gfs_put_inode,
      put_super:gfs_put_super,
      statfs:gfs_statfs,
      clear_inode:gfs_clear_inode,
      remount_fs:gfs_remount_fs,
      write_super:gfs_write_super,
      dentry_to_fh:gfs_dentry_to_fh,
      fh_to_dentry:gfs_fh_to_dentry,
      show_options:gfs_show_options,
      write_super_lockfs:gfs_write_super_lockfs,
      unlockfs:gfs_unlockfs,
};
