/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bits.h"
#include "dio.h"
#include "glock.h"
#include "glops.h"
#include "file.h"
#include "log.h"
#include "quota.h"
#include "rgrp.h"
#include "super.h"
#include "trans.h"

/**
 * mc_hash: find the mh_cache hash bucket for a buffer
 * @bh: the buffer
 *
 * Returns: The bucket number
 */

static unsigned int
mc_hash(struct buffer_head *bh)
{
	uint64_t blkno;
	unsigned int h;

	blkno = (uint64_t) ((bh)->b_blocknr);
	h = gfs_hash(&blkno, sizeof (uint64_t)) & GFS_MCHASH_MASK;

	return h;
}

/**
 * gfs_add_to_mh_cache - add buffers to the cache of metadata
 * @rgd: a RG
 * @bh: an array of buffers
 * @num: the number of buffers in the array
 *
 */

void
gfs_add_to_mh_cache(struct gfs_rgrpd *rgd, struct buffer_head **bh,
		    unsigned int num)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_mh_cache *mc;
	unsigned int x, bucket;
	uint64_t gen;

	for (x = 0; x < num; x++) {
		gfs_meta_check(sdp, bh[x]);

		mc = gmalloc_mh_cache();

		mc->mc_block = (uint64_t) ((bh[x])->b_blocknr);
		memcpy(&mc->mc_mh, (bh[x])->b_data,
		       sizeof (struct gfs_meta_header));

		gen = gfs64_to_cpu(mc->mc_mh.mh_generation) + 2;
		mc->mc_mh.mh_generation = cpu_to_gfs64(gen);

		bucket = mc_hash(bh[x]);

		spin_lock(&sdp->sd_mh_cache_lock);
		list_add(&mc->mc_list_hash, &sdp->sd_mh_cache[bucket]);
		list_add(&mc->mc_list_rgd, &rgd->rd_mh_cache);
		spin_unlock(&sdp->sd_mh_cache_lock);

		atomic_inc(&sdp->sd_mh_cache_count);
	}

	return;
}

/**
 * gfs_fish_from_mh_cache - Try to fill in a buffer with data from the cache
 * @sdp: the filesystem
 * @bh: the buffer to fill in
 *
 * Returns: TRUE if the buffer was cached, FALSE otherwise
 */

int
gfs_fish_from_mh_cache(struct gfs_sbd *sdp, struct buffer_head *bh)
{
	struct list_head *tmp, *head;
	struct gfs_mh_cache *mc;
	unsigned int bucket = mc_hash(bh);

	spin_lock(&sdp->sd_mh_cache_lock);

	for (head = &sdp->sd_mh_cache[bucket], tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		mc = list_entry(tmp, struct gfs_mh_cache, mc_list_hash);
		if (mc->mc_block != (uint64_t) ((bh)->b_blocknr))
			continue;

		list_del(&mc->mc_list_hash);
		list_del(&mc->mc_list_rgd);

		spin_unlock(&sdp->sd_mh_cache_lock);

		atomic_dec(&sdp->sd_mh_cache_count);

		gfs_prep_new_buffer(bh);
		memcpy((bh)->b_data, &mc->mc_mh,
		       sizeof (struct gfs_meta_header));

		gfree_mh_cache(mc);

		return TRUE;
	}

	spin_unlock(&sdp->sd_mh_cache_lock);

	return FALSE;
}

/**
 * gfs_zap_mh_cache - Get rid of the data in the cache of metadata headers
 * @rgd: a RG
 *
 */

void
gfs_zap_mh_cache(struct gfs_rgrpd *rgd)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_mh_cache *mc;

	spin_lock(&sdp->sd_mh_cache_lock);

	while (!list_empty(&rgd->rd_mh_cache)) {
		mc = list_entry(rgd->rd_mh_cache.next, struct gfs_mh_cache,
				mc_list_rgd);
		list_del(&mc->mc_list_hash);
		list_del(&mc->mc_list_rgd);

		spin_unlock(&sdp->sd_mh_cache_lock);
		atomic_dec(&sdp->sd_mh_cache_count);
		gfree_mh_cache(mc);
		spin_lock(&sdp->sd_mh_cache_lock);
	}

	spin_unlock(&sdp->sd_mh_cache_lock);

	return;
}

/**
 * rgrp_verify - Verify that a resource group is consistent
 * @sdp: the filesystem
 * @rgd: the rgrp
 *
 * Somebody should have already called gfs_glock_rg() on this RG.
 */

static void
rgrp_verify(struct gfs_rgrpd *rgd)
{
	struct gfs_bitmap *bits = NULL;
	uint32_t length = rgd->rd_ri.ri_length;
	uint32_t count[4], tmp;
	int buf, x;

	for (x = 0; x < 4; x++) {
		count[x] = 0;

		for (buf = 0; buf < length; buf++) {
			bits = &rgd->rd_bits[buf];
			count[x] +=
			    gfs_bitcount(rgd,
					 (rgd->rd_bh[buf])->b_data +
					 bits->bi_offset, bits->bi_len, x);
		}
	}

	GFS_ASSERT_RGRPD(count[0] == rgd->rd_rg.rg_free, rgd,
			 printk("free data mismatch:  %u != %u\n",
				count[0], rgd->rd_rg.rg_free);
	    );

	tmp = rgd->rd_ri.ri_data -
	    (rgd->rd_rg.rg_usedmeta + rgd->rd_rg.rg_freemeta) -
	    (rgd->rd_rg.rg_useddi + rgd->rd_rg.rg_freedi) - rgd->rd_rg.rg_free;
	GFS_ASSERT_RGRPD(count[1] == tmp, rgd,
			 printk("used data mismatch:  %u != %u\n",
				count[1], tmp);
	    );

	GFS_ASSERT_RGRPD(count[2] == rgd->rd_rg.rg_freemeta, rgd,
			 printk("free metadata mismatch:  %u != %u\n",
				count[2], rgd->rd_rg.rg_freemeta);
	    );

	tmp =
	    rgd->rd_rg.rg_usedmeta + (rgd->rd_rg.rg_useddi +
				      rgd->rd_rg.rg_freedi);
	GFS_ASSERT_RGRPD(count[3] == tmp, rgd,
			 printk("used metadata mismatch:  %u != %u\n",
				count[3], tmp);
	    );

}

/**
 * gfs_blk2rgrpd - Find resource group for a given data block number
 * @sdp: The GFS superblock
 * @n: The data block number
 *
 * Returns: Ths resource group, or NULL if not found
 */

struct gfs_rgrpd *
gfs_blk2rgrpd(struct gfs_sbd *sdp, uint64_t blk)
{
	struct list_head *tmp, *head;
	struct gfs_rgrpd *rgd = NULL;
	struct gfs_rindex *ri;

	spin_lock(&sdp->sd_rg_mru_lock);

	for (head = &sdp->sd_rg_mru_list, tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		rgd = list_entry(tmp, struct gfs_rgrpd, rd_list_mru);
		ri = &rgd->rd_ri;

		if (ri->ri_data1 <= blk && blk < ri->ri_data1 + ri->ri_data) {
			list_del(&rgd->rd_list_mru);
			list_add(&rgd->rd_list_mru, &sdp->sd_rg_mru_list);
			break;
		} else
			rgd = NULL;
	}

	spin_unlock(&sdp->sd_rg_mru_lock);

	return rgd;
}

/**
 * gfs_rgrpd_get_first - get the first RG
 * @sdp: The GFS superblock
 *
 * Returns: The first rgrp in the filesystem
 */

struct gfs_rgrpd *
gfs_rgrpd_get_first(struct gfs_sbd *sdp)
{
	struct gfs_rgrpd *rgd;

	GFS_ASSERT_SBD(!list_empty(&sdp->sd_rglist), sdp,);
	rgd = list_entry(sdp->sd_rglist.next, struct gfs_rgrpd, rd_list);

	return rgd;
}

/**
 * gfs_rgrpd_get_next - get the next RG
 * @rgd: A RG
 *
 * Returns: The next rgrp
 */

struct gfs_rgrpd *
gfs_rgrpd_get_next(struct gfs_rgrpd *rgd)
{
	struct gfs_rgrpd *new_rgd = NULL;

	if (rgd->rd_list.next != &rgd->rd_sbd->sd_rglist)
		new_rgd =
		    list_entry(rgd->rd_list.next, struct gfs_rgrpd, rd_list);

	return new_rgd;
}

/**
 * gfs_clear_rgrpd - Clear up rgrps
 * @sdp: The GFS superblock
 *
 */

void
gfs_clear_rgrpd(struct gfs_sbd *sdp)
{
	struct gfs_rgrpd *rgd;
	struct gfs_glock *gl;

	sdp->sd_rg_forward = NULL;

	while (!list_empty(&sdp->sd_rg_recent)) {
		rgd =
		    list_entry(sdp->sd_rg_recent.next, struct gfs_rgrpd,
			       rd_recent);
		list_del(&rgd->rd_recent);
	}

	while (!list_empty(&sdp->sd_rglist)) {
		rgd =
		    list_entry(sdp->sd_rglist.next, struct gfs_rgrpd, rd_list);
		gl = rgd->rd_gl;

		list_del(&rgd->rd_list);
		list_del(&rgd->rd_list_mru);

		if (gl) {
			gfs_glock_force_drop(gl);
			if (atomic_read(&gl->gl_lvb_count))
				gfs_unhold_lvb(gl);
			gl->gl_object = NULL;
			gfs_put_glstruct(gl);
		}

		if (rgd->rd_bits)
			kfree(rgd->rd_bits);
		if (rgd->rd_bh)
			kfree(rgd->rd_bh);

		kfree(rgd);
	}

}

/**
 * gfs_compute_bitstructs - Compute the bitmap sizes
 * @rgd: The resource group descriptor
 *
 */

static void
compute_bitstructs(struct gfs_rgrpd *rgd)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_bitmap *bits;
	uint32_t length = rgd->rd_ri.ri_length;
	uint32_t bytes_left, bytes;
	int x;

	rgd->rd_bits = gmalloc(length * sizeof (struct gfs_bitmap));
	memset(rgd->rd_bits, 0, length * sizeof (struct gfs_bitmap));

	bytes_left = rgd->rd_ri.ri_bitbytes;

	for (x = 0; x < length; x++) {
		bits = &rgd->rd_bits[x];

		if (length == 1) {
			bytes = bytes_left;
			bits->bi_offset = sizeof (struct gfs_rgrp);
			bits->bi_start = 0;
			bits->bi_len = bytes;
		} else if (x == 0) {
			bytes = sdp->sd_sb.sb_bsize - sizeof (struct gfs_rgrp);
			bits->bi_offset = sizeof (struct gfs_rgrp);
			bits->bi_start = 0;
			bits->bi_len = bytes;
		} else if (x + 1 == length) {
			bytes = bytes_left;
			bits->bi_offset = sizeof (struct gfs_meta_header);
			bits->bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
			bits->bi_len = bytes;
		} else {
			bytes =
			    sdp->sd_sb.sb_bsize -
			    sizeof (struct gfs_meta_header);
			bits->bi_offset = sizeof (struct gfs_meta_header);
			bits->bi_start = rgd->rd_ri.ri_bitbytes - bytes_left;
			bits->bi_len = bytes;
		}

		bytes_left -= bytes;
	}

	GFS_ASSERT_RGRPD(!bytes_left, rgd,);
	GFS_ASSERT_RGRPD((rgd->rd_bits[length - 1].bi_start +
			  rgd->rd_bits[length - 1].bi_len) * GFS_NBBY ==
			 rgd->rd_ri.ri_data, rgd,
			 printk("start=%u len=%u offset=%u\n",
				rgd->rd_bits[length - 1].bi_start,
				rgd->rd_bits[length - 1].bi_len,
				rgd->rd_bits[length - 1].bi_offset);
			 gfs_rindex_print(&rgd->rd_ri, TRUE);
	    );

	rgd->rd_bh = gmalloc(length * sizeof (struct buffer_head *));
	memset(rgd->rd_bh, 0, length * sizeof (struct buffer_head *));

}

/**
 * gfs_ri_update - Pull in a new resource index from the disk
 * @gl: The glock covering the rindex inode
 *
 * Returns: 0 on successful update, error code otherwise
 */

static int
gfs_ri_update(struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_rgrpd *rgd;
	char buf[sizeof (struct gfs_rindex)];
	int error;

	GFS_ASSERT_SBD(!do_mod(ip->i_di.di_size, sizeof (struct gfs_rindex)),
		       sdp,);

	gfs_clear_rgrpd(sdp);

	for (sdp->sd_rgcount = 0;; sdp->sd_rgcount++) {
		error =
		    gfs_internal_read(ip, buf,
				      sdp->sd_rgcount *
				      sizeof (struct gfs_rindex),
				      sizeof (struct gfs_rindex));
		if (!error)
			break;
		if (error != sizeof (struct gfs_rindex)) {
			if (error > 0)
				error = -EIO;
			goto fail;
		}

		rgd = gmalloc(sizeof (struct gfs_rgrpd));
		memset(rgd, 0, sizeof (struct gfs_rgrpd));

		INIT_LIST_HEAD(&rgd->rd_mh_cache);
		rgd->rd_sbd = sdp;

		list_add_tail(&rgd->rd_list, &sdp->sd_rglist);
		list_add_tail(&rgd->rd_list_mru, &sdp->sd_rg_mru_list);

		gfs_rindex_in(&rgd->rd_ri, buf);

		compute_bitstructs(rgd);

		error =
		    gfs_get_glstruct(sdp, rgd->rd_ri.ri_addr, &gfs_rgrp_glops,
				     ST_CREATE, &rgd->rd_gl);
		if (error)
			goto fail;

		error = gfs_hold_lvb(rgd->rd_gl);
		if (error)
			goto fail;

		rgd->rd_gl->gl_object = rgd;
		rgd->rd_rg_vn = rgd->rd_rb_vn = rgd->rd_gl->gl_vn - 1;
	}

	sdp->sd_riinode_vn = ip->i_gl->gl_vn;

	return 0;

      fail:
	gfs_clear_rgrpd(sdp);

	return error;
}

/**
 * gfs_rindex_hold - Grab a lock on the rindex
 * @sdp: The GFS superblock
 * @ri_gh: the glock holder
 *
 * We grab a lock in the rindex inode to make sure that it doesn't
 * change whilst we are performing an operation. We keep this lock
 * for quite long periods of time compared to other locks. This
 * doesn't matter, since its shared and it is very, very rarely
 * accessed in the exclusive mode.
 *
 * Returns: 0 on success, error code otherwise
 */

int
gfs_rindex_hold(struct gfs_sbd *sdp, struct gfs_holder *ri_gh)
{
	struct gfs_inode *ip = sdp->sd_riinode;
	struct gfs_glock *gl = ip->i_gl;
	int error;

	error = gfs_glock_nq_init(gl, LM_ST_SHARED, 0, ri_gh);
	if (error)
		goto out;

	if (sdp->sd_riinode_vn != gl->gl_vn) {
		down(&sdp->sd_rindex_lock);
		if (sdp->sd_riinode_vn != gl->gl_vn) {
			error = gfs_ri_update(ip);
			if (error)
				gfs_glock_dq_uninit(ri_gh);
		}
		up(&sdp->sd_rindex_lock);
	}

      out:

	return error;
}

/**
 * gfs_rgrp_read - Read in a RG's bitmaps
 * @rgd: the struct gfs_rgrpd describing the RG to read in
 *
 * Read in RG bitmaps.  Must call gfs_rgrp_relse() it free the bitmaps.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_rgrp_read(struct gfs_rgrpd *rgd)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_glock *gl = rgd->rd_gl;
	unsigned int x, length = rgd->rd_ri.ri_length;
	int error;

	for (x = 0; x < length; x++) {
		GFS_ASSERT_RGRPD(!rgd->rd_bh[x], rgd,);
		rgd->rd_bh[x] = gfs_dgetblk(sdp, rgd->rd_ri.ri_addr + x, gl);
	}

	for (x = 0; x < length; x++) {
		error = gfs_dreread(sdp, rgd->rd_bh[x], DIO_START);
		if (error)
			goto fail;
	}

	for (x = length; x--;) {
		error = gfs_dreread(sdp, rgd->rd_bh[x], DIO_WAIT);
		if (error)
			goto fail;
		gfs_metatype_check(sdp, rgd->rd_bh[x],
				   (x) ? GFS_METATYPE_RB : GFS_METATYPE_RG);
	}

	if (rgd->rd_rg_vn != gl->gl_vn) {
		gfs_rgrp_in(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
		rgd->rd_rg_vn = gl->gl_vn;
	}

	return 0;

      fail:
	for (x = 0; x < length; x++) {
		gfs_drelse(rgd->rd_bh[x]);
		rgd->rd_bh[x] = NULL;
	}

	return error;
}

/**
 * gfs_rgrp_relse - Release RG bitmaps read in with gfs_rgrp_read()
 * @rgd: the struct gfs_rgrpd describing the RG to read in
 *
 */

void
gfs_rgrp_relse(struct gfs_rgrpd *rgd)
{
	int x, length = rgd->rd_ri.ri_length;

	for (x = 0; x < length; x++) {
		gfs_drelse(rgd->rd_bh[x]);
		rgd->rd_bh[x] = NULL;
	}

}

/**
 * rgrp_lvb_ok - check to see if a RG LVB looks accurate
 * @rgd: the RG
 *
 * Returns: TRUE if the RG LVB is ok
 */

static __inline__ int
rgrp_lvb_ok(struct gfs_rgrpd *rgd)
{
	struct gfs_rgrp_lvb *rb = &rgd->rd_rb;

	if (rb->rb_magic != GFS_MAGIC)
		return FALSE;

	if (!rb->rb_owner)
		return TRUE;

	if (rb->rb_owner == rgd->rd_sbd->sd_lockstruct.ls_jid + 1 &&
	    test_bit(RDF_LVB_OWNED, &rgd->rd_flags))
		return TRUE;

	return FALSE;
}

/**
 * gfs_rgrp_lvb_ok - export rgrp_lvb_ok to other files
 * @rgd: the RG
 *
 * Returns: TRUE if the RG LVB is ok
 */

int
gfs_rgrp_lvb_ok(struct gfs_rgrpd *rgd)
{
	int ret;
	ret = rgrp_lvb_ok(rgd);
	return ret;
}

/**
 * gfs_rgrp_lvb_fill - copy RG usage data out of the struct gfs_rgrp into the struct gfs_rgrp_lvb
 * @rgd: the resource group data structure
 *
 */

void
gfs_rgrp_lvb_fill(struct gfs_rgrpd *rgd)
{
	struct gfs_rgrp *rg = &rgd->rd_rg;
	struct gfs_rgrp_lvb *rb = &rgd->rd_rb;

	rb->rb_magic = GFS_MAGIC;
	rb->rb_free = rg->rg_free;
	rb->rb_useddi = rg->rg_useddi;
	rb->rb_freedi = rg->rg_freedi;
	rb->rb_usedmeta = rg->rg_usedmeta;
	rb->rb_freemeta = rg->rg_freemeta;

	set_bit(RDF_RB_DIRTY, &rgd->rd_flags);

}

/**
 * gfs_rgrp_lvb_init - Init the data of a RG LVB
 * @rgd: the resource group data structure
 *
 * Returns:  0 on success, -EXXX on failure
 */

int
gfs_rgrp_lvb_init(struct gfs_rgrpd *rgd)
{
	struct gfs_glock *gl = rgd->rd_gl;
	struct gfs_holder rgd_gh;
	int error;

	error = gfs_glock_nq_init(gl, LM_ST_EXCLUSIVE, 0, &rgd_gh);
	if (error)
		return error;

	gfs_rgrp_lvb_fill(rgd);
	rgd->rd_rb.rb_owner = 0;
	gfs_rgrp_lvb_out(&rgd->rd_rb, gl->gl_lvb);
	gfs_sync_lvb(gl);

	clear_bit(RDF_RB_DIRTY, &rgd->rd_flags);

	gfs_glock_dq_uninit(&rgd_gh);

	return 0;
}

/**
 * rgrp_lvb_own - Mark a RG LVB as being owned by a machine
 * @rgd: the resource group data structure
 *
 * If a machine ever sees a RG LVB that's owned by another
 * machine, it knows that the data in that LVB is stale because
 * the owning machine crashed.
 */

static void
rgrp_lvb_own(struct gfs_rgrpd *rgd)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_glock *gl = rgd->rd_gl;
	struct gfs_rgrp_lvb *rb = &rgd->rd_rb;

	if (!rgrp_lvb_ok(rgd)) {
		gfs_rgrp_lvb_fill(rgd);
		rb->rb_owner = 0;
	}
	if (rb->rb_owner != sdp->sd_lockstruct.ls_jid + 1) {
		rb->rb_owner = sdp->sd_lockstruct.ls_jid + 1;
		gfs_rgrp_lvb_out(rb, gl->gl_lvb);
		gfs_sync_lvb(gl);
	}

	set_bit(RDF_LVB_OWNED, &rgd->rd_flags);

}

/**
 * gfs_get_alloc - allocate a struct gfs_alloc structure for an inode
 * @ip: the inode
 *
 * Returns: the struct gfs_alloc
 */

struct gfs_alloc *
gfs_get_alloc(struct gfs_inode *ip)
{
	struct gfs_alloc *al = ip->i_alloc;

	GFS_ASSERT_INODE(!al, ip,);

	al = gmalloc(sizeof (struct gfs_alloc));
	memset(al, 0, sizeof (struct gfs_alloc));

	ip->i_alloc = al;

	return al;
}

/**
 * gfs_put_alloc - throw away the struct gfs_alloc for an inode
 * @ip: the inode
 *
 */

void
gfs_put_alloc(struct gfs_inode *ip)
{
	struct gfs_alloc *al = ip->i_alloc;

	GFS_ASSERT_INODE(al, ip,);

	kfree(al);

	ip->i_alloc = NULL;

	return;
}

/**
 * try_rgrp_fit - See if a given reservation will fit in a given RG
 * @rgd: the RG data
 * @al: the struct gfs_alloc structure describing the reservation
 *
 * Sets the $ir_datares field in @res.
 * Sets the $ir_metares field in @res.
 *
 * Returns: 1 on success, 0 on failure
 */

static int
try_rgrp_fit(struct gfs_rgrpd *rgd, struct gfs_alloc *al)
{
	uint32_t freeblks = rgd->rd_rg.rg_free;
	uint32_t freemeta = rgd->rd_rg.rg_freemeta;
	uint32_t metares = al->al_metablks;
	uint32_t datares = al->al_datablks;

	/*  First take care of the data blocks required  */

	if (freeblks < al->al_datablks)
		return 0;

	freeblks -= al->al_datablks;

	/*  Then take care of the dinodes  */

	if (rgd->rd_rg.rg_freedi < al->al_diblks)
		metares += al->al_diblks - rgd->rd_rg.rg_freedi;

	/*  Then take care of the metadata blocks  */

	while (freemeta < metares) {
		if (freeblks < GFS_META_CLUMP)
			return 0;

		freeblks -= GFS_META_CLUMP;
		freemeta += GFS_META_CLUMP;

		datares += GFS_META_CLUMP;
	}

	al->al_rgd = rgd;
	al->al_metares = metares;
	al->al_datares = datares;

	return 1;
}

/**
 * recent_rgrp_first - get first RG from recent list
 * @sdp: The GFS superblock
 * @rglast: address of the rgrp used last
 *
 * Returns: The first rgrp in the recent list
 */

static struct gfs_rgrpd *
recent_rgrp_first(struct gfs_sbd *sdp, uint64_t rglast)
{
	struct list_head *tmp, *head;
	struct gfs_rgrpd *rgd = NULL;

	spin_lock(&sdp->sd_rg_recent_lock);

	if (list_empty(&sdp->sd_rg_recent))
		goto out;

	if (!rglast)
		goto first;

	for (head = &sdp->sd_rg_recent, tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		rgd = list_entry(tmp, struct gfs_rgrpd, rd_recent);
		if (rgd->rd_ri.ri_addr == rglast)
			goto out;
	}

      first:
	rgd = list_entry(sdp->sd_rg_recent.next, struct gfs_rgrpd, rd_recent);

      out:
	spin_unlock(&sdp->sd_rg_recent_lock);

	return rgd;
}

/**
 * recent_rgrp_next - get next RG from recent list
 * @cur_rgd: current rgrp
 *
 * Returns: The next rgrp in the recent list
 */

static struct gfs_rgrpd *
recent_rgrp_next(struct gfs_rgrpd *cur_rgd)
{
	struct gfs_sbd *sdp = cur_rgd->rd_sbd;
	struct list_head *tmp, *head;
	struct gfs_rgrpd *rgd;

	spin_lock(&sdp->sd_rg_recent_lock);

	for (head = &sdp->sd_rg_recent, tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		rgd = list_entry(tmp, struct gfs_rgrpd, rd_recent);
		if (rgd == cur_rgd) {
			if (cur_rgd->rd_recent.next != &sdp->sd_rg_recent)
				rgd =
				    list_entry(cur_rgd->rd_recent.next,
					       struct gfs_rgrpd, rd_recent);
			else
				rgd = NULL;

			goto out;
		}
	}
	rgd = NULL;

      out:
	spin_unlock(&sdp->sd_rg_recent_lock);

	return rgd;
}

/**
 * recent_rgrp_remove - remove an RG from recent list
 * @rgd: The rgrp to remove
 *
 */

static void
recent_rgrp_remove(struct gfs_rgrpd *rgd)
{

	spin_lock(&rgd->rd_sbd->sd_rg_recent_lock);
	list_del(&rgd->rd_recent);
	spin_unlock(&rgd->rd_sbd->sd_rg_recent_lock);

}

/**
 * recent_rgrp_add - add an RG to recent list
 * @new_rgd: The rgrp to add
 *
 */

static void
recent_rgrp_add(struct gfs_rgrpd *new_rgd)
{
	struct gfs_sbd *sdp = new_rgd->rd_sbd;
	struct list_head *tmp, *head;
	struct gfs_rgrpd *rgd = NULL;
	unsigned int count = 0;
	unsigned int max = sdp->sd_rgcount / sdp->sd_journals;

	spin_lock(&sdp->sd_rg_recent_lock);

	for (head = &sdp->sd_rg_recent, tmp = head->next; tmp != head;
	     tmp = tmp->next) {
		rgd = list_entry(tmp, struct gfs_rgrpd, rd_recent);
		if (rgd == new_rgd)
			goto out;

		if (++count >= max)
			goto out;
	}
	list_add_tail(&new_rgd->rd_recent, &sdp->sd_rg_recent);

      out:
	spin_unlock(&sdp->sd_rg_recent_lock);

}

/**
 * forward_rgrp_get - get an rgrp to try next from full list
 * @sdp: The GFS superblock
 *
 * Returns: The rgrp to try next
 */

static struct gfs_rgrpd *
forward_rgrp_get(struct gfs_sbd *sdp)
{
	struct gfs_rgrpd *rgd;
	unsigned int rg = 0, x;

	spin_lock(&sdp->sd_rg_forward_lock);

	rgd = sdp->sd_rg_forward;
	if (!rgd) {
		if (sdp->sd_rgcount >= sdp->sd_journals)
			rg = sdp->sd_rgcount * sdp->sd_lockstruct.ls_jid /
			    sdp->sd_journals;

		for (x = 0, rgd = gfs_rgrpd_get_first(sdp); x < rg; x++, rgd = gfs_rgrpd_get_next(rgd))	/*  Do Nothing  */
			;

		sdp->sd_rg_forward = rgd;
	}

	spin_unlock(&sdp->sd_rg_forward_lock);

	return rgd;
}

/**
 * forward_rgrp_set - set the forward rgrp pointer
 * @sdp: the filesystem
 * @rgd: The new forward rgrp
 *
 */

static void
forward_rgrp_set(struct gfs_sbd *sdp, struct gfs_rgrpd *rgd)
{

	spin_lock(&sdp->sd_rg_forward_lock);
	sdp->sd_rg_forward = rgd;
	spin_unlock(&sdp->sd_rg_forward_lock);

}

/**
 * get_local_rgrp - Choose and lock a rgrp for allocation
 * @ip: the inode to reserve space for
 * @rgp: the chosen and locked rgrp
 *
 * Try to acquire rgrp in way which avoids contending with others.
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
get_local_rgrp(struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_rgrpd *rgd, *begin, *next = NULL;
	struct gfs_alloc *al = ip->i_alloc;
	int flags = LM_FLAG_TRY;
	int error = 0;
	int skipped = 0;
	int loops = 0;
	int update_recent = FALSE;

	/*  Try recently successful rgrps  */

	rgd = recent_rgrp_first(sdp, ip->i_last_rg_alloc);

	while (rgd) {
		error =
		    gfs_glock_nq_init(rgd->rd_gl, LM_ST_EXCLUSIVE, LM_FLAG_TRY,
				      &al->al_rgd_gh);
		switch (error) {
		case 0:
			if (try_rgrp_fit(rgd, al))
				goto out;

			next = recent_rgrp_next(rgd);
			recent_rgrp_remove(rgd);
			gfs_glock_dq_uninit(&al->al_rgd_gh);
			rgd = next;
			break;

		case GLR_TRYFAILED:
			rgd = recent_rgrp_next(rgd);
			break;

		default:
			GFS_ASSERT_RGRPD(error < 0, rgd,);
			goto fail;
		}
	}

	/*  Go through full list of rgrps  */

	update_recent = TRUE;
	begin = rgd = forward_rgrp_get(sdp);

	for (;;) {
		error =
		    gfs_glock_nq_init(rgd->rd_gl, LM_ST_EXCLUSIVE, flags,
				      &al->al_rgd_gh);
		switch (error) {
		case 0:
			if (try_rgrp_fit(rgd, al))
				goto out;
			gfs_glock_dq_uninit(&al->al_rgd_gh);
			break;

		case GLR_TRYFAILED:
			GFS_ASSERT_RGRPD(flags == LM_FLAG_TRY, rgd,);
			skipped++;
			break;

		default:
			GFS_ASSERT_RGRPD(error < 0, rgd,);
			goto fail;
		}

		rgd = gfs_rgrpd_get_next(rgd);
		if (!rgd)
			rgd = gfs_rgrpd_get_first(sdp);

		if (rgd == begin) {
			if (++loops >= 2 || !skipped) {
				error = -ENOSPC;
				goto fail;
			}
			flags = 0;
		}
	}

      out:
	ip->i_last_rg_alloc = rgd->rd_ri.ri_addr;

	if (update_recent) {
		recent_rgrp_add(rgd);
		rgd = gfs_rgrpd_get_next(rgd);
		forward_rgrp_set(sdp, rgd);
	}

      fail:

	return error;
}

/**
 * gfs_inplace_reserv - Reserve space in the filesystem
 * @ip: the inode to reserve space for
 *
 * Acquire resource group locks to allow for the maximum allocation
 * described by "res".
 *
 * This should probably become more complex again, but for now, let's go
 * for simple (one resource group) reservations.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_inplace_reserv(struct gfs_inode *ip)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al = ip->i_alloc;
	struct gfs_glock *gl;
	int error;

	GFS_ASSERT_INODE(al->al_diblks || al->al_datablks
			 || al->al_metablks, ip,);

	error = gfs_rindex_hold(sdp, &al->al_ri_gh);
	if (error)
		goto out;

	error = get_local_rgrp(ip);
	if (error) {
		gfs_glock_dq_uninit(&al->al_ri_gh);
		goto out;
	}

	gl = al->al_rgd->rd_gl;
	if (!list_empty(&gl->gl_depend))
		gfs_sync_dependencies(gl, 0);

      out:

	return error;
}

/**
 * gfs_inplace_relse - release an inplace reservation
 * @ip: the inode the reservation was taken out on
 *
 * Release a reservation made by gfs_inplace_reserv().
 */

void
gfs_inplace_relse(struct gfs_inode *ip)
{
	struct gfs_alloc *al = ip->i_alloc;

	GFS_ASSERT_INODE(al->al_dialloced <= al->al_diblks, ip,
			 printk
			 ("al_dialloced = %u, al_diblks = %u, trans = %d\n",
			  al->al_dialloced, al->al_diblks, al->al_trans);
	    );
	GFS_ASSERT_INODE(al->al_metaalloced <= al->al_metares, ip,
			 printk
			 ("al_metaalloced = %u, al_metablks = %u, trans = %d\n",
			  al->al_metaalloced, al->al_metablks, al->al_trans);
	    );
	GFS_ASSERT_INODE(al->al_dataalloced <= al->al_datares, ip,
			 printk
			 ("al_dataalloced = %u, al_datares = %u, trans = %d\n",
			  al->al_dataalloced, al->al_datares, al->al_trans);
	    );

	al->al_rgd = NULL;
	gfs_glock_dq_uninit(&al->al_rgd_gh);
	gfs_glock_dq_uninit(&al->al_ri_gh);

}

/**
 * gfs_get_block_type - Check a block in a RG is of given type
 * @rgd: the resource group holding the block
 * @block: the block number
 *
 * Returns: The block type (GFS_BLKST_*)
 */

unsigned char
gfs_get_block_type(struct gfs_rgrpd *rgd, uint64_t block)
{
	struct gfs_bitmap *bits = NULL;
	uint32_t length, rgrp_block, buf_block;
	unsigned int buf;
	unsigned char type;

	length = rgd->rd_ri.ri_length;
	rgrp_block = block - rgd->rd_ri.ri_data1;

	for (buf = 0; buf < length; buf++) {
		bits = &rgd->rd_bits[buf];
		if (rgrp_block < (bits->bi_start + bits->bi_len) * GFS_NBBY)
			break;
	}

	GFS_ASSERT_RGRPD(buf < length, rgd,);
	buf_block = rgrp_block - bits->bi_start * GFS_NBBY;

	type = gfs_testbit(rgd,
			   (rgd->rd_bh[buf])->b_data + bits->bi_offset,
			   bits->bi_len, buf_block);

	return type;
}

/**
 * blkalloc_internal - allocate a single block
 * @rgd: the resource group descriptor
 * @goal: the goal block in the RG
 * @old_state: the type of block to find
 * @new_state: the resulting block type
 * @do_it: if FALSE, we just find the block we would allocate
 *
 * This function never fails.
 *
 * Returns:  returns the block allocated
 */

static uint32_t
blkalloc_internal(struct gfs_rgrpd *rgd,
		  uint32_t goal,
		  unsigned char old_state, unsigned char new_state, int do_it)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_bitmap *bits = NULL;
	uint32_t length = rgd->rd_ri.ri_length;
	uint32_t blk = 0;
	unsigned int buf, x;

	for (buf = 0; buf < length; buf++) {
		bits = &rgd->rd_bits[buf];
		if (goal < (bits->bi_start + bits->bi_len) * GFS_NBBY)
			break;
	}

	GFS_ASSERT_RGRPD(buf < length, rgd,);
	goal -= bits->bi_start * GFS_NBBY;

	/*  "x <= length" because we're skipping over some of the first
	   buffer when the goal is non-zero.  */

	for (x = 0; x <= length; x++) {
		blk = gfs_bitfit(rgd,
				 (rgd->rd_bh[buf])->b_data + bits->bi_offset,
				 bits->bi_len, goal, old_state);
		if (blk != BFITNOENT)
			break;

		buf = (buf + 1) % length;
		bits = &rgd->rd_bits[buf];
		goal = 0;
	}

	GFS_ASSERT_RGRPD(x <= length, rgd,);

	if (do_it) {
		gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[buf]);
		gfs_setbit(rgd,
			   (rgd->rd_bh[buf])->b_data + bits->bi_offset,
			   bits->bi_len, blk, new_state);
	}

	return bits->bi_start * GFS_NBBY + blk;
}

/**
 * blkfree_internal - Free a block
 * @sdp: the filesystem
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 * @new_state: the new state of the block
 *
 */

static struct gfs_rgrpd *
blkfree_internal(struct gfs_sbd *sdp, uint64_t bstart, uint32_t blen,
		 unsigned char new_state)
{
	struct gfs_rgrpd *rgd;
	struct gfs_bitmap *bits = NULL;
	uint32_t length, rgrp_blk, buf_blk;
	unsigned int buf;

	rgd = gfs_blk2rgrpd(sdp, bstart);
	GFS_ASSERT_SBD(rgd, sdp, printk("block = %" PRIu64 "\n", bstart);
	    );

	length = rgd->rd_ri.ri_length;
	rgrp_blk = bstart - rgd->rd_ri.ri_data1;

	while (blen--) {
		for (buf = 0; buf < length; buf++) {
			bits = &rgd->rd_bits[buf];
			if (rgrp_blk <
			    (bits->bi_start + bits->bi_len) * GFS_NBBY)
				break;
		}

		GFS_ASSERT_RGRPD(buf < length, rgd,);
		buf_blk = rgrp_blk - bits->bi_start * GFS_NBBY;
		rgrp_blk++;

		gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[buf]);
		gfs_setbit(rgd,
			   (rgd->rd_bh[buf])->b_data + bits->bi_offset,
			   bits->bi_len, buf_blk, new_state);
	}

	return rgd;
}

/**
 * clump_alloc - Allocate a clump of metadata
 * @rgd: the resource group descriptor
 * @first: returns the first block allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
clump_alloc(struct gfs_rgrpd *rgd, uint32_t * first)
{
	struct gfs_sbd *sdp = rgd->rd_sbd;
	struct gfs_meta_header mh;
	struct buffer_head **bh;
	uint32_t goal, blk;
	unsigned int x;
	int error = 0;

	memset(&mh, 0, sizeof (struct gfs_meta_header));
	mh.mh_magic = GFS_MAGIC;
	mh.mh_type = GFS_METATYPE_NONE;

	bh = gmalloc(GFS_META_CLUMP * sizeof(struct buffer_head *));
	memset(bh, 0, sizeof (GFS_META_CLUMP * sizeof (struct buffer_head *)));

	goal = rgd->rd_last_alloc_data;

	for (x = 0; x < GFS_META_CLUMP; x++) {
		blk = blkalloc_internal(rgd, goal, GFS_BLKST_FREE,
					GFS_BLKST_FREEMETA, TRUE);
		if (!x && first)
			*first = blk;

		bh[x] = gfs_dgetblk(sdp, rgd->rd_ri.ri_data1 + blk, rgd->rd_gl);

		gfs_prep_new_buffer(bh[x]);

		gfs_meta_header_out(&mh, (bh[x])->b_data);
		((struct gfs_meta_header *) (bh[x])->b_data)->mh_generation = 0;	/*  Zero out the generation number.  */

		error = gfs_dwrite(sdp, bh[x], DIO_DIRTY | DIO_START);
		if (error)
			goto out;

		goal = blk;
	}

	rgd->rd_last_alloc_data = goal;

	for (x = 0; x < GFS_META_CLUMP; x++) {
		error = gfs_dwrite(sdp, bh[x], DIO_WAIT);
		if (error)
			goto out;
	}

	gfs_add_to_mh_cache(rgd, bh, GFS_META_CLUMP);

	GFS_ASSERT_RGRPD(rgd->rd_rg.rg_free >= GFS_META_CLUMP, rgd,);
	rgd->rd_rg.rg_free -= GFS_META_CLUMP;
	rgd->rd_rg.rg_freemeta += GFS_META_CLUMP;

      out:
	for (x = 0; x < GFS_META_CLUMP; x++)
		if (bh[x]) {
			gfs_dwrite(sdp, bh[x], DIO_WAIT);
			gfs_drelse(bh[x]);
		}
	kfree(bh);

	return error;
}

/**
 * gfs_blkalloc - Allocate a data block
 * @ip: the inode to allocate the data block for
 * @block: the block allocated
 *
 */

void
gfs_blkalloc(struct gfs_inode *ip, uint64_t * block)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al = ip->i_alloc;
	struct gfs_rgrpd *rgd = al->al_rgd;
	uint32_t goal, blk;
	int same;

	GFS_ASSERT_INODE(rgd, ip,);

	same = (rgd->rd_ri.ri_addr == ip->i_di.di_goal_rgrp);
	goal = (same) ? ip->i_di.di_goal_dblk : rgd->rd_last_alloc_data;

	blk =
	    blkalloc_internal(rgd, goal, GFS_BLKST_FREE, GFS_BLKST_USED, TRUE);
	rgd->rd_last_alloc_data = blk;

	if (!same) {
		ip->i_di.di_goal_rgrp = rgd->rd_ri.ri_addr;
		ip->i_di.di_goal_mblk = 0;
	}
	ip->i_di.di_goal_dblk = blk;

	*block = rgd->rd_ri.ri_data1 + blk;

	GFS_ASSERT_RGRPD(rgd->rd_rg.rg_free, rgd,);
	rgd->rd_rg.rg_free--;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	al->al_dataalloced++;

	gfs_trans_add_quota(sdp, +1, ip->i_di.di_uid, ip->i_di.di_gid);

	return;
}

/**
 * gfs_metaalloc - Allocate a metadata block to a file
 * @ip:  the file
 * @block: the block allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_metaalloc(struct gfs_inode *ip, uint64_t * block)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_alloc *al = ip->i_alloc;
	struct gfs_rgrpd *rgd = al->al_rgd;
	uint32_t goal, blk;
	int same;
	int error = 0;

	GFS_ASSERT_INODE(rgd, ip,);

	same = (rgd->rd_ri.ri_addr == ip->i_di.di_goal_rgrp);

	if (!rgd->rd_rg.rg_freemeta) {
		error = clump_alloc(rgd, &goal);
		if (error)
			return error;

		al->al_dataalloced += GFS_META_CLUMP;
	} else
		goal = (same) ? ip->i_di.di_goal_mblk : rgd->rd_last_alloc_meta;

	blk =
	    blkalloc_internal(rgd, goal, GFS_BLKST_FREEMETA, GFS_BLKST_USEDMETA,
			      TRUE);
	rgd->rd_last_alloc_meta = blk;

	if (!same) {
		ip->i_di.di_goal_rgrp = rgd->rd_ri.ri_addr;
		ip->i_di.di_goal_dblk = 0;
	}
	ip->i_di.di_goal_mblk = blk;

	*block = rgd->rd_ri.ri_data1 + blk;

	GFS_ASSERT_RGRPD(rgd->rd_rg.rg_freemeta, rgd,);
	rgd->rd_rg.rg_freemeta--;
	rgd->rd_rg.rg_usedmeta++;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	al->al_metaalloced++;

	gfs_trans_add_quota(sdp, +1, ip->i_di.di_uid, ip->i_di.di_gid);

	return error;
}

/**
 * gfs_dipick - Look for a free dinode
 * @dip: the directory that the inode is going in
 * @inum: the inode that the allocator will pick when you call gfs_dialloc()
 *
 */

void
gfs_dipick(struct gfs_inode *dip, struct gfs_inum *inum)
{
	struct gfs_rgrpd *rgd = dip->i_alloc->al_rgd;
	unsigned char from, to;
	uint32_t goal;

	GFS_ASSERT_INODE(rgd, dip,);

	/*  If there are free dinodes use those first  */

	if (rgd->rd_rg.rg_freedi) {
		*inum = rgd->rd_rg.rg_freedi_list;
		GFS_ASSERT_RGRPD(inum->no_formal_ino && inum->no_addr, rgd,);
		return;
	}

	/*  Else, pick a metadata block  */

	if (rgd->rd_rg.rg_freemeta) {
		from = GFS_BLKST_FREEMETA;
		to = GFS_BLKST_USEDMETA;
		goal = rgd->rd_last_alloc_meta;
	} else {
		from = GFS_BLKST_FREE;
		to = GFS_BLKST_FREEMETA;
		goal = rgd->rd_last_alloc_data;
	}

	inum->no_formal_ino = 0;
	inum->no_addr = blkalloc_internal(rgd, goal, from, to, FALSE);
	inum->no_addr += rgd->rd_ri.ri_data1;

	return;
}

/**
 * dialloc_preexist - Allocate a preexisting unused dinode
 * @dip: the directory that the inode is going in
 * @expected_inum: the number of the dinode to be allocated
 * @expected_gl: the glock of the dinode to be allocated
 *
 * Returns: 0 on success, -EXXX on failure
 */

static int
dialloc_preexist(struct gfs_inode *dip,
		 struct gfs_inum *expected_inum, struct gfs_glock *expected_gl)
{
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_alloc *al = dip->i_alloc;
	struct gfs_rgrpd *rgd = al->al_rgd;
	struct gfs_dinode *di;
	struct buffer_head *bh;
	struct gfs_inum next;
	uint32_t flags;
	int error;

	GFS_ASSERT_RGRPD(GFS_INUM_EQUAL
			 (&rgd->rd_rg.rg_freedi_list, expected_inum), rgd,);

	error =
	    gfs_dread(sdp, expected_inum->no_addr, expected_gl,
		      DIO_START | DIO_WAIT, &bh);
	if (error)
		return error;

	gfs_metatype_check(sdp, bh, GFS_METATYPE_DI);

	di = (struct gfs_dinode *) (bh)->b_data;

	flags = di->di_flags;
	flags = gfs32_to_cpu(flags);
	GFS_ASSERT_RGRPD(flags & GFS_DIF_UNUSED, rgd,);

	gfs_inum_in(&next, (char *) &di->di_next_unused);

	gfs_drelse(bh);

	rgd->rd_rg.rg_freedi--;
	rgd->rd_rg.rg_useddi++;
	rgd->rd_rg.rg_freedi_list = next;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	al->al_dialloced++;

	return error;
}

/**
 * gfs_dialloc - Allocate a dinode
 * @dip: the directory that the inode is going in
 * @expected_inum: The expected inode number
 * @expected_gl: the glock of the dinode to be allocated
 *
 * Assumes expected_gl is held exclusively by the calling process.
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_dialloc(struct gfs_inode *dip,
	    struct gfs_inum *expected_inum, struct gfs_glock *expected_gl)
{
	struct gfs_sbd *sdp = dip->i_sbd;
	struct gfs_alloc *al = dip->i_alloc;
	struct gfs_rgrpd *rgd = al->al_rgd;
	uint32_t goal, blk;
	int error = 0;

	GFS_ASSERT_INODE(rgd, dip,);

	if (rgd->rd_rg.rg_freedi) {
		error = dialloc_preexist(dip, expected_inum, expected_gl);
		return error;
	}

	if (!rgd->rd_rg.rg_freemeta) {
		error = clump_alloc(rgd, NULL);
		if (error)
			return error;

		al->al_dataalloced += GFS_META_CLUMP;
	}

	goal = expected_inum->no_addr - rgd->rd_ri.ri_data1;
	blk =
	    blkalloc_internal(rgd, goal, GFS_BLKST_FREEMETA, GFS_BLKST_USEDMETA,
			      TRUE);
	rgd->rd_last_alloc_meta = blk;

	GFS_ASSERT_RGRPD(rgd->rd_ri.ri_data1 + blk == expected_inum->no_addr,
			 rgd,);

	GFS_ASSERT_RGRPD(rgd->rd_rg.rg_freemeta, rgd,);
	rgd->rd_rg.rg_freemeta--;
	rgd->rd_rg.rg_useddi++;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	al->al_dialloced++;
	al->al_metaalloced++;

	return error;
}

/**
 * gfs_blkfree - free a piece of data
 * @ip: the inode these blocks are being free from
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 *
 */

void
gfs_blkfree(struct gfs_inode *ip, uint64_t bstart, uint32_t blen)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_rgrpd *rgd;

	rgd = blkfree_internal(sdp, bstart, blen, GFS_BLKST_FREE);

	rgd->rd_rg.rg_free += blen;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	gfs_trans_add_quota(sdp, -(int64_t) blen, ip->i_di.di_uid,
			    ip->i_di.di_gid);
	gfs_wipe_buffers(ip, rgd, bstart, blen, FALSE);

	return;
}

/**
 * gfs_metafree - free a piece of metadata
 * @ip: the inode these blocks are being free from
 * @bstart: the start of a run of blocks to free
 * @blen: the length of the block run
 *
 */

void
gfs_metafree(struct gfs_inode *ip, uint64_t bstart, uint32_t blen)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct gfs_rgrpd *rgd;

	rgd = blkfree_internal(sdp, bstart, blen, GFS_BLKST_FREEMETA);

	GFS_ASSERT_RGRPD(rgd->rd_rg.rg_usedmeta >= blen, rgd,);
	rgd->rd_rg.rg_usedmeta -= blen;
	rgd->rd_rg.rg_freemeta += blen;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	gfs_trans_add_quota(sdp, -(int64_t) blen, ip->i_di.di_uid,
			    ip->i_di.di_gid);
	gfs_wipe_buffers(ip, rgd, bstart, blen, TRUE);

	return;
}

/**
 * gfs_difree - free a piece of metadata
 * @ip: the inode representing the dinode to free
 * @rgd: the resource group that contains the dinode
 *
 * Returns: 0 on success, -EXXX on failure
 */

int
gfs_difree(struct gfs_inode *ip, struct gfs_rgrpd *rgd)
{
	struct gfs_sbd *sdp = ip->i_sbd;
	struct buffer_head *dibh;
	struct gfs_dinode *di;
	struct gfs_rgrpd *tmp_rgd;
	uint32_t incarn;
	uint64_t ctime;
	uint32_t flags;
	int error;

	tmp_rgd =
	    blkfree_internal(sdp, ip->i_num.no_addr, 1, GFS_BLKST_FREEMETA);
	GFS_ASSERT_INODE(rgd == tmp_rgd, ip,);

	error = gfs_get_inode_buffer(ip, &dibh);
	if (error)
		return error;

	di = (struct gfs_dinode *) (dibh)->b_data;

	gfs_trans_add_bh(sdp, ip->i_gl, dibh);

	incarn = gfs32_to_cpu(di->di_header.mh_incarn) + 1;
	di->di_header.mh_incarn = cpu_to_gfs32(incarn);

	ctime = CURRENT_TIME;
	di->di_ctime = cpu_to_gfs64(ctime);

	flags = (gfs32_to_cpu(di->di_flags)) | GFS_DIF_UNUSED;
	di->di_flags = cpu_to_gfs32(flags);

	gfs_drelse(dibh);

	GFS_ASSERT_RGRPD(rgd->rd_rg.rg_useddi, rgd,);
	rgd->rd_rg.rg_useddi--;
	rgd->rd_rg.rg_freemeta++;

	gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
	gfs_rgrp_out(&rgd->rd_rg, (rgd->rd_bh[0])->b_data);
	rgrp_lvb_own(rgd);

	gfs_trans_add_quota(sdp, -1, ip->i_di.di_uid, ip->i_di.di_gid);
	gfs_wipe_buffers(ip, rgd, ip->i_num.no_addr, 1, TRUE);

	return error;
}

/**
 * gfs_rlist_add - add a RG to a list of RGs
 * @sdp: the filesystem
 * @rlist: the list of resource groups
 * @block: the block
 *
 * Figure out what RG a block belongs to and add that RG to the list
 *
 */

void
gfs_rlist_add(struct gfs_sbd *sdp, struct gfs_rgrp_list *rlist, uint64_t block)
{
	struct gfs_rgrpd *rgd;
	struct gfs_rgrpd **tmp;
	unsigned int new_space;
	unsigned int x;

	GFS_ASSERT_SBD(rlist->rl_rgrps <= rlist->rl_space, sdp,);
	GFS_ASSERT_SBD(!rlist->rl_ghs, sdp,);

	rgd = gfs_blk2rgrpd(sdp, block);
	GFS_ASSERT_SBD(rgd, sdp, printk("block = %" PRIu64 "\n", block);
	    );

	for (x = 0; x < rlist->rl_rgrps; x++)
		if (rlist->rl_rgd[x] == rgd)
			return;

	if (rlist->rl_rgrps == rlist->rl_space) {
		new_space = rlist->rl_space + 10;

		tmp = gmalloc(new_space * sizeof (struct gfs_rgrpd *));

		if (rlist->rl_rgd) {
			memcpy(tmp, rlist->rl_rgd,
			       rlist->rl_space * sizeof (struct gfs_rgrpd *));
			kfree(rlist->rl_rgd);
		}

		rlist->rl_space = new_space;
		rlist->rl_rgd = tmp;
	}

	rlist->rl_rgd[rlist->rl_rgrps++] = rgd;
}

/**
 * gfs_rlist_alloc - all RGs have been added to the rlist, allocated holders for them
 * @rlist: the list of resource groups
 * @state: the lock state to acquire the RG lock in
 * @flags: the modifier flags for the holder structures
 *
 */

void
gfs_rlist_alloc(struct gfs_rgrp_list *rlist, unsigned int state, int flags)
{
	unsigned int x;

	rlist->rl_ghs = gmalloc(rlist->rl_rgrps * sizeof (struct gfs_holder));
	for (x = 0; x < rlist->rl_rgrps; x++)
		gfs_init_holder(rlist->rl_rgd[x]->rd_gl, state, flags,
				&rlist->rl_ghs[x]);

}

/**
 * gfs_rlist_free - free a resource group list
 * @list: the list of resource groups
 *
 */

void
gfs_rlist_free(struct gfs_rgrp_list *rlist)
{
	unsigned int x;

	if (rlist->rl_rgd)
		kfree(rlist->rl_rgd);

	if (rlist->rl_ghs) {
		for (x = 0; x < rlist->rl_rgrps; x++)
			gfs_uninit_holder(&rlist->rl_ghs[x]);
		kfree(rlist->rl_ghs);
	}

}

/**
 * gfs_reclaim_metadata - reclaims unused metadata
 * @sdp: the file system
 * @stats: stats on reclaimation
 *
 * This function will look through the resource groups and
 * free the unused metadata.
 *
 * Returns: 0 on success, -EXXX on error
 */

int
gfs_reclaim_metadata(struct gfs_sbd *sdp, struct gfs_reclaim_stats *stats)
{
	struct gfs_holder ji_gh, ri_gh, rgd_gh, t_gh;
	struct gfs_rgrpd *rgd;
	struct gfs_rgrp *rg;
	struct gfs_dinode *di;
	struct gfs_inum next;
	struct buffer_head *bh;
	uint32_t flags;
	uint32_t goal;
	unsigned int x;
	int error = 0;

	/*  Acquire the jindex lock here so we don't deadlock with a
	   process writing the the jindex inode.  :-(  */

	error = gfs_jindex_hold(sdp, &ji_gh);
	if (error)
		goto fail;

	error = gfs_rindex_hold(sdp, &ri_gh);
	if (error)
		goto fail_jindex_relse;

	for (rgd = gfs_rgrpd_get_first(sdp); rgd; rgd = gfs_rgrpd_get_next(rgd)) {
		error =
		    gfs_glock_nq_init(rgd->rd_gl, LM_ST_EXCLUSIVE, GL_NOCACHE,
				      &rgd_gh);
		if (error)
			goto fail_rindex_relse;

		rgrp_verify(rgd);

		rg = &rgd->rd_rg;

		if (!rg->rg_freedi && !rg->rg_freemeta) {
			gfs_glock_dq_uninit(&rgd_gh);
			continue;
		}

		gfs_zap_mh_cache(rgd);

		error = gfs_lock_fs_check_clean(sdp, LM_ST_EXCLUSIVE, &t_gh);
		if (error)
			goto fail_gunlock_rg;

		error =
		    gfs_trans_begin(sdp, TRANS_RECLAIM, rgd->rd_ri.ri_length,
				    0);
		if (error)
			goto fail_unlock_fs;

		next = rg->rg_freedi_list;

		for (x = rg->rg_freedi; x--;) {
			GFS_ASSERT_RGRPD(next.no_formal_ino
					 && next.no_addr, rgd,);

			blkfree_internal(sdp, next.no_addr, 1, GFS_BLKST_FREE);

			error =
			    gfs_dread(sdp, next.no_addr, rgd->rd_gl,
				      DIO_FORCE | DIO_START | DIO_WAIT, &bh);
			if (error)
				goto fail_end_trans;

			di = (struct gfs_dinode *) (bh)->b_data;
			flags = di->di_flags;
			flags = gfs32_to_cpu(flags);
			GFS_ASSERT_RGRPD(flags & GFS_DIF_UNUSED, rgd,);

			gfs_inum_in(&next, (char *) &di->di_next_unused);

			gfs_drelse(bh);

			rg->rg_freedi--;
			rg->rg_free++;
			stats->rc_inodes++;
		}

		GFS_ASSERT_RGRPD(!next.no_formal_ino && !next.no_addr, rgd,);
		rg->rg_freedi_list = next;

		goal = 0;
		for (x = rg->rg_freemeta; x--;) {
			goal =
			    blkalloc_internal(rgd, goal, GFS_BLKST_FREEMETA,
					      GFS_BLKST_FREE, TRUE);
			rg->rg_freemeta--;
			rg->rg_free++;
			stats->rc_metadata++;
		}

		gfs_trans_add_bh(sdp, rgd->rd_gl, rgd->rd_bh[0]);
		gfs_rgrp_out(rg, (rgd->rd_bh[0])->b_data);

		gfs_trans_end(sdp);

		gfs_glock_dq_uninit(&t_gh);

		gfs_glock_dq_uninit(&rgd_gh);
	}

	gfs_glock_dq_uninit(&ri_gh);

	gfs_glock_dq_uninit(&ji_gh);

	return 0;

      fail_end_trans:
	gfs_trans_end(sdp);

      fail_unlock_fs:
	gfs_glock_dq_uninit(&t_gh);

      fail_gunlock_rg:
	gfs_glock_dq_uninit(&rgd_gh);

      fail_rindex_relse:
	gfs_glock_dq_uninit(&ri_gh);

      fail_jindex_relse:
	gfs_glock_dq_uninit(&ji_gh);

      fail:
	return error;
}
