/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif				/*  MODVERSIONS  */
#include <asm/uaccess.h>
#include "pool.h"
#include "fixed_div64.h"

typedef struct label_list_s {
	pool_label_t ll_label;
	kdev_t ll_dev;
	struct label_list_s *next;
} label_list_t;

extern pool_t pools[POOL_MAX_DEVICES];
extern int pool_size[POOL_MAX_DEVICES];
extern int mp_type;
extern int mp_stripe;

/*********************
 ** local functions **
 ********************/
static int pool_loaded(char *);
static label_list_t *find_and_sort_labels(void);

static int
pool_loaded(char *poolname)
{
	int i;
	for (i = 0; i < POOL_MAX_DEVICES; i++) {
		down_read(&(pools[i].p_rwsema));
		if ((!strcmp(pools[i].p_name, poolname)) &&
		    (pools[i].p_flags & (PF_LOWEROPEN | PF_VALID))) {
			up_read(&(pools[i].p_rwsema));
			return 1;
		} else {
			up_read(&(pools[i].p_rwsema));
		}
	}
	return 0;
}

static label_list_t *
find_and_sort_labels(void)
{
	int i;
	int num_devices = 0;
	kdev_t *devices = NULL;
	pool_label_t tmp_label;
	label_list_t *tmp, *head, *sorted, *smallest;

	smallest = sorted = tmp = head = NULL;

  /** Find all the labels **/
	num_devices = read_partition_list(&devices);
	if (num_devices <= 0) {
		goto fail;
	}
	for (i = 0; i < num_devices; i++) {
		switch (scan_dev_for_pool_label(devices[i], &tmp_label)) {
		case 0:	/* Label not found */
			break;
		case 1:	/* Label found */
			if (!
			    (tmp =
			     (label_list_t *) pmalloc(sizeof (label_list_t)))) {
				goto fail;
			}
			memcpy(&(tmp->ll_label), &tmp_label,
			       sizeof (pool_label_t));
			tmp->ll_dev = devices[i];
			tmp->next = head;
			head = tmp;
			break;
		default:	/* error */
			/* goto fail; 
			   Don't fail, simply ignore that device. */
			break;
		}
	}
	pfree(devices, num_devices * sizeof (kdev_t));

  /** sort the labels... **/
	while (head) {
		for (tmp = head, smallest = head; tmp; tmp = tmp->next) {
			if (strcmp(smallest->ll_label.pl_pool_name,
				   tmp->ll_label.pl_pool_name) < 0) {
				smallest = tmp;
			}
		}
		tmp = head;
		if (tmp == smallest) {
			head = tmp->next;
		} else {
			for (; tmp->next != smallest; tmp = tmp->next) ;
			tmp->next = smallest->next;
		}
		smallest->next = sorted;
		sorted = smallest;
	}

	return sorted;

      fail:
	while (head) {
		tmp = head;
		head = head->next;
		pfree(tmp, sizeof (label_list_t));
	}
	if (num_devices > 0) {
		pfree(devices, num_devices * sizeof (kdev_t));
	}
	return NULL;
}

/**
 * open_devices - open all pool devices block and memexp
 * @pp
 * @beginning
 *
 * (pool_t *)pp is the pool to have its devices opened.
 * (int)beginning is the subpool to begin opening devices at.
 * In the case of a new pool (or one that isn't open) this
 * beginning would be zero.  However, if the pool is being grown,
 * then beginning should be the index of the first of the added
 * subpools.
 *
 * Returns: 0 on success, 1 on read-only, <0 on failure
 */

int
open_devices(pool_t * pp, int beginning)
{
	subpool_t *sp;
	device_t *devp;
	kdev_t dev;
	int i, j, k;
	int readonly = 0;
	int error;

	for (j = beginning; j < pp->p_num_spools; j++) {
		sp = &(pp->p_spools[j]);
		for (i = 0; i < sp->sp_devs; i++) {
			devp = &(sp->sp_devlist[i]);
			write_lock(&(devp->d_rwlock));
			for (k = 0; k < devp->d_num_good_paths; k++) {
				dev = devp->d_dev[k];

				error = bdopen(dev);
				if(error > 0){
					readonly = 1;
				} else if (error < 0) {
					printk
					    ("Pool: open failed: error %d opening %d,%d\n",
					     error, MAJOR(dev), MINOR(dev));
					k--;
					write_unlock(&(devp->d_rwlock));
					goto disk_backout;
				}
			}
			write_unlock(&(devp->d_rwlock));
		}
	}

	return (readonly);

      disk_backout:
	for (; j >= beginning; j--) {
		sp = &(pp->p_spools[j]);
		for (; i >= 0; i--) {
			devp = &(sp->sp_devlist[i]);
			write_lock(&(devp->d_rwlock));
			for (; k >= 0; k--) {
				dev = devp->d_dev[0];
				bdclose(dev);
			}
			write_unlock(&(devp->d_rwlock));
		}
	}
	return (error);
}

/**
 * close_devices - close all device pool has open
 * @pp: pointer to the pool structure
 *
 * Returns: Linux errno
 */

int
close_devices(pool_t * pp)
{
	subpool_t *sp;
	device_t *devp;
	kdev_t dev;
	int i, j, k;
	int error = 0;

	/* Close all storage devices */

	for (j = 0; j < pp->p_num_spools; j++) {

		sp = &(pp->p_spools[j]);

		for (i = 0; i < sp->sp_devs; i++) {

			devp = &(sp->sp_devlist[i]);
			read_lock(&(devp->d_rwlock));
			for (k = 0; k < devp->d_num_good_paths; k++) {
				dev = devp->d_dev[k];

				fsync_dev(dev);
				error = bdclose(dev);
				if (error)
					printk
					    ("Pool: close_devices: error %d\n",
					     error);
			}
			read_unlock(&(devp->d_rwlock));
		}

	}

	return (0);
}

/**
 * gen_pool
 * @add: the structure containing info on pool to be made
 * @pp: the pool_t structure to be filled
 *
 * This function takes the information in the pool_add_t structure, which
 * is basically pool information which was read from pool labels on the
 * disk, reformatts it, and copies it into the pool_t structure.
 *
 * Returns: Linux errno
 */

static int
gen_pool(pool_t * add, pool_t * pp)
{
	subpool_t *sp = NULL;
	device_t *nd = NULL;
	int i, j;
	int size;
	int error = 0;

	pp->p_blksize = 0;	/* this is obiviously a special case */

	/* Perform a few simple tests.  Most error checking should be done by 
	 * passemble.  Here we just check the pool version.
	 * The on-disk pool label must match down to the subrevison.  Thus, versions
	 * 0.3.0 and 0.4.0 don't match and are assumed to have different on disk
	 * structures.   On the other hand opening a 0.4.2 label with a 0.4.1 or
	 * 0.4.5 kernel is not an error */

	if ((add->p_version >> 16) != (POOL_VERSION >> 16) ||
	    ((add->p_version >> 8) & 0xFF) != ((POOL_VERSION >> 8) & 0xFF)) {
		printk("Pool: ----------> %x <-----------\n", add->p_version);
		printk("Pool: Version mismatch in pool: %s:\n", add->p_name);
		printk("Pool: \ton disk: v%d.%d.%d\n", add->p_version >> 16,
		       (add->p_version >> 8) & 0xFF, add->p_version & 0xFF);
		printk("Pool: \tin core: v%d.%d.%d\n", POOL_VERSION >> 16,
		       (POOL_VERSION >> 8) & 0xFF, POOL_VERSION & 0xFF);
		error = -EINVAL;
		goto fail;
	}

	size = add->p_num_spools * sizeof (subpool_t);

	pp->p_num_spools = add->p_num_spools;
	if (!(pp->p_spools = (subpool_t *) pmalloc(size))) {
		error = -ENOMEM;
		goto fail;
	}
	memcpy(pp->p_spools, add->p_spools, size);

	/* fill in the devices that constitute the subpools */

	for (i = 0; i < add->p_num_spools; i++) {
		sp = &(pp->p_spools[i]);
		size = sp->sp_devs * sizeof (device_t);

		if (!(sp->sp_devlist = (device_t *) pmalloc(size))) {
			error = -ENOMEM;
			goto fail;
		}

		memcpy(sp->sp_devlist, add->p_spools[i].sp_devlist, size);

		for (j = 0; j < sp->sp_devs; j++) {
			nd = &(sp->sp_devlist[j]);
			nd->d_rwlock = RW_LOCK_UNLOCKED;
			if (!pp->p_blksize) {	/* set the hardsector size if we don't know it */
				pp->p_blksize =
				    get_device_sector_size(nd->d_dev[0]);
			} else if (pp->p_blksize !=
				   get_device_sector_size(nd->d_dev[0])) {
				printk
				    ("Pool: ERROR :: All devices in a given pool"
				     " must have the same block size.\n");
				error = -EINVAL;
				goto fail;
			}
		}

		/* FIXME: assuming 512 byte blocks */
		/* stripe width should be a multiple of 4k */
		if (((sp->sp_stripe >> 3) << 3) != sp->sp_stripe) {
			printk("Pool: stripe width must be a mutliple of 4k\n");
			error = -EINVAL;
			goto fail;
		}
	}

	pp->p_pool_id = add->p_pool_id;
	pp->p_version = add->p_version;
	memset(pp->p_name, 0, POOL_NAME_SIZE);
	sprintf(pp->p_name, "%s", add->p_name);
	pp->p_capacity = add->p_capacity;
	pp->p_mptype = add->p_mptype;
	pp->p_mpstripe = add->p_mpstripe;
	pp->p_oblk = 0;

	return error;

      fail:
	free_pool(pp);
	return error;
}

/**
 * add_pool - make the kernel aware of a new pool
 * @add: ptr to struct containing new pool info
 *
 * This function takes the pool configuration infomation in 'add'
 * and constructs a pool_t structure from that.  It then finds an
 * appropriate minor number and places the structure in the 
 * global pool array 'pools'.
 *
 * Returns: Linux errno
 */

static int
add_pool(pool_t * add)
{
	pool_t *new_pp = NULL;
	int i = 0;
	int error = 0;

	/* see if this pool has already been added */

	for (i = 1; i < POOL_MAX_DEVICES; i++) {

		down_read(&(pools[i].p_rwsema));

		if (pools[i].p_flags & PF_FREE) {
			up_read(&(pools[i].p_rwsema));
			continue;
		}

		if (add->p_pool_id == pools[i].p_pool_id) {
			if (pools[i].p_flags & PF_VALID) {
				up_read(&(pools[i].p_rwsema));

				return 0;
			} else {
				up_read(&(pools[i].p_rwsema));
				error = -EBUSY;
				goto fail;
			}
		} else
			up_read(&(pools[i].p_rwsema));
	}

	/* the pool is not already configured so find space for it */

	if (add->p_minor &&
	    (add->p_minor <= (POOL_MAX_DEVICES / 2)) &&
	    (add->p_version == POOL_VERSION)) {
		down_write(&(pools[add->p_minor].p_rwsema));

		if ((pools[add->p_minor].p_flags & PF_FREE) &&
		    (pools[add->p_minor].p_flags & PF_VALID)) {

			pools[add->p_minor].p_flags = PF_INIT;	/* clear valid and free flags */
			ASSERT(!POOL_BUSY(&pools[add->p_minor]));

			up_write(&(pools[add->p_minor].p_rwsema));
			i = add->p_minor;
		} else {
			printk("Pool: Can not assemble pool \"%s\".\n"
			       "Pool: There is a minor number conflict with the pool \"%s\"\n",
			       add->p_name, pools[add->p_minor].p_name);
			up_write(&(pools[add->p_minor].p_rwsema));
			error = -EEXIST;
			goto fail;
		}
	} else {
		/* Minor number is not assigned, so pick one, starting at **
		 ** minor number (POOL_MAX_DEVICES/2)+1 .................. */
		for (i = ((POOL_MAX_DEVICES / 2) + 1); i < POOL_MAX_DEVICES;
		     i++) {

			down_write(&(pools[i].p_rwsema));

			if (pools[i].p_flags & PF_FREE
			    && pools[i].p_flags & PF_VALID) {

				pools[i].p_flags = PF_INIT;	/* clear valid and free flags */
				ASSERT(!POOL_BUSY(&pools[i]));

				up_write(&(pools[i].p_rwsema));
				break;
			} else {
				up_write(&(pools[i].p_rwsema));
			}
		}
	}

	if (i == POOL_MAX_DEVICES) {
		error = -EMFILE;
		goto fail;
	}

	new_pp = &pools[i];

	/* clear out all the old stats */
	memset(&new_pp->p_stats, 0, sizeof (pool_stat_t));

	error = gen_pool(add, new_pp);

	if (error != 0) {
		down_write(&(pools[i].p_rwsema));
		new_pp->p_flags |= PF_VALID | PF_FREE;
		up_write(&(pools[i].p_rwsema));
		goto fail;
	}

	error = open_devices(new_pp, 0);

	if (error > 0) {
		down_write(&(pools[i].p_rwsema));
		new_pp->p_flags |= PF_READONLY;
		up_write(&(pools[i].p_rwsema));
	} else if (error < 0) {
		down_write(&(pools[i].p_rwsema));
		new_pp->p_flags |= PF_VALID | PF_FREE;
		up_write(&(pools[i].p_rwsema));
		goto fail;
	}

	down_write(&(pools[i].p_rwsema));
	new_pp->p_flags |= PF_VALID | PF_LOWEROPEN;

	register_pool(new_pp, i);

	/* track pool_size in 1k blocks */
	pool_size[i] = new_pp->p_capacity >> 1;
	up_write(&(pools[i].p_rwsema));

      fail:
	return error;
}

/**
 * free_pool - free memory associated with the pool_t
 * @pool: ptr to the pool_t struct
 *
 * free memory associated with the pool_t, leaving you with an
 * emptied pool_t structure.  The pool_t struct itself is not freed
 *
 * Returns: 0
 */
int
free_pool(pool_t * pool)
{
	int i;
	subpool_t *sp;

	if (pool) {
		if (pool->p_spools) {
			for (i = 0; i < pool->p_num_spools; i++) {
				sp = &(pool->p_spools[i]);
				if (sp->sp_devlist) {
					pfree(sp->sp_devlist,
					      sp->sp_devs * sizeof (device_t));
				}
			}
			pfree(pool->p_spools,
			      pool->p_num_spools * sizeof (subpool_t));
			pool->p_num_spools = 0;
			pool->p_spools = NULL;
		}
		memset(pool->p_name, 0, POOL_NAME_SIZE);
	}

	return (0);
}

/**
 * remove_pool - removes pool from list of pools in the kernel
 * @minor: minor # of the pool to be removed
 *
 * This routine checks that a pool is not in use and then removes it from
 * the list of managed pools.
 *
 * Returns: Linux errno
 */
int
remove_pool(uint32_t major, uint32_t minor)
{
	pool_t *pp;
	int error;

	pp = &pools[minor];

	down_write(&(pp->p_rwsema));
	if (pp->p_flags & PF_FREE) {
		error = -EINVAL;
		goto fail;
	}

	if (!(pp->p_flags & PF_VALID) || POOL_BUSY(pp) || (pp->p_oblk == -1)) {
		error = -EBUSY;
		goto fail;
	}
	pp->p_oblk = -1;
	up_write(&(pp->p_rwsema));

	/* No need to worry about not holding the lock here because **
	 ** this pool can not be opened while p_oblk is -1.......... */
	fsync_dev(MKDEV(major, minor));

	down_write(&(pp->p_rwsema));
	if (pp->p_flags & PF_LOWEROPEN) {
		close_devices(pp);
	}

	free_pool(pp);

	pp->p_flags = PF_VALID | PF_FREE;

	unregister_pool(minor);
	up_write(&(pp->p_rwsema));

	return (0);

      fail:

	up_write(&(pp->p_rwsema));
	return (error);
}

/*
 * restore_pool_paths - update multipath routes
 * @pool: the pool to update
 *
 * This function will restore device paths that have
 * been marked as bad.
 *
 * NOTE:  This function must not be called in a interupt context.
 *
 * Returns: 0 on success, -1 on failure
 */
int
restore_pool_paths(pool_t * pool)
{
	int i;
	device_t *dev;
	pool_label_t *label;
	label_list_t *head, *tmp;

	down_read(&(pool->p_rwsema));

	if (!((pool->p_flags & PF_VALID) && (pool->p_flags & PF_LOWEROPEN)) ||
	    (pool->p_mpstripe < 0)) {
		up_read(&(pool->p_rwsema));
		return 0;
	}

	head = find_and_sort_labels();
	if (!head) {
		up_read(&(pool->p_rwsema));
		return -1;
	}
	while (head) {
		if (strcmp(pool->p_name, head->ll_label.pl_pool_name)) {
			tmp = head;
			head = head->next;
			pfree(tmp, sizeof (label_list_t));
			continue;
		}
		label = &(head->ll_label);

		if ((label->pl_sp_id >= pool->p_num_spools) ||
		    (label->pl_sp_devid >=
		     pool->p_spools[label->pl_sp_id].sp_devs)) {
			printk
			    ("Pool: Warning:  A Pool label was found on a device which is\n"
			     "Pool:           outside current pool bounds.  It is likely\n"
			     "Pool:           that new labels were written after the\n"
			     "Pool:           pools were assembled.\n");
			goto fail;
		}

		if (pool->p_pool_id != label->pl_pool_id) {
			printk
			    ("Pool: Warning:  Current active Pool IDs do not match the Pool\n"
			     "Pool:           IDs on the devices.  It is likely that new\n"
			     "Pool:           labels were written after the pools were\n"
			     "Pool:           assembled.\n");
			goto fail;
		}

		dev =
		    &(pool->p_spools[label->pl_sp_id].
		      sp_devlist[label->pl_sp_devid]);

		/* leave in possiblilty for adding future paths 
		   if(dev->d_num_paths == dev->d_num_good_paths){
		   continue;
		   }
		 */

		write_lock(&(dev->d_rwlock));
		for (i = 0; i < dev->d_num_good_paths; i++) {
			if (dev->d_dev[i] == head->ll_dev) {
				break;
			}
		}

		if ((i >= dev->d_num_good_paths) && (i < POOL_MAX_MULTIPATH)) {
			int error = 0;
			kdev_t tmp_dev;
			tmp_dev = head->ll_dev;
			error = bdopen(head->ll_dev);

			if ((error > 0) && !(pool->p_flags & PF_READONLY)){
				bdclose(head->ll_dev);
				printk("Pool: Unable to add read-only device (%d,%d)\n"
				       "      to a pool (%s) that is read/write.\n",
				       MAJOR(head->ll_dev),
				       MINOR(head->ll_dev),
				       pool->p_name);
			} else if (error) {
				printk
				    ("Pool: Open failed: error %d opening %d,%d\n",
				     error, MAJOR(head->ll_dev),
				     MINOR(head->ll_dev));
			} else {
				int j;
				for (j = i + 1; j < dev->d_num_paths; j++) {
					if (tmp_dev == dev->d_dev[j])
						dev->d_dev[j] = dev->d_dev[i];
				}
				dev->d_dev[i] = tmp_dev;
				dev->d_num_good_paths++;
				dev->d_num_paths =
				    (dev->d_num_paths >
				     dev->d_num_good_paths) ? dev->
				    d_num_paths : dev->d_num_good_paths;
				printk
				    ("Pool: Addition or restoration of device (%d, %d) in pool, %s\n",
				     MAJOR(head->ll_dev), MINOR(head->ll_dev),
				     pool->p_name);
			}
		}
		tmp = head;
		head = head->next;
		pfree(tmp, sizeof (label_list_t));
		write_unlock(&(dev->d_rwlock));
	}

	up_read(&(pool->p_rwsema));
	return 0;

      fail:
	up_read(&(pool->p_rwsema));
	while (head) {
		tmp = head;
		head = head->next;
		pfree(tmp, sizeof (label_list_t));
	}
	return -1;
}

/**
 * grow_pool - like assemble, only grows an existing pool online
 * @add: the pool to be grown
 *
 * Very similar to add_pool() in most respects.  However, a new pool
 * minor number is not assigned.
 * See the pool_t definition in pool_common.h for a description of
 * the locking protocol for the pool structure.
 * 	
 * Returns: Linux errno
 */

static int
grow_pool(pool_t * add)
{
	pool_t *new_pp = NULL;
	pool_t *cpy_pp = NULL;
	subpool_t *sp;
	device_t *devp = NULL;
	int i, j;
	int error = 0;

	/* see if this pool has already been added */

	for (i = 1; i < POOL_MAX_DEVICES; i++) {

		down_read(&(pools[i].p_rwsema));

		if (pools[i].p_flags & PF_FREE) {
			up_read(&(pools[i].p_rwsema));
			continue;
		}

		if (add->p_pool_id == pools[i].p_pool_id) {
			if (pools[i].p_flags & PF_VALID) {
				up_read(&(pools[i].p_rwsema));
				new_pp = &pools[i];
			} else {
				up_read(&(pools[i].p_rwsema));
				error = -EBUSY;
				goto fail;
			}
		} else {
			up_read(&(pools[i].p_rwsema));
		}
	}

	if (!new_pp) {
		error = -ENODEV;
		goto fail;
	}

	if (!(cpy_pp = (pool_t *) pmalloc(sizeof (pool_t)))) {
		error = -ENOMEM;
		goto fail;
	}

	memcpy(cpy_pp, new_pp, sizeof (pool_t));

	error = gen_pool(add, cpy_pp);

	if (error != 0) {
		pfree(cpy_pp, sizeof (pool_t));
		goto fail;
	}

	/* With this call we open the newly added devices */
	error = open_devices(cpy_pp, new_pp->p_num_spools);

	if (error > 0) {
		down_read(&(new_pp->p_rwsema));
		if(!(new_pp->p_flags & PF_READONLY)){
			up_read(&(new_pp->p_rwsema));
			free_pool(cpy_pp);
			pfree(cpy_pp, sizeof (pool_t));
			printk("Pool:  Growing %s would add a read-only device to a read/write pool.\n", new_pp->p_name);
			goto fail;
		}
		up_read(&(new_pp->p_rwsema));
	} else if (error < 0) {
		free_pool(cpy_pp);	/* frees everything except pool_t */
		pfree(cpy_pp, sizeof (pool_t));
		goto fail;
	}

	down_write(&(new_pp->p_rwsema));

	if (new_pp->p_capacity == cpy_pp->p_capacity) {
		up_write(&(new_pp->p_rwsema));
		free_pool(cpy_pp);	/* frees everything except pool_t */
		pfree(cpy_pp, sizeof (pool_t));
		error = -EIO;
		goto fail;
	}

	for (i = 0; i < new_pp->p_num_spools; i++) {
		sp = &(new_pp->p_spools[i]);
		for (j = 0; j < sp->sp_devs; j++) {
			devp = &(sp->sp_devlist[j]);
		}
	}

	free_pool(new_pp);
	new_pp->p_capacity = cpy_pp->p_capacity;
	memcpy(new_pp->p_name, cpy_pp->p_name, POOL_NAME_SIZE);
	new_pp->p_num_spools = cpy_pp->p_num_spools;
	new_pp->p_spools = cpy_pp->p_spools;
	pfree(cpy_pp, sizeof (pool_t));
	new_pp->p_flags |= PF_VALID | PF_LOWEROPEN;

	printk("Pool: New Pool Capacity:           %" PRId64 "\n",
	       new_pp->p_capacity);

	up_write(&(new_pp->p_rwsema));
      fail:
	return error;
}

/**
 * assemble_pools - assemble pools in the kernel
 * @poolname:
 *
 * Returns: Linux errno
 */
int
assemble_pools(char *poolname, int flags)
{
	int error = 0;
	int i, j, k;
	int grow = (flags == POOL_GROW) ? 1 : 0;
	pool_label_t *label = NULL;
	label_list_t *tmp = NULL;
	label_list_t *head = NULL;
	pool_t *np = NULL;
	subpool_t *ns = NULL;
	device_t *nd = NULL;
	uint64_t pblk_off = 0;
	uint64_t spblk_off = 0;

	head = find_and_sort_labels();

	while (head) {
		if (!strlen(poolname) ||
		    !strcmp(poolname, head->ll_label.pl_pool_name)) {
			if (!grow & pool_loaded(head->ll_label.pl_pool_name)) {
				tmp = head;
				head = head->next;
				pfree(tmp, sizeof (label_list_t));
				continue;
			}
			label = &(head->ll_label);
			if (!(np = pmalloc(sizeof (pool_t)))) {
				error = -ENOMEM;
				goto fail;
			}
			memset(np, 0, sizeof (pool_t));
			np->p_pool_id = label->pl_pool_id;
			np->p_version = label->pl_version;
			np->p_minor = label->pl_minor;
			np->p_num_spools = label->pl_subpools;
			np->p_mptype = 0;
			np->p_mpstripe = -1;

			memcpy(np->p_name, label->pl_pool_name, POOL_NAME_SIZE);

			if (!
			    (np->p_spools =
			     pmalloc(np->p_num_spools * sizeof (subpool_t)))) {
				error = -ENOMEM;
				goto fail;
			}

			memset(np->p_spools, 0,
			       np->p_num_spools * sizeof (subpool_t));

			for (; ((head) &&
				(!strcmp
				 (head->ll_label.pl_pool_name, np->p_name)));) {
				label = &(head->ll_label);
				if (label->pl_pool_id != np->p_pool_id) {
					printk
					    ("Pool: Multiple pools with the name \"%s\" found.\n",
					     label->pl_pool_name);
					tmp = head;
					head = head->next;
					pfree(tmp, sizeof (label_list_t));
					continue;
				}
				ns = &(np->p_spools[label->pl_sp_id]);
				ns->sp_devs = label->pl_sp_devs;
				ns->sp_type = label->pl_sp_type;
				ns->sp_stripe = label->pl_striping;
				if ((ns->sp_devlist == NULL) && (ns->sp_devs)) {
					if (!
					    (ns->sp_devlist =
					     pmalloc(ns->sp_devs *
						     sizeof (device_t)))) {
						error = -ENOMEM;
						goto fail;
					}
					memset(ns->sp_devlist, 0,
					       ns->sp_devs * sizeof (device_t));
				}

				if (label->pl_blocks) {
					nd = &(ns->
					       sp_devlist[label->pl_sp_devid]);
					if (nd->d_num_paths) {
						np->p_mptype = mp_type;
						np->p_mpstripe = mp_stripe;
						if (!np->p_mptype) {
							printk
							    ("Pool: Multipathing is not supported by this pool.\n");
							error = -ENOSYS;
							goto fail;
						}
						/*  ATTENTION -- potentially dangerous.  The way we test for **
						 **  multipathing could render forward compatibility efforts  **
						 **  useless !!!!!!!                                          
						 if(!test_for_false_multipathing(nd->d_dev[0], head->ll_dev)){
						 */
						if (nd->d_num_paths >=
						    POOL_MAX_MULTIPATH) {
							printk
							    ("Pool: More than %d paths to device found.\n"
							     "Pool: Ignoring (%d,%d).\n",
							     POOL_MAX_MULTIPATH,
							     MAJOR(head->
								   ll_dev),
							     MINOR(head->
								   ll_dev));
						} else {
							nd->d_dev[nd->
								  d_num_paths] =
							    head->ll_dev;
							nd->d_num_paths++;
							nd->d_num_good_paths++;
						}
						/*
						   } else {
						   error = -EIO;
						   goto fail;
						   }
						 */
					} else {
						nd->d_num_paths++;
						nd->d_num_good_paths++;
						nd->d_nblks = label->pl_blocks;
						nd->d_dev[0] = head->ll_dev;
					}
				}
				tmp = head;
				head = head->next;
				pfree(tmp, sizeof (label_list_t));
			}

			pblk_off = 0;	/* logical blk offsets - scope is entire pool */
			spblk_off = 0;	/* logical blk offsets - scope is one subpool */

			for (i = 0; i < np->p_num_spools; i++) {
				ns = &(np->p_spools[i]);

				ns->sp_start = pblk_off;
				spblk_off = 0;

				if (ns->sp_devs == 0) {	/* expected a label here */
					printk
					    ("Pool: Pool \"%s\", subpool %d has no devices.\n",
					     np->p_name, i);
					/* FIXME: I'm not sure that this is the best error code to return
					   but it's the best i can find ATM */
					error = -ENODEV;
					goto fail;
				}
				for (j = 0; j < ns->sp_devs; j++) {
					nd = &(ns->sp_devlist[j]);

					if (nd->d_nblks == 0) {	/* expected a label here */
						printk
						    ("Pool: Pool \"%s\", subpool %d, device %d has no blocks.\n",
						     np->p_name, i, j);
						/* FIXME: I'm not sure that this is the best error code to return
						   but it's the best i can find ATM */
						error = -ENOSPC;
						goto fail;
					}
					nd->d_start = spblk_off;
					spblk_off += nd->d_nblks;
				}
				ns->sp_len = spblk_off;	/* check on this brassow */
				pblk_off += spblk_off;
			}
			np->p_capacity = pblk_off;

			if (grow) {
				error = grow_pool(np);
				while (head) {
					tmp = head;
					head = head->next;
					pfree(tmp, sizeof (label_list_t));
				}
			} else {
				error = add_pool(np);
			}

			if (error != 0) {
				goto fail;
			}

			for (k = 0; k < np->p_num_spools; k++) {
				if (np->p_spools[k].sp_devlist) {
					pfree(np->p_spools[k].sp_devlist,
					      np->p_spools[k].sp_devs *
					      sizeof (device_t));
				}
			}
			pfree(np->p_spools,
			      np->p_num_spools * sizeof (subpool_t));
			pfree(np, sizeof (pool_t));
		} else {
			tmp = head;
			head = head->next;
			pfree(tmp, sizeof (label_list_t));
		}
	}

	/* this 'if' should be an 'ASSERT' */
	if (head) {
		printk("Pool: LABELS ARE LEFT OVER!!!!!!!!!\n");
	}
	return error;

      fail:
	while (head) {
		tmp = head;
		head = head->next;
		pfree(tmp, sizeof (label_list_t));
	}
	if (np) {
		if (np->p_spools) {
			for (i = 0; i < np->p_num_spools; i++) {
				ns = &(np->p_spools[i]);

				if (ns->sp_devlist) {
					pfree(ns->sp_devlist,
					      ns->sp_devs * sizeof (device_t));
				}
			}
			pfree(np->p_spools,
			      np->p_num_spools * sizeof (subpool_t));
		}
		pfree(np, sizeof (pool_t));
	}

	return error;
}

/**
 * count_pool - count the number of pools registered in the kernel
 * @count: ptr to int that will hold the number of pools in the kernel
 *
 * Returns: Linux errno
 */
int
count_pool(int *count)
{
	int i, j;
	pool_t *pp;

	for (i = 1, j = 0; i < POOL_MAX_DEVICES; i++) {

		pp = &pools[i];

		down_read(&(pp->p_rwsema));

		if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID)) {
			up_read(&(pp->p_rwsema));
			continue;
		}
		up_read(&(pp->p_rwsema));
		j++;
	}

	*count = j;

	return (0);
}

/**
 * list_pool - gather info on add pools registered in the kernel
 * @list - ptr where the info about the pools will be placed
 *
 * routine runs through the list of in-use pools and returns
 * to the caller.  It provides a mapping between pool devices
 * label devices and provides a versioning system by returning the
 * time at which the pool was last read in.  It also provides
 * about whether the device is in use or not.
 *
 * Returns: Linux errno
 */
int
list_pool(uint32_t major, pool_list_t * list)
{
	int i, j, k;
	pool_status_t *stats;
	pool_t *pp;

	if (list->pl_count > POOL_MAX_DEVICES)
		list->pl_count = POOL_MAX_DEVICES;

	if (!(stats = (pool_status_t *)
	      pmalloc(list->pl_count * sizeof (pool_status_t)))) {
		return -ENOMEM;
	}
	memset(stats, 0, list->pl_count * sizeof (pool_status_t));

	for (i = 1, j = 0; i < POOL_MAX_DEVICES; i++) {

		pp = &pools[i];

		down_read(&(pp->p_rwsema));

		if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID)) {
			up_read(&(pp->p_rwsema));
			continue;
		}

		stats[j].ps_pool_major = major;
		stats[j].ps_pool_minor = (uint32_t) i;
		stats[j].ps_pool_id = pp->p_pool_id;
		stats[j].ps_capacity = pp->p_capacity;
		stats[j].ps_num_spools = pp->p_num_spools;
		stats[j].ps_mptype = pp->p_mptype;
		/* FIXME: Assuming 512 byte blocks */
		stats[j].ps_mpstripe = pp->p_mpstripe / 2;

		for (k = 0; k < POOL_NAME_SIZE; k++)
			stats[j].ps_name[k] = pp->p_name[k];
		if (POOL_BUSY(pp))
			stats[j].ps_in_use = 1;
		else
			stats[j].ps_in_use = 0;

		up_read(&(pp->p_rwsema));

		if (++j >= list->pl_count) {
			break;
		}
	}

	list->pl_count = j;

	if (!(j > 0)) {
		pfree(stats, list->pl_count * sizeof (pool_status_t));
		return (0);
	}

	if (copy_to_user(list->pl_list, stats,
			 (int) (sizeof (pool_status_t) * j))) {
		pfree(stats, list->pl_count * sizeof (pool_status_t));
		return -EFAULT;
	}

	pfree(stats, list->pl_count * sizeof (pool_status_t));
	return (0);
}

/**
 * list_spool - generate list of subpools for a given pool
 * @pp: the specific pool from which to extract subpool info
 * @param: ptr into which the subpool info will be placed.
 *
 * Generate a list of subpools for a given pool and copy their sizes out
 * to user space.  Note: the space must be available in 'param' to put
 * the information in.
 *
 * Returns: Linux errno
 */
int
list_spool(pool_t * pp, pool_splist_t * param)
{
	pool_spstatus_t *list;
	subpool_t *spool;
	int i;
	int max;

	max = pp->p_num_spools > param->spl_count ? param->spl_count :
	    pp->p_num_spools;

	if (!(list = pmalloc(max * sizeof (pool_spstatus_t)))) {
		return -ENOMEM;
	}
	memset(list, 0, max * sizeof (pool_spstatus_t));

	for (i = 0; i < max; i++) {
		spool = &pp->p_spools[i];

		list[i].sps_start = spool->sp_start;
		list[i].sps_len = spool->sp_len;
		list[i].sps_stripe = spool->sp_stripe;
		list[i].sps_numdevs = spool->sp_devs;
		list[i].sps_type = spool->sp_type;
	}

	if (copy_to_user(param->spl_list, list,
			 (int) (sizeof (pool_spstatus_t) * max))) {
		pfree(list, sizeof (pool_spstatus_t) * max);
		return -EFAULT;
	}

	param->spl_count = max;

	pfree(list, sizeof (pool_spstatus_t) * max);
	return (0);
}

/**
 * list_pooldevs - generate device info for all devices in a given pool
 * @pp: the pool from which to extract the device information
 * @param: ptr to where the device info will be place
 *
 * Generate a list of devices for a given pool including various information
 * specific to each device.
 *
 * Returns: Linux errno
 */
int
list_pooldevs(pool_t * pp, pool_devlist_t * param)
{
	int error = 0;
	int i, j, k, n;
	subpool_t *spool;
	pool_devstatus_t *list;

	if (!(list = pmalloc(param->pdl_count * sizeof (pool_devstatus_t)))) {
		return -ENOMEM;
	}
	memset(list, 0, param->pdl_count * sizeof (pool_devstatus_t));

	for (i = 0, k = 0; (k < param->pdl_count) && (i < pp->p_num_spools);
	     i++) {
		spool = &pp->p_spools[i];
		for (j = 0; j < spool->sp_devs; j++, k++) {
			read_lock(&(spool->sp_devlist[j].d_rwlock));
			for (n = 0; n < spool->sp_devlist[j].d_num_paths; n++) {
				list[k].pds_dev[n] =
				    kdev_t_to_nr(spool->sp_devlist[j].d_dev[n]);
			}
			list[k].pds_num_paths =
			    spool->sp_devlist[j].d_num_paths;
			list[k].pds_num_good_paths =
			    spool->sp_devlist[j].d_num_good_paths;
			list[k].pds_blocks = spool->sp_devlist[j].d_nblks;
			read_unlock(&(spool->sp_devlist[j].d_rwlock));
		}
	}

	if (copy_to_user(param->pdl_list, list,
			 (int) (sizeof (pool_devstatus_t) *
				param->pdl_count))) {
		error = -EFAULT;
	}

	pfree(list, sizeof (pool_devstatus_t) * param->pdl_count);
	return error;
}

/**
 * pool_blkmap - maps logical storage view to a real storage device and blk
 * @pp: pool in which blk mapping is being requested
 * @psector: logical pool sector
 * @pbcount: logical pool blk count
 * @dev_sector: real device sector(s)
 * @dev_bcount: real device blk count(s)
 * @out_dev: real device(s)
 *
 * This is where the magic happens.  This function will take the logical
 * pool sector and blk count and convert it to specific devices' sectors
 * and blk counts.
 *
 * Returns: Linux errno
 */
int
pool_blkmap(pool_t * pp, uint64_t psector, uint32_t pbcount,
	    uint64_t * dev_sector, uint32_t * dev_bcount, device_t ** out_dev)
{
	subpool_t *sp = NULL;
	device_t *devp = NULL;
	int i, j;
	uint32_t current_spool = 0;
	uint32_t current_dev;
	uint64_t blk;
	uint64_t sp_offset;
	uint32_t devnum = 0;	/* # devices in selected spool */
	int error;

	/* lock is to prevent pool resizing during mapping */
	down_read(&(pp->p_rwsema));

	/* select subpool */

	for (j = current_spool; j < pp->p_num_spools; j++) {
		sp = &(pp->p_spools[j]);

		if ((psector >= sp->sp_start) &&
		    (psector < sp->sp_start + sp->sp_len)) {
			break;
		}
	}

	if (j == pp->p_num_spools) {
		printk("Pool: block not in any subpools at line %d!\n",
		       __LINE__);
		error = -EIO;
		goto fail;
	}

	current_spool = j;
	devnum = sp->sp_devs;

	/* Determine device within selected sub-pool */

	if (sp->sp_stripe) {
		/*
		   current_dev = (uint32_t)(((psector - sp->sp_start) 
		   / sp->sp_stripe) % devnum);
		 */
		uint64_t tmp = psector - sp->sp_start;

		do_div(tmp, sp->sp_stripe);
		tmp = do_mod(tmp, devnum);

		current_dev = (uint32_t) tmp;
	} else {
		for (i = 0; i < devnum; i++) {
			devp = &(sp->sp_devlist[i]);

			read_lock(&(devp->d_rwlock));
			if (((psector - sp->sp_start) >= devp->d_start) &&
			    ((psector - sp->sp_start) <
			     (devp->d_start + devp->d_nblks))) {
				read_unlock(&(devp->d_rwlock));
				break;
			} else {
				read_unlock(&(devp->d_rwlock));
			}
		}

		if (i == devnum) {
			printk
			    ("\nPool: block requested past end of subpool at line %d!\n",
			     __LINE__);
			error = -EIO;
			goto fail;
		}

		current_dev = i;
	}

	devp = &(sp->sp_devlist[current_dev]);

	/* Determine the block number on the device */
	if (sp->sp_stripe) {
		uint64_t tmp1, tmp2;
		tmp1 = tmp2 = (psector - sp->sp_start);

		do_div(tmp1, (sp->sp_stripe * sp->sp_devs));
		tmp2 = do_mod(tmp2, sp->sp_stripe);

		tmp1 *= sp->sp_stripe;
		blk = tmp1 + tmp2;
	} else {
		read_lock(&(devp->d_rwlock));
		blk = psector - sp->sp_start - devp->d_start;
		read_unlock(&(devp->d_rwlock));
	}

	/* account for partition header */

	blk += 16;

	*dev_sector = blk;
	*out_dev = devp;

	up_read(&(pp->p_rwsema));

	return (0);

      fail:

	up_read(&(pp->p_rwsema));

	return error;
}
