/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* label.c - Manipulates pool labels */

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/time.h>
#include <string.h>

#include "sisstring.h"
#include "sislogger.h"
#include "global.h"
#include "iddev.h"
#include "libpool.h"
#include "libpool_internal.h"
#include "NSP.cf"
#include "osi_user.h"
#include "osi_list.h"
#include "pool_ondisk.h"

#define POOL_MAGIC_BYTES 8


#ifdef DEBUG
/* FIXME: THis is only here for devel debugging */
static void _print_scan_labels(void *labels)
{
   osi_list_t   *tmp;
   struct label_list *l = NULL,
                *head = labels;
   pool_label_t *label;
   int ext2sb_check_result;
   char msg[128];
   struct sstring *emsg;

   emsg = ss_init(127);

   printf("  %-25s%25s\n", "Device", "Pool Label");
   printf("  ======                                  ==========\n");
   osi_list_foreach(tmp, &(head->ll_list)) {
      l = osi_list_entry(tmp, struct label_list, ll_list);
      label = &(l->ll_label);
      ext2sb_check_result = _ext2sb_check(l->ll_dev_path, emsg);
      if(ext2sb_check_result <= 0){
         printf("  %-25s%25s\n", l->ll_dev_path,
                (label->pl_magic == POOL_MAGIC) ?
                label->pl_pool_name :
                (ext2sb_check_result) ?
                "< none >" :
                "<- ext fs ->");
      }
      else {
         ss2cstring(msg, 128, emsg);
         printf("%s\n", msg);
      }
   }
   return;
}

static void _print_labels(struct label_list *head) {
   osi_list_t   *tmp;
   struct label_list *l;
   pool_label_t *label;

   osi_list_foreach(tmp, &(head->ll_list)) {
      l = osi_list_entry(tmp, struct label_list, ll_list);
      label = &l->ll_label;

      printf("Magic: %llu\n", label->pl_magic);
      printf("Version: %u\n", label->pl_version);
      printf("Pool name: %s\n", label->pl_pool_name);
      printf("Path: %s\n", l->ll_dev_path);

   }

   return;
}

#endif /* DEBUG */



/**
 * _generate_poolid - attempt to generate a unique ID for the pool
 * @id: the variable to put the ID into
 * @blocks: the number of blocks in the device (helps to seed the ID)
 * @emsg: storage for error message if error occurs
 *
 * generates a (hopefully) unique pool ID based on the number of blocks in the
 * pool and the time that the pool id was generated.
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: static
 */
static int _generate_poolid(uint64 *id, uint64 blocks, struct sstring *emsg)
{
   struct timeval mytime;
   if(gettimeofday(&mytime, NULL)) {
      ss_sprintf(emsg, "%s(%d): Unable to get time of day: %s",
                 __FILE__, __LINE__, strerror(errno));
      return errno;
   }
   *id = (blocks << 40) +
      ( (0x000fffff & mytime.tv_sec) << 20 ) +
      ( 0x000fffff & mytime.tv_usec ) ;
   return 0;
}

/**
 * _align_blkdevs - align all the block devices
 * @sp: subpool to align
 * @id: id to set (if not set)
 * @emsg: error message
 *
 * Description: Reserve 8k for the pool labels, align the remainder to 8kbytes.
 *
 */
static int _align_blkdevs(user_subpool_t *sp, uint64 *id, uint64 *min_devblk,
                            struct sstring *emsg)
{
   int i;
   int error = 0;
   user_device_t *current_dev = NULL;
   uint64         blocks = 0;
   uint64         sp_blocks = 0;

   for(i = 0; i < sp->num_devs; i++) {
      current_dev = &(sp->devs[i]);

      if((error = get_blk_size(current_dev->dev_path[0], &blocks, emsg)) != 0) {
         break;
      }

      debug_print("True size of %s: %"PRIu64"\n",
                  current_dev->dev_path[0], blocks);

      /* FIXME: Assuming 512 byte blocks */
      /* we need to round all block devices down to the nearest 8k */
      blocks = blocks & 0xFFFFFFF0;

      /*FIXME: assuming 512 byte blocks */
      /* Make sure 8k is reserved for the pool label */
      current_dev->blocks = blocks-16;
      sp_blocks += current_dev->blocks;
      debug_print("%"PRIu64" %"PRIu64"\n", sp_blocks, current_dev->blocks);
      if(*id == 0) {
         if((error = _generate_poolid(id, current_dev->blocks, emsg)) != 0) {
            break;
         }
      }
      debug_print("ID = %"PRIu64"\n", *id);

      if(sp->striping != 0) {
         if(*min_devblk > current_dev->blocks) {
            *min_devblk = current_dev->blocks;
         }
      }
      else {
         debug_print("logical size of %s: %"PRIu64"\n",
                     current_dev->dev_path[0], current_dev->blocks);
      }
   }

   return error;
}

/**
 * _trunc_strbds - truncate striped block devices
 * @sp: subpool to operate on
 * @min_devsize: size of the smallest block device (to be set)
 * @emsg: error message
 *
 * Description: Conform all block devices to the smallest device in
 * that subpool.
 *
 */
static int _trunc_strbds(user_subpool_t *sp, uint64 *min_devsize,
                         struct sstring *emsg)
{
   int i;
   int error = 0;
   user_device_t *current_dev = NULL;

   /* FIXME: assuming 512 byte blocks */
   /* min_devblcok must be larger than the stripe size */
   if(*min_devsize < sp->striping) {
      ss_sprintf(emsg, "subpool stripe size is larger"
                 " than the smallest device");
      error = EINVAL;
      goto end;
   }

   /* min_devblock must be divisable by the stripe size */
   *min_devsize = ((*min_devsize) / (sp->striping)) * (sp->striping);

   for(i = 0; i < sp->num_devs; i++) {
      current_dev = &(sp->devs[i]);
      current_dev->blocks = *min_devsize;
      debug_print("logical size of %s: %"PRIu64"\n",
                  current_dev->dev_path[0], current_dev->blocks);
   }

   /* FIXME: Should i print the number of bytes lost to the pool label
    * and 8k aliagnment? */
  end:
   return error;
}


/**
 * _build_labels - actually fill the label structure and add it to the list
 * @pool: pool that contains pool info
 * @sp: subpool we're working on
 * @head: head of label list
 * @id: pool id
 * @emsg: error message
 *
 */
static int _build_labels(user_pool_t *pool, user_subpool_t *sp,
                         struct label_list *head, uint64 *id,
                         struct sstring *emsg)
{
   int i;
   int error = 0;
   struct label_list *label;
   user_device_t *current_dev = NULL;

   for(i = 0; i < sp->num_devs; i++) {
      current_dev = &(sp->devs[i]);

      label = (struct label_list *) malloc(sizeof(struct label_list));
      if(label == NULL) {
         ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
         error = errno;

         /* Cleanup previously allocated labels */
         destroy_labels(head);
         break;
      }
      memset(label, 0, sizeof(struct label_list));

      label->ll_label.pl_magic = POOL_MAGIC;
      label->ll_label.pl_version = POOL_VERSION;
      label->ll_dev_path = strdup(current_dev->dev_path[0]);

      label->ll_label.pl_blocks = current_dev->blocks;
      label->ll_label.pl_pool_id = *id;

      strcpy(label->ll_label.pl_pool_name,pool->name);
      label->ll_label.pl_subpools = pool->num_sps;
      label->ll_label.pl_sp_id = sp->id;
      label->ll_label.pl_sp_devs = sp->num_devs;
      label->ll_label.pl_sp_devid = current_dev->devid;
      label->ll_label.pl_sp_type = sp->type;
      label->ll_label.pl_striping = sp->striping;
      label->ll_label.pl_minor = pool->minor;

      osi_list_add_prev(&(label->ll_list), &(head->ll_list));
   }

   return error;
}

/**
 * _make_labels - This makes labels out of a single userland pool structure
 * @pool: the userland pool structure (generally from config file)
 * @labels: the label list to fill
 * @emsg: storage for error message if error occurs
 *
 * Converts the information in the user_pool_t structure to struct label_list
 * format
 *
 * Returns: zero on success, non-zero on failure
 * FIXME: Return values need to be better defined
 *
 * Scope: static
 */
static int _make_labels(user_pool_t *pool, struct label_list **labels,
                        struct sstring *emsg)
{
   int            error = 0,
                  i;
   uint64         id = pool->id;
   struct label_list *head = NULL;
   user_subpool_t *current_sp = NULL;
   uint64         sp_blocks = 0;
   uint64         min_devblk = 0xFFFFFFFF;

   head = (struct label_list *) malloc(sizeof(struct label_list));

   if(!head) {
      ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
      error = errno;
      goto end;
   }

   osi_list_init(&(head->ll_list));

   for(i = 0; i < pool->num_sps; i++) {
      current_sp = &(pool->sp[i]);
      sp_blocks = 0;
      /* Preprocess the devices to get the sizes lined up properly */
      if((error = _align_blkdevs(current_sp, &id, &min_devblk, emsg)) != 0){
         free(head);
         break;
      }

      /* Now truncate all devices to the smallest device size if striping */
      if(current_sp->striping != 0) {
         if((error = _trunc_strbds(current_sp, &min_devblk, emsg)) != 0) {
            free(head);
            break;
         }
      }

      /* now actually build the list of labels */
      if((error = _build_labels(pool, current_sp, head, &id, emsg)) != 0) {
         free(head);
         break;
      }
   }

   if(error == 0)
      *labels = head;

end:
   return error;
}

/**
 * _get_disklabel - actually read in a disklabel and check what it is
 * @d: structure to fill in
 * @ll: label that has already been read in
 * @emsg: error message
 *
 * Returns: 0 on success, non-zero on failure
 *
 * Scope: static
 */
#define TYPE_SIZE  128
static int _get_disklabel(struct disklabels *d, struct label_list *ll)
{
   int fd = -1;
   char type[TYPE_SIZE] = {0};
   int error = 0;

   d->dev_path = strdup(ll->ll_dev_path);
   if(d->dev_path == NULL) {
      log_err("Unable to allocated device path");
      error = errno;
      goto end;
   }

   if(ll->ll_dev_path)
      log_debug("%s %"PRIu64, ll->ll_dev_path, ll->ll_label.pl_magic);

      /* FIXME: Should i check for actual errors here? */
   if(ll->ll_label.pl_magic != POOL_MAGIC) {
      if((fd = open(d->dev_path, O_RDONLY)) < 0){
         log_debug("Unable to open device \"%s\": %s",
                  d->dev_path, strerror(errno));
         /* If I can't open a device, just mark it as unknown and keep going */
         d->name = strdup(ERROR_LABEL);
         d->lt = unknown_label;
         return 0;
      }
      error = identify_device(fd, type, TYPE_SIZE);

      if(error == 0) {
         log_notice("%s - %s", d->dev_path, type, error);
         d->name = strdup(type);
         d->lt = known_label;
      }
      else if(error == 1) {
         log_notice("%s - %s", d->dev_path, UNKNOWN_LABEL);
         d->name = strdup(UNKNOWN_LABEL);
         d->lt = unknown_label;
         error = 0;
      }
      else if(error == -1) {
         log_warn("Error identifying device %s: %s",
                   d->dev_path, strerror(errno));
         d->name = strdup(ERROR_LABEL);
         d->lt = error_label;
         error = 0;
      }
   }
   else {
      log_notice("%s - %s", d->dev_path, ll->ll_label.pl_pool_name);
      d->name = strdup(ll->ll_label.pl_pool_name);
      d->lt = pool_label;
   }

   close(fd);

 end:
   return error;
}

/**
 * _get_disklabels - gets basic information about each disk
 * @labels: list of labels
 * @dl: disklabel array
 * @diskcount: number of disks in disklabel array
 * @emsg: storage for error message if error occurs
 *
 * reads through each label and checks to see what kind it is, filling in the
 * disklabel name field with the type, or pool name if it is a pool label
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: static
 */
static int _get_disklabels(struct label_list *labels, disklabels_t **dl,
                           int *diskcount, struct sstring *emsg)
{
   int error = 0;
   struct label_list *ll, *head = labels;
   osi_list_t *tmp = NULL;
   int count = 0;

   osi_list_foreach(tmp, &(head->ll_list)) {
      count++;
   }
   *diskcount = count;
   *dl = (disklabels_t *) malloc(sizeof(disklabels_t) * count);
   if(*dl == NULL) {
      ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
      error = errno;
      goto end;
   }
   count--;
   tmp = NULL;
   osi_list_foreach(tmp, &(head->ll_list)) {
      ll = osi_list_entry(tmp, struct label_list, ll_list);
      _get_disklabel(&(*dl)[count], ll);
      count--;
   }

  end:
   return error;
}

/**
 * _strip_nonpool - strip all non-pool disklabels out of label list
 * @labels: list of disk labels
 *
 * Pull all non-pool disklabels out of the label list
 *
 * Scope: static
 */
static void _strip_nonpool(struct label_list **config)
{
   struct label_list *ll, *head = *config;
   osi_list_t *tmp;
   osi_list_foreach(tmp, &(head->ll_list)) {
      ll = osi_list_entry(tmp, struct label_list, ll_list);
      if(ll->ll_label.pl_magic != POOL_MAGIC)
         osi_list_del(&(ll->ll_list));

   }
   return;
}

/**
 * _sort_pool_labels - sorts a list of labels by pool name
 * @l: the list of labels
 * @emsg: storage for error message if error occurs
 *
 * Rebuilds the list of labels in order sorted by pool name.  There must not
 * be any non-pool labels in the list when it is run.
 *
 * Returns: zero on success, non-zero on failure.  Currently if this fails,
 *          the original list is completely mangled
 *
 * Scope: static
 */
static int _sort_pool_labels(struct label_list **l, struct sstring *emsg)
{
   int error = 0;
   osi_list_t *tmp = NULL, *small = NULL;
   struct label_list *head, *smallist, *tmplist, *sorted;

   sorted = (struct label_list *) malloc(sizeof(struct label_list));
   if(sorted == NULL) {
      ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
      error = errno;
      goto end;
   }

   osi_list_init(&(sorted->ll_list));
   head = *l;

   while(!osi_list_empty(&(head->ll_list))) {
      osi_list_foreach(tmp, &(head->ll_list)) {
         if(!small)
            small = tmp;
         tmplist = osi_list_entry(tmp, struct label_list, ll_list);
         smallist = osi_list_entry(small, struct label_list, ll_list);
         if(tmplist->ll_label.pl_magic != POOL_MAGIC) {
            ss_sprintf(emsg, "%s(%d): Non-pool label detected on %s!",
                       __FILE__, __LINE__, tmplist->ll_dev_path);
            error = 1;
            goto fail;
         }
         if(strcmp(smallist->ll_label.pl_pool_name,
                  tmplist->ll_label.pl_pool_name) < 0) {
            small = tmp;
         }
      }
      osi_list_del(small);
      smallist = osi_list_entry(small, struct label_list, ll_list);
      osi_list_add(small, &(sorted->ll_list));
/*      fprintf(stderr, "%d\n", _list_count(&(head->ll_list))); */
      small = NULL;
   }
   free(*l);
   *l = sorted;
   (*l)->ll_sorted = 1;
   return error;

fail:
   debug_print("Cleaning up after pool label sort failure.");
   tmp = NULL;
   /* Add the removed entries back into the original list */
   osi_list_foreach_safe(tmp, &(sorted->ll_list), small) {
      osi_list_del(tmp);
      osi_list_add(tmp, &(head->ll_list));
   }
   /* remove the head of the other list */
   free(sorted);
  end:
   return error;
}

/**
 * _count_pools - counts the number of pools in the label list
 * @l: label list
 * @pool_count: variable to store the pool count in
 *
 * Returns: zero on success
 *
 * Scope: static
 */
static void _count_pools(struct label_list *l, int *pool_count)
{
   char poolname[POOL_NAME_SIZE] = "";
   osi_list_t *tmp = NULL;
   struct label_list *label;

   *pool_count = 0;

   osi_list_foreach(tmp, &(l->ll_list)) {
      label = osi_list_entry(tmp, struct label_list, ll_list);
      if(label->ll_label.pl_magic != POOL_MAGIC)
         continue;
      if(strcmp(poolname, label->ll_label.pl_pool_name)) {
         (*pool_count)++;
         strcpy(poolname, label->ll_label.pl_pool_name);
         /*fprintf(stderr, "%d %s\n", *pool_count, poolname);*/
      }
   }

   return;
}

/**
 * _get_pools - get the info on detected pools ondisk
 * @config:
 * @p:
 * @pool_count:
 * @emsg: storage for error message if error occurs
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: static
 */
static int _get_pools(struct label_list *config, user_pool_t **p,
                      int pool_count, struct sstring *emsg)
{
   int error = 0;
   char pl_name[POOL_NAME_SIZE] = "";
   int pool_num = -1;
   uint32 subpool_num = 0, *dev_num = NULL;
   int subpool_set = 0;
   struct label_list *ll, *head = NULL;
   user_pool_t *pool = NULL;
   user_subpool_t *sp = NULL;
   user_device_t *dev = NULL;
   osi_list_t *tmp;

   pool = *p = (user_pool_t *) malloc(sizeof(user_pool_t) * pool_count);
   if(pool == NULL) {
      ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__, strerror(errno));
      error = errno;
      goto end;
   }
   memset(pool, 0, sizeof(user_pool_t) * pool_count);

   head = config;

   osi_list_foreach(tmp, &(head->ll_list)) {
      ll = osi_list_entry(tmp, struct label_list, ll_list);

      /* Pools are sorted by name */
      if(strcmp(ll->ll_label.pl_pool_name, pl_name)) {
         strcpy(pl_name, ll->ll_label.pl_pool_name);
         pool_num++;
         strcpy(pool[pool_num].name, pl_name);
         pool[pool_num].id = ll->ll_label.pl_pool_id;
         pool[pool_num].minor = ll->ll_label.pl_minor;
         pool[pool_num].num_sps = ll->ll_label.pl_subpools;
         pool[pool_num].sp = (user_subpool_t*) malloc(sizeof(user_subpool_t) *
               pool[pool_num].num_sps);
         if(pool[pool_num].sp == NULL) {
            ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                       strerror(errno));
            error = errno;

            destroy_pl(pool, pool_num);

            goto end;
         }
         memset(pool[pool_num].sp, 0, sizeof(user_subpool_t) *
               pool[pool_num].num_sps);
         subpool_num = 0;
         subpool_set = 0;
      }
      if(pool[pool_num].id != ll->ll_label.pl_pool_id) {
         log_err("Warning: %s has a label with same pool name but different"
                 " id from previously discovered disk for %s...ignoring",
                 ll->ll_dev_path, pool[pool_num].name);
         continue;
      }
      /* subpools are not sorted within the pool */
      if(!subpool_set || subpool_num != ll->ll_label.pl_sp_id) {
         if(ll->ll_label.pl_sp_id >= pool[pool_num].num_sps) {
            log_warn("Warning: Subpool %d out of bounds for pool %s",
                     ll->ll_label.pl_sp_id, pool[pool_num].name);
            /* If we detect an out-of-bounds subpool, keep processing but
               ignore this pool */
            continue;
         }
         subpool_set = 1;
         subpool_num = ll->ll_label.pl_sp_id;
         sp = &(pool[pool_num].sp[subpool_num]);
         sp->id = ll->ll_label.pl_sp_id;
         sp->striping = ll->ll_label.pl_striping;
         sp->num_devs = ll->ll_label.pl_sp_devs;
         sp->type  = ll->ll_label.pl_sp_type;
         /* Make sure the device structures haven't already been allocated */
         if(!sp->devs) {
            sp->devs = (user_device_t *) malloc(sizeof(user_device_t) *
                  sp->num_devs);
            if(sp->devs == NULL) {
               ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                          strerror(errno));
               error = errno;

               destroy_pl(pool, pool_num);
               goto end;
            }
            memset(sp->devs, 0, sizeof(user_device_t) * sp->num_devs);
         }
      }
      /* FIXME: should probably hold *all* the paths - currently only shows
         first path found */
      dev = &(sp->devs[ll->ll_label.pl_sp_devid]);
      if(dev->initialized == 0) {
         dev->initialized = 1;
         dev->sp_id = subpool_num;
         dev->devid = ll->ll_label.pl_sp_devid;
         dev->blocks = ll->ll_label.pl_blocks;
         dev->dev_path[0] = strdup(ll->ll_dev_path);
         if(dev->dev_path == NULL) {
            ss_sprintf(emsg, "%s(%d): %s", __FILE__, __LINE__,
                       strerror(errno));
            error = errno;
            destroy_pl(pool, pool_num);
            goto end;
         }
      }
   }

end:
   if(dev_num)
      free(dev_num);
   return error;
}

/**
 * _check_for_prev_write - returns an error if it detects a previous write
 * @label: label to check against
 *
 * Description: Read in label from disk and verify it doesn't contain
 * a pool label already - prevents user from writing multiple labels
 * to the same disk that has multiple paths
 */
static int _check_for_prev_write(struct label_list *label)
{
   int error = 0;
   struct label_list *dl = NULL;
   struct disklabels d = {0};
   pool_label_t *cur_disk = NULL, *new_disk = NULL;

   if(get_pool_label(label->ll_dev_path, &dl)) {
      stack;
      error = STD_ERROR;
      goto end;
   }
   _get_disklabel(&d, dl);

   cur_disk = &dl->ll_label;
   new_disk = &label->ll_label;

   /* the only way this series of tests should succeed is if the
    * caller passed in a label that is in the same pool, but should be
    * on a different disk, than the label that is on the disk it
    * should be written to. */
   if(d.lt == pool_label) {
      /* if pool id is the same we have the same pool */
      if(new_disk->pl_pool_id == cur_disk->pl_pool_id) {
         /* if subpool id or device id is different, we wrote to this
            disk already */
         if((new_disk->pl_sp_id != cur_disk->pl_sp_id) ||
            (new_disk->pl_sp_devid != cur_disk->pl_sp_devid)) {
            log_err("Error: %s already contains label for pool \"%s\""
                    "\nMake sure there are no redudant paths to a"
                    " device in configuration file.",
                    label->ll_dev_path, d.name);
            error = STD_ERROR;
            goto end;
         }
      }
   }

  end:
   if(dl) {
      if(dl->ll_dev_path)
         free(dl->ll_dev_path);
      free(dl);
   }

   return error;
}


/**
 * get_pool_label
 *
 */
int get_pool_label(char *path, struct label_list **l)
{
   int fdesc = -1;
   char buffer[SIZEOF_POOL_LABEL];
   struct label_list *label = NULL;
   int error = 0;

   label = (struct label_list *) malloc(sizeof(struct label_list));
   if(!label) {
      log_err("Out of memory allocating label_list");
      error = ENOMEM;
      goto end;
   }
   label->ll_dev_path = (char*) malloc(sizeof(char) * (strlen(path)+1));
   if(!label->ll_dev_path) {
      free(label);
      label=NULL;
      log_err("Out of memory allocating path for label_list");
      error = ENOMEM;
      goto end;
   }

   if((fdesc = open(path, O_RDONLY)) < 0){
      log_warn("Unable to open device \"%s\": %s", path, strerror(errno));
      error = 1;
      goto end;
   }
   if(read(fdesc, buffer, sizeof(pool_label_t)) < sizeof(pool_label_t)){
      close (fdesc);
      log_warn("Unable to read device \"%s\": %s", path, strerror(errno));
      error = 1;
      goto end;
   }

   pool_label_in(&(label->ll_label), buffer);
   close(fdesc);

  end:
   if(label)
      if(label->ll_dev_path)
         strcpy(label->ll_dev_path, path);
   *l = label;
   return error;
}


/**
 * write_labels - take a set of pool structures and write them to disk
 * @p: The pool info to write to disk
 * @emsg: storage for error message if error occurs
 *
 * FIXME: Do i want to support an array of user_pool_t structures?
 * TODO: Check for ext2, etc labels on disks before writing to them
 * FIXME: need to have some way to check that the pool labels are valid before
 *        writing them to disk
 *
 * Returns: zero on success, non-zero on error
 *
 * Scope: external
 */
int write_labels(user_pool_t *p, struct sstring *emsg)
{
   char         buffer[SIZEOF_POOL_LABEL];
   int          fd,
                error = 0;
   osi_list_t   *tmp;
   struct label_list *label = NULL,
                *head = NULL;

   if((error = _make_labels(p, &head, emsg)) != 0) {
      goto end;
   }
#ifdef DEBUG
   _print_labels(head);
#endif /* DEBUG */

   osi_list_foreach(tmp, &(head->ll_list)) {
      label = osi_list_entry(tmp, struct label_list, ll_list);
      memset(buffer, 0, SIZEOF_POOL_LABEL);
      pool_label_out(&label->ll_label, buffer);

      if((error = _check_for_prev_write(label))) {
         stack;
         break;
      }

      debug_print("Opening %s\n", label->ll_dev_path);
      if ((fd = open(label->ll_dev_path, O_WRONLY)) < 0) {
         ss_sprintf(emsg, "%s(%d): Failed to open %s.\n", __FILE__, __LINE__,
                    label->ll_dev_path);
         error = -errno;
         break;
      }

      if (write(fd, buffer, SIZEOF_POOL_LABEL) < SIZEOF_POOL_LABEL) {
         ss_sprintf(emsg, "%s(%d): Failed to write to %s\n",
                     __FILE__, __LINE__, label->ll_dev_path);
         error = -errno;
         fsync(fd);
         close(fd);
         break;
      }

      fsync(fd);
      close(fd);
   }

   destroy_labels(head);

   free(head);

end:
   return error;
}

/**
 * erase_label - erase a given pool label
 * @p: pool information
 * @emsg: storage for error message if error occurs
 *
 * Erase the pool label described in p - this really just removes the
 * first 8 bytes of the pool label to wipe out the magic field.  Without
 * the magic field, pool will not recognize the rest of the label.  This is to
 * prevent the following scenario -
 *
 *  o user writes pool label on /dev/sda
 *
 *  o user creates partition table on /dev/sda
 *
 *  o user erases pool label from /dev/sda
 *
 * If the pool erase wiped the entire SIZEOF_POOL_LABEL section, it would also
 * overwrite the partition table, which is not in the users best interest.
 *
 * FIXME: watch for invalid user_pool_t structures...
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int erase_label(user_pool_t *p, struct sstring *emsg)
{
   char  buffer[POOL_MAGIC_BYTES];
   int   fd,
         i,
         j,
         error = 0,
         oerr = 0,
         werr = 0,
         assembled = 0;
   user_device_t *dev;

   if((error = set_pool_assembled(p, &assembled, emsg)) != 0) {
      goto end;
   }

   if(assembled != 0) {
      ss_sprintf(emsg, "Pool \"%s\" is assembled - cannot erase labels", p->name);
      error = EBUSY;
      goto end;
   }

   memset(buffer, 0, POOL_MAGIC_BYTES);

   for(i = 0; i < p->num_sps; i++) {
      for(j = 0; j < p->sp[i].num_devs; j++) {
         dev = &(p->sp[i].devs[j]);
         if(dev->dev_path == NULL || dev->dev_path[0] == NULL) {
            continue;
         }
         if((fd = open(dev->dev_path[0], O_WRONLY)) < 0) {
            debug_print("Failed to open %s.\n", dev->dev_path[0]);
            oerr++;
            werr++;
            error = STD_ERROR;
            /* Don't bother writing to it, but try to get the rest... */
            continue;
         }
         if(write(fd, buffer, POOL_MAGIC_BYTES) < POOL_MAGIC_BYTES) {
            debug_print("Failed to wipe magic completely for %s\n",
                        dev->dev_path[0]);
            werr++;
            error = STD_ERROR;
         }
         fsync(fd);
         close(fd);
      }
   }

   if(error != 0) {
      ss_sprintf(emsg, "%s(%d): Failed to open %d devices, failed to erase %d"
                 " devices.", __FILE__, __LINE__, oerr, werr);
   }

end:
   return error;
}

/**
 * print_pool_labels - scan the disks and print out the pool ondisk structures
 * @name: name of pool to print - if NULL, print all detected pool labels
 * @emsg: storage for error message if error occurs
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int print_pool_labels(char *name, struct sstring *emsg)
{
   int error = 0;
   struct label_list *labels = NULL;
   osi_list_t *tmp;
   struct label_list *ll, *head;
   struct pool_label_s *p = NULL;

   if((error = scan_labels(&labels, emsg)))
      goto fail;

   _strip_nonpool(&labels);

   if((error = _sort_pool_labels(&labels, emsg)))
      goto fail;

   head = labels;

   osi_list_foreach(tmp, &(head->ll_list)) {
      ll = osi_list_entry(tmp,struct label_list, ll_list);
      p = &(ll->ll_label);
      if(name != NULL) {
         if(strcmp(name, p->pl_pool_name) != 0) {
            continue;
         }
      }
      if(p->pl_magic == POOL_MAGIC) {
         log_info("Pool = %s", p->pl_pool_name);
         log_info("  ID = %"PRIx64, p->pl_pool_id);
         log_info("  Magic = %"PRIx64, p->pl_magic);
         log_info("  Version = %x", p->pl_version);
         log_info("  Subpools = %x", p->pl_subpools);
         log_info("  Sp_id = %x", p->pl_sp_id);
         log_info("  Sp_type = %x", p->pl_sp_type);
         log_info("  Blocks = %"PRIx64, p->pl_blocks);
         log_info("  Striping = %x", p->pl_striping);
         log_info("  Sp_dmep_devs = %x", p->pl_sp_dmepdevs);
         log_info("  Sp_dmepid = %x", p->pl_sp_dmepid);
         log_info("  SP_weight = %x", p->pl_sp_weight);
         log_info("  Pool_minor = %x", p->pl_minor);
         log_info("  Sp_devs = %x", p->pl_sp_devs);
         log_info("  Sp_devid = %x", p->pl_sp_devid);
         log_info("  Device_name = %s", ll->ll_dev_path);
         log_info("  Device_Major = %x", ll->ll_major);
         log_info("  Device_Minor = %x", ll->ll_minor);
      }
   }

  fail:
   return error;

}

/**
 * scan_disks - scan the disks and gather info on what's on them
 * @di: diskinfo structure to fill
 * @emsg: storage for error message if error occurs
 *
 * Fill the diskinfo structure based on information pulled from the disks
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int scan_disks(diskinfo_t *di, struct sstring *emsg)
{
   int error = 0;
   int diskcount = 0, poolcount = 0;
   user_pool_t *p = NULL;
   disklabels_t *dl = NULL;
   struct label_list *labels = NULL;

   if((error = scan_labels(&labels, emsg)))
      goto fail;

   if((error = _get_disklabels(labels, &dl, &diskcount, emsg)))
      goto fail;

   di->dl = dl;
   di->dl_count = diskcount;

   _strip_nonpool(&labels);

   if((error = _sort_pool_labels(&labels, emsg)))
      goto fail;

   _count_pools(labels, &poolcount);

#ifdef DEBUG
   _print_scan_labels(labels);
#endif /* DEBUG */

   if((error = _get_pools(labels, &p, poolcount, emsg)))
      goto fail;

   di->pl = p;
   di->pl_count = poolcount;

   destroy_labels(labels);
   free(labels);

fail:

   return error;
}

/**
 * rename_pool - rename a pool
 * @oldpool: pool to rename
 * @newpool: name to call it
 * @emsg: storage for error message if error occurs
 *
 * Renames the pool specified by oldpool to newpool
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int rename_pool(char *oldpool, char *newpool, struct sstring *emsg)
{
   int error = 0,
       i;
   diskinfo_t di = {0};
   user_pool_t *p = NULL;

   if(check_poolname(newpool) != 0) {
      ss_sprintf(emsg, "Pool name must be smaller than %d characters",
                 POOL_NAME_SIZE);
      error = STD_ERROR;
      goto end;
   }

   /* get the old pool label */
   error = scan_disks(&di, emsg);
   if(!error) {
      for(i = 0; i < di.pl_count; i++) {
         if(!strcmp(oldpool, di.pl[i].name)) {
            p = &(di.pl[i]);
         }
      }
      if(p != NULL) {
         if(strlen(newpool) == 0) {
            ss_sprintf(emsg, "%s(%d): New name is of zero length",
                       __FILE__, __LINE__, newpool);
            error = STD_ERROR;
            goto end;
         }
         if((strchr(newpool, ' ') != NULL) ||
            (strchr(newpool, '\t') != NULL)) {
            ss_sprintf(emsg, "%s(%d): New name '%s' has whitespace in it",
                       __FILE__, __LINE__, newpool);
            error = STD_ERROR;
            goto end;
         }
         if((strchr(newpool, '/') != NULL)) {
            ss_sprintf(emsg, "%s(%d): New name '%s' has '/' in it",
                       __FILE__, __LINE__, newpool);
            error = STD_ERROR;
            goto end;
         }
         /* change the name in the old pool label */
         memset(p->name, 0, POOL_NAME_SIZE);
         strcpy(p->name, newpool);
      }
      else {
         ss_sprintf(emsg, "%s(%d): Unable to find pool \"%s\"",
                    __FILE__, __LINE__, oldpool);
         error = STD_ERROR;
      }
   }
   if(!error) {
      /* write the new pool label back */
      error = write_labels(p, emsg);
   }
  end:
   destroy_di(&di);

   return error;
}

/**
 * reminor_pool - reminor a pool
 * @p: pool to rename
 * @num: minor number to set it to
 * @emsg: storage for error message if error occurs
 *
 * Changes a pool's minor number to num
 *
 * Returns: zero on success, non-zero on failure
 *
 * Scope: external
 */
int reminor_pool(char *poolname, uint16 num, struct sstring *emsg)
{
   int error = 0,
       i;
   diskinfo_t di;

   user_pool_t *p = NULL;

   if(check_minornum(num) != 0) {
      ss_sprintf(emsg, "Minor number %d out of bounds", num);
      error = STD_ERROR;
      goto end;
   }

   /* get the old pool label */
   error = scan_disks(&di, emsg);
   if(!error) {
      for(i = 0; i < di.pl_count; i++) {
         if(!strcmp(poolname, di.pl[i].name)) {
            p = &(di.pl[i]);
         }
      }
      if(p != NULL) {
         /* change the minor number in the old pool label */
         p->minor = num;
      }
      else {
         ss_sprintf(emsg, "%s(%d): Unable to find pool \"%s\"",
                    __FILE__, __LINE__, poolname);
         error = STD_ERROR;
      }
   }
   if(!error) {
      /* write the new pool label back */
      error = write_labels(p, emsg);
   }

   destroy_di(&di);

  end:
   return error;
}


/*
 * vim: et ts=3 sw=3 noai cin :
 */

