/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* misc.c -  Misc functions for libpool */

#include "sisstring.h"
#include "libpool.h"
#include "libpool_internal.h"
#define WANT_SPTYPE_NAMES
#include "pool_sptypes.h"

static int _flags = 0;

/**
 * check_flag - check if given flag is set
 * @flag - flag to check
 *
 * Checks if flag is set or not set
 *
 * Returns: positive if flag is set, zero if flag is not set
 *
 * Scope: external
 */
inline int check_flag(int flag)
{
  return _flags & flag;
}

/**
 * add_flag - adds a flag to the library's flag structure
 * @flag - flag to add
 *
 * Add the given flag to the library's flag structure for later checking - this
 * can then be accessed by the library functions themselves or the program
 * that is accessing the library.
 *
 * Returns: zero on success, non-zero on failure
 */
inline int add_flag(int flag)
{
   if(_flags |= flag)
      return 0;
   else
      return 1;
}

/**
 * del_flags - remove the given flag from the library's flag structure
 * @flag: flag to remove
 *
 * Remove the given flag from the library's flag structure - remember this
 * affects all components of the library and calling program that are using the
 * flag for this instance of the library.
 *
 * Returns: zero ATM
 * FIXME: is there any other valid return value?
 */
inline int del_flag(int flag)
{
   if(check_flag(flag))
        _flags &= ~flag;

   return 0;
}

/**
 * get_sptype_num - convert a string to type number
 * @type: name of type
 * @num: type number
 *
 * Returns: zero on match, non-zero on no match
 */
int get_sptype_num(char *type, uint32 *num)
{
   int i;
   int error = STD_ERROR;
   for(i = 0; sptype_names[i].name[0]; i++){
      if(!strcmp(type, sptype_names[i].name)) {
         *num = sptype_names[i].label;
         error = 0;
         break;
      }
   }
   return error;
}


/**
 * get_sptype_name - convert type number to string equivalent
 * @type: type number (defined in pool_sptypes.h)
 * @name: store the name in here
 *
 * Returns: zero on match, non-zero on no match
 */
int get_sptype_name(uint32 type, char **name){
  int 	x;
  char	buffer[POOL_LINE_MAX];
  int    error = STD_ERROR;
  for (x = 0; sptype_names[x].name[0]; x++) {
    if (type == sptype_names[x].label){
      snprintf(buffer, POOL_LINE_MAX, "%s", sptype_names[x].name);
      *name = (char *)malloc(strlen(buffer));
      strcpy(*name, buffer);
      error = 0;
      break;
    }
  }
  return error;
}

/**
 * destroy_sptype_name - clean up the name allocated in get_sptype_name
 * @name: Name to be destroyed
 *
 * Frees the memory allocated in get_sptype_name
 *
 * Scope: external
 *
 */
void destroy_sptype_name(char **name)
{
   if(*name)
      free(*name);
   return;
}

/**
 * destroy_pl -
 * @pool:
 * @pool_count:
 *
 */
void destroy_pl(user_pool_t *pool, int pool_count)
{
   int i, j, k, l;

   if(pool == NULL)
      return;

   for(i = 0; i < pool_count; i++) {
      if(pool[i].sp != NULL) {
         for(j = 0; j < pool[i].num_sps; j++) {
            if(pool[i].sp[j].devs != NULL) {
               for(k = 0; k < pool[i].sp[j].num_devs; k++) {
                  if(pool[i].sp[j].devs[k].dev_path != NULL) {
                     for(l = 0; l < pool[i].sp[j].devs[k].path_count; l++) {
                        if(pool[i].sp[j].devs[k].dev_path[l]) {
                           free(pool[i].sp[j].devs[k].dev_path[l]);
                        }
                     }
                  }
               }
               free(pool[i].sp[j].devs);
            }
         }
         free(pool[i].sp);
      }
   }
   free(pool);

}

/**
 * destroy_di - Destroys the contents within di structure, but not the di itself
 * @di: diskinfo_t structure to destroy
 *
 * Frees the allocated members of the diskinfo_t structure.  Zeros out the
 * counters for the arrays.  The diskinfo_t structure itself is not freed.
 *
 */
void destroy_di(diskinfo_t *di)
{
   int i;

   if(di != NULL) {
      if(di->dl_count) {
         for(i = 0; i < di->dl_count; i++) {
            free(di->dl[i].name);
            free(di->dl[i].dev_path);
         }
         free(di->dl);
         di->dl = NULL;
      }
      if(di->pl_count) {
         destroy_pl(di->pl, di->pl_count);
         di->pl = NULL;
      }
   }
   di->dl_count = 0;
   di->pl_count = 0;
   return;
}

/**
 * destroy_labels - Free the label list
 * @l: list of labels
 *
 * Iterates through the label list and removes each item
 *
 * Scope: Internal to library
 */
void destroy_labels(struct label_list *l)
{
   osi_list_t   *tmp = NULL,
                *safety = NULL;
   struct label_list *label = NULL,
                *head = l;

   /* FIXME: Verify that all memory allocated is freed by this */
   osi_list_foreach_safe(tmp, &(head->ll_list), safety) {
      label = osi_list_entry(tmp, struct label_list, ll_list);
      if(label->ll_dev_path)
         free(label->ll_dev_path);
      osi_list_del(tmp);
      free(label);
   }
   return;
}

/*
 * vim: et ts=3 sw=3 noai cin :
 */

