/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

/* pool_assemble.c - Assembles and disassembles pools. */

#include <sys/stat.h>
#include <ctype.h>

#include "sislogger.h"
#include "siscli.h"
#include "sisstring.h"
#include "libpool.h"
#include "osi_user.h"
#include "pool_release.cf"
#include "copyright.cf"

#define STRING_SIZE 512
#define MP_NOT_CHANGED -1042
static int _is_pool_assembled(char *name, struct sstring *emsg);
static int _assemble_pool(char *poolname, struct sstring *emsg);
static int _disassemble_pool(char *poolname, struct sstring *emsg);

static void _usage(char *name) {

   printf("Usage:\n");
   printf("\n");
   printf("%s [options] [pool(s)]\n", name);
   printf("\n");
   printf("Options:\n");
   printf("  -a               Assemble specified pool or all pools if none "
          "specified.\n");
   printf("  -D               Enable debugging output.\n");
   printf("  -h               This usage information.\n");
   printf("  -q               Enable quiet output.\n");
   printf("  -r               Remove specified pool or all pools if none "
          "specified.\n");
   printf("  -m <name/num>    Tune multipathing -\n"
          "                   <name>: \"none\" or \"failover\"\n"
          "                   <num>: stripe size(kB) for round-robin"
          " multipathing\n");
   printf("  -V               Version information.\n");
   printf("  -v               Enable verbose output.\n");

   return;
}



int main(int argc, char **argv)
{
   int        error = 0,
              tmperr,
              count,
              c,
              i,
              j,
              verbose = 0,
              quiet = 0,
              len;
   uint64_t   mod_ver = 0;
   char       action = 0,
              **pool_names = NULL;
   int        mp = MP_NOT_CHANGED;
   diskinfo_t di = {0};
   struct sstring *emsg = NULL;
   char       msg[STRING_SIZE];

   while((c = getopt(argc, argv, "aDhm:qrVv")) != -1 && !error) {
      switch(c) {
         case 'a':
            if(action) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'D':
            add_flag(POOL_DEBUG);
            verbose = 3;
            break;

         case 'h':
            _usage(argv[0]);
            return error;
            break;

         case 'm':
            if(strcmp(optarg, "none") == 0) {
               mp = -1;
            }
            else if (strcmp(optarg, "failover") == 0) {
               mp = 0;
            }
            else {
               mp = atoi(optarg);
            }
            break;

         case 'q':
            if(check_flag(POOL_VERBOSE)) {
               mutually_exclusive(argv[0], c, 'v');
               error = 1;
            }
            else {
               add_flag(POOL_QUIET);
               quiet++;
            }
            break;

         case 'r':
            if(action) {
               mutually_exclusive(argv[0], c, action);
               error = 1;
            }
            else
               action = c;
            break;

         case 'V':
            PRINT_VERSION(argv[0], POOL_RELEASE_NAME, GFS_COPYRIGHT);
            /*printf("%s %s (built %s %s)\n", argv[0],
                  GFS_RELEASE_NAME, __DATE__, __TIME__);
                   printf("%s\n", GFS_COPYRIGHT);*/
            return error;
            break;

         case 'v':
            if(check_flag(POOL_QUIET)) {
               mutually_exclusive(argv[0], c, 'q');
               error = 1;
            }
            else {
               verbose++;
               add_flag(POOL_VERBOSE);
               break;
            }

         case ':':
         case '?':
            fprintf(stderr, "Please use '-h' for usage.\n");
            exit(1);
            break;

         default:
            fprintf(stderr, "Bad programmer! You forgot to catch the"
                 " %c flag\n", c);
            exit(1);
            break;
      }

   }

   if(argc > optind) {
      pool_names = &(argv[optind]);
   }

   if(!error) {
      if((error = ccs_initlog("pool.ccs", "pool/tool_config",
                              verbose, quiet))) {
         if((error = init_defaultlog(verbose, quiet, NULL)) != 0) {
            log_err("Unable to initialize tool logger");
            goto end;
         }
      }

      emsg = ss_init(STRING_SIZE-1);

      if((error = open_comm(emsg))) {
         if(error == ENXIO) {
            log_err("Unable to open pool comm device.  "
                      "Verify pool module is loaded");
            cstring2ss(emsg, "", 0);
         }
         else {
            ss_len(emsg, &len);
            if(len > 0) {
               ss2cstring(msg, STRING_SIZE, emsg);
               log_crit(msg);
            }
         }
      }
      else if ((error = check_version(emsg))) {
         if(error > 0)
            log_crit("Pool module version %"PRIu64" incompatible with"
                  " this tool.", mod_ver);
         else
            log_crit("Unable to retrieve pool module version.");
      }
      else {

         /* Clear out block devices in the /dev/pool directory */
         clean_pool_dir(emsg);

         switch(action) {

            /* Assemble */
            case 0:
            case 'a':

               /* Grab the current info off the disk */
               log_info("Scanning disk labels...");
               if((error = scan_disks(&di, emsg))) {
                  log_info("FAILED.");
                  ss2cstring(msg, STRING_SIZE, emsg);
                  log_err(msg);

                  break;
               }
               log_info("ok.");

               if(!di.pl_count && pool_names == NULL) {
                  log_warn("No pool labels found on disk");
                  break;
               }

               if(pool_names != NULL) {
                  /* look for the requested pool(s) */
                  for(j = 0; pool_names[j] != NULL; j++) {
                     for(i = 0; i < di.pl_count; i++) {
                        if(!strcmp(di.pl[i].name, pool_names[j]))
                           break;
                     }

                     if(i == di.pl_count) {
                        log_err("%s not found on disk.", pool_names[j]);
                        error |= 1;
                        continue;
                     }
                     tmperr = _assemble_pool(pool_names[j], emsg);
                     error |= tmperr;
                     ss_len(emsg, &len);
                     if (tmperr != 0 && len > 0) {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                        cstring2ss(emsg, "", 0);
                     }
                     else if(mp != MP_NOT_CHANGED) {
                        tmperr = tune_multipath(pool_names[j], mp, emsg);
                        error |= tmperr;
                        ss_len(emsg, &len);
                        if(tmperr != 0 && len > 0) {
                           ss2cstring(msg, STRING_SIZE, emsg);
                           log_err(msg);
                           cstring2ss(emsg, "", 0);
                        }
                     }
                  }
               }
               else {
                  count = 0;

                  /* assemble all pools that were found */
                  for(i = 0; i < di.pl_count; i++) {
                     /* need to update device nodes here so we can query the
                      * assembled pool immediately - the cache in the libary
                      * might be able to be fixed so this is unnecessary...
                      */
                     tmperr = _assemble_pool(di.pl[i].name, emsg);
                     if(tmperr == 0) {
                        if(mp != MP_NOT_CHANGED) {
                           tmperr = tune_multipath(di.pl[i].name, mp, emsg);
                           if(tmperr != 0) {
                              ss2cstring(msg, STRING_SIZE, emsg);
                              log_err(msg);
                              cstring2ss(emsg, "", 0);
                           }
                        }
                     }
                     else {
                        ss_len(emsg, &len);
                        /* If the pool is already loaded, ignore the error */
                        if(tmperr == EEXIST) {
                           tmperr = 0;
                        }
                        else if (len > 0) {
                           ss2cstring(msg, STRING_SIZE, emsg);
                           log_err(msg);
                           cstring2ss(emsg, "", 0);
                        }
                     }
                     /* OR the errors so if any errors occur, pool_assemble
                      * detects them
                      */
                     error |= tmperr;
                     /* FIXME: should i abort on error or try to load as many
                        pools as possible */
                  }

                  if(count == 0 && error == 0) {
                     log_notice("No additional pools assembled.");
                  }
               }

               update_device_nodes(emsg);
               break;

            /* Disassemble (remove) */
            case 'r':
               if(pool_names) {
                  for(j = 0; pool_names[j] != NULL; j++) {
                     tmperr = _disassemble_pool(pool_names[j],emsg);
                     if(tmperr != 0){
                        error |= tmperr;
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                        cstring2ss(emsg, "", 0);
                     }
                  }
               }
               else {
                  count = 0;

                  log_info("Scanning kernel for assembled pools...");
                  if((error = get_loaded_pools(&di, emsg))) {
                     log_info("FAILED.");
                     ss2cstring(msg, STRING_SIZE, emsg);
                     log_err(msg);
                     break;
                  }

                  log_info("ok.");
                  if(!di.pl_count) {
                     log_warn("No pools are assembled.");
                     break;
                  }

                  /* disassemble all loaded pools */
                  for(i = 0; i < di.pl_count; i++) {
                     tmperr = _disassemble_pool(di.pl[i].name, emsg);
                     if(tmperr == 0) {
                        update_device_nodes(emsg);
                        count++;
                     }
                     else {
                        ss2cstring(msg, STRING_SIZE, emsg);
                        log_err(msg);
                        cstring2ss(emsg, "", 0);
                     }
                     error |= tmperr;
                  }

                  if(count == 0) {
                     log_warn("No pools disassembled.");
                  }
                  destroy_di(&di);
               }

               break;

            default:
               fprintf(stderr, "Please use -h' for usage.\n");
               error = 1;
               break;
         }
      }
   }
  end:
   close_comm(emsg);
   /* Until we have well defined error codes, the exit status will either
      be 0 or 1 */
   if(error != 0) {
      error = 1;
   }
   return error;

}

/* Gets a single character from the input stream and flushes the rest */
static char _get_a_char(FILE *fp) {
   int c, d;

   d = c = fgetc(fp);
   while (d != '\n' && d != EOF)
      d = fgetc (fp);

   return (char)c;
}

/*
 * _check_pooldevs -
 * @poolname:
 *
 * Description: check to see if /dev/pool/{poolname} exists, and if so
 * check what kind of file it is.  If it's a block device, skip it - otherwise
 * prompt the user as to whether to remove it or not.
 *
 */
int _check_pooldevs(char *poolname)
{
   int error = 0;
   struct stat st;
   char path[POOL_NAME_SIZE+10];

   strcpy(path, "/dev/pool/");
   strcat(path, poolname);

   if((error = stat(path, &st)) < 0) {
      /* if there is no file, it's good and we can just leave */
      if(errno == ENOENT) {
         error = 0;
         goto end;
      }

      /* otherwise return the system error code */
      error = errno;
      log_err("Error accessing %s: %s", path, strerror(errno));
      goto end;
   }
   if(S_ISBLK(st.st_mode)) {
      goto end;
   }
   /* dirs need interaction */
   if(S_ISDIR(st.st_mode)) {
      if(check_flag(POOL_FORCE) != 1) {
         printf("%s exists - remove? (y/n): ", path);
         if(tolower(_get_a_char(stdin)) != 'y') {
            error = 1;
            goto end;
         }
      }
      if(rmdir(path) != 0) {
         log_err("Unable to remove %s: %s", path, strerror(errno));
         error = errno;
      }
      goto end;
   }
   /* There is an accessible file there, determine what to do about it */
   /* anything else needs interaction */
   if(check_flag(POOL_FORCE) == 0) {
      printf("%s exists - remove? (y/n): ", path);
      if(tolower(_get_a_char(stdin)) != 'y') {
         error = 1;
         goto end;
      }
   }
   if(unlink(path) != 0) {
      log_err("Unable to remove %s: %s", path, strerror(errno));
      error = errno;
   }

end:
   return error;
}

/*
 * _is_pool_assembled - queries kernel to see if specified pool is assembled
 * @name: pool to check
 *
 * Returns: positive if pool is already assembled, zero if not, and negative
 *          on error.
 */
static int _is_pool_assembled(char *name, struct sstring *emsg)
{
   diskinfo_t di = {0};
   int found = 0;
   int i;
   int error;

   log_info("Scanning to see if %s is assembled...", name);
   if((error = get_loaded_pools(&di, emsg)) != 0) {
      log_info("FAILED.");
      found = -1;
   }
   else {
      log_info("ok.");
      for(i = 0; i < di.pl_count; i++) {
         if(!strcmp(name, di.pl[i].name)) {
            found = 1;
            break;
         }
      }
   }
   destroy_di(&di);
   return found;
}

/*
 * _assemble_pool - assembles pools (duh)
 * @poolname: pool to assemble, or NULL if all should be assembled
 *
 * Returns: zero on success, non-zero on failure
 */
static int _assemble_pool(char *poolname, struct sstring *emsg)
{
   int error = 0;

   if((error = _check_pooldevs(poolname)) != 0) {
      log_warn("%s not assembled", poolname);
      cstring2ss(emsg, "", 0);
      goto end;
   }

   /* Check to see if pool is already loaded... */
   error = _is_pool_assembled(poolname, emsg);
   if( error < 0) {
      log_err("Unable to check kernel for assembled pools...aborting.");
   }
   else if (error > 0) {
      ss_sprintf(emsg, "%s already loaded", poolname);
      error = EEXIST;
   }
   else {
      /* pool has been found on disk - now try to assemble it */
      log_info("Assembling %s...", poolname);
      if((error = assemble_pool(poolname, emsg)) != 0) {
         log_info("failed.");
      }
      else {
         log_info("ok.");
         log_warn("%s assembled.", poolname);
      }
   }

end:
   return error;
}

/*
 * _disassemble_pool - disassembles pools (duh)
 * @poolname: pool to disassemble, or NULL if all should be disassembled
 *
 * Returns: zero on success, non-zero on failure
 */
static int _disassemble_pool(char *poolname, struct sstring *emsg)
{
   int error;

   log_info("Disassembling %s...", poolname);
   if((error = disassemble_pool(poolname, emsg))) {
      log_info("failed.");
   }
   else {
      log_info("ok.");
      log_warn("%s disassembled.", poolname);
   }

   return error;
}
/*
 * vim: et ts=3 sw=3 noai cin :
 */

