/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <sys/uio.h>
#include <string.h>

#include <stdio.h>
#include <stdlib.h>

/* ccs_writev - similar to writev
 * @fd: file descriptor
 * @iov: the io vector
 * @count: number of iov's in the array
 * @retries: max number of writes to complete total write
 *
 * This function operates like writev, with the exception that it
 * does not tolerate partial failures.  It will either write the
 * entire amount in the iov's, or it will return an error.  It
 * will make as many as <retries> calls to writev to accomplish
 * this.  So, if writev returns an error, ccs_write will return
 * an error.  If writev successfully writes all iov's, ccs_write
 * will return success.  If writev does not write all iov's (or
 * partial iov's), ccs_write will reissue the writev from the
 * correct offset.
 *
 * Returns: 0 on success, -ERRNO on error, or amount not written on failure
 */

int ccs_writev(int fd, struct iovec *iov, int count, int retries){
  int tmp, tries, i, total;

  for(i=0, total=0; i < count; i++){
    total += iov[i].iov_len;
  }
  tmp = writev(fd, iov, count);

  if(tmp < 0){
    return tmp;
  } else if (tmp < total){
    struct iovec retry[count];

    memcpy(retry, iov, sizeof(retry));
    total -= tmp;

    for(tries=0; tries < retries && total; tries++){
      for(i=0; (tmp >= retry[i].iov_len) && (i < count); i++){
	tmp -= retry[i].iov_len;
	retry[i].iov_len = 0;
      }

      retry[i].iov_base += tmp;
      retry[i].iov_len -= tmp;
      
      tmp = writev(fd, retry, count);
      if(tmp < 0) return tmp;
      total -= tmp;
    }
    return total;
  } else {
    return 0;
  }
}

