/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <wait.h>
#include <sys/types.h>
#include <dirent.h>
#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "ccs_ft.h"

#include "global.h"
#include "copyright.cf"
#include "ccs_release.cf"
#include "ccs_log.h"
#include "ccs_lockfile.h"
#include "ccs_iov.h"
#include "ccs_hash.h"

#define CCS_DEFAULT_PATH "/etc/sistina/ccs-build"
#define CCS_SERVD_LOCKFILE "/var/run/sistina/ccs_servd.pid"

static char *ccs_path = CCS_DEFAULT_PATH;
static int verbose=0;

static void _destroy_lockfile(void);
static void print_start_msg(void);
static int set_signals(void);
static void print_usage(FILE *stream);
static int is_valid_archive_dir(char *path);
static int process_request(int afd);

int main(int argc, char *argv[]){
  int c;
  int port = CCS_DEFAULT_BEPORT;

  int daemon = 1;
  int sock_fd;
  int accept_fd;
  struct sockaddr_in addr;
  int len = sizeof(struct sockaddr_in);

  while((c = getopt(argc, argv, "hnP:p:Vv")) != -1){
    switch(c){
    case 'h':
      print_usage(stdout);
      exit(EXIT_SUCCESS);

    case 'n':
      daemon = 0;
      break;

    case 'P':
      port = atoi(optarg);
      if(!port){
	fprintf(stderr, "Invalid port number, %s\n", optarg);
	exit(EXIT_FAILURE);
      }
      break;

    case 'p':
      ccs_path = optarg;
      break;

    case 'V':
      printf("%s %s (built %s %s)\n", argv[0], CCS_RELEASE_NAME,
	     __DATE__, __TIME__);
      printf("%s\n", SISTINA_COPYRIGHT);
      exit(EXIT_SUCCESS);
      break;
    case 'v':
      verbose = 1;
      break;
    default:
      print_usage(stderr);
      exit(EXIT_FAILURE);
    }
  }

  if(argc > optind){
    fprintf(stderr, "Too many arguments.\n"
	    "Try 'ccs_servd -h' for help.\n");
    exit(EXIT_FAILURE);
  }    

  if(ccs_path[0] != '/'){
    char tmp[256];
    
    if(!getcwd(tmp, 256)){
      fprintf(stderr, "Unable to get current working directory.\n");
      exit(EXIT_FAILURE);
    }
    sprintf(tmp+strlen(tmp), "/%s", ccs_path);
    ccs_path = strdup(tmp);
    if(!ccs_path){
      fprintf(stderr, "Unable to allocate memory for archive directory.\n");
      exit(EXIT_FAILURE);
    }
  }
  

  if(!is_valid_archive_dir(ccs_path)){
    exit(EXIT_FAILURE);
  }

  openlog("ccs_servd", LOG_PID, LOG_DAEMON);

  if(daemon){
    int pid;

    if(test_lockfile(CCS_SERVD_LOCKFILE)){
      exit(EXIT_FAILURE);
    }

    if((pid = fork()) < 0){
      fprintf(stderr, "Unable to fork().\n");
      exit(EXIT_FAILURE);
    } else if (pid !=0){
      /* close the parent */
      exit(EXIT_SUCCESS);
    }

    setsid();
    chdir("/");
    umask(0);

    close(0); close(1); close(2);
    open("/dev/null", O_RDONLY);  /* reopen stdin */
    open("/dev/null", O_WRONLY);  /* reopen stdout */
    open("/dev/null", O_WRONLY);  /* reopen stderr */    
    if(create_lockfile(CCS_SERVD_LOCKFILE)){
      exit(EXIT_FAILURE);
    }
  }

  atexit(&_destroy_lockfile);  /* destroy lockfile at exit */
  set_signals();
  print_start_msg();

  sock_fd = socket(PF_INET, SOCK_STREAM, 0);

  if(sock_fd < 0){
    log_sys_err("Socket creation failed.");
    exit(EXIT_FAILURE);
  } else {
    int trueint = 1;
    setsockopt(sock_fd, SOL_SOCKET, SO_REUSEADDR, &trueint, sizeof(int));
  }

  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  addr.sin_port = htons(port);

  if(bind(sock_fd, (struct sockaddr *)&addr, sizeof(struct sockaddr_in)) < 0){
    log_sys_err("Unable to bind socket.");
    close(sock_fd);
    exit(EXIT_FAILURE);
  }

  listen(sock_fd, 5);
  
  while(1){
    int pid;

    accept_fd = accept(sock_fd, (struct sockaddr *)&addr, &len);

    if(accept_fd < 0){
      log_sys_err("Unable to accept connection");
      continue;
    }

    pid = fork();

    if(pid < 0){
      log_sys_err("Unable to fork.");
      /* Rely on exit to close sock_fd and accept_fd */
      exit(EXIT_FAILURE);
    } else if(pid){
      /* parent doesn't need accept_fd */
      close(accept_fd);
    } else {
      /* child doesn't need sock_fd */
      close(sock_fd);
      break;
    }
    wait(NULL);
  }

  if(verbose){
    unsigned char *saddr = (char *)(&addr.sin_addr);
    log_msg("Connection accepted from host: %u.%u.%u.%u.\n",
	    saddr[0],saddr[1],saddr[2],saddr[3]);
  }

  if(process_request(accept_fd)){
    close(accept_fd);
    exit(EXIT_FAILURE);
  }
  close(accept_fd);

  exit(EXIT_SUCCESS);
}


static void _destroy_lockfile(void){
  destroy_lockfile(CCS_SERVD_LOCKFILE);
}

static void print_start_msg(void){
  log_msg("Starting ccs_servd %s:\n", CCS_RELEASE_NAME);
  log_msg(" Built: " __DATE__" " __TIME__"\n");
  log_msg(" %s\n", SISTINA_COPYRIGHT);
}

static void sig_handler(int sig){
  switch(sig){
  case SIGINT:
    log_msg("Stopping ccs_servd, SIGINT received.\n");
    exit(EXIT_SUCCESS);
  case SIGQUIT:
    log_msg("Stopping ccs_servd, SIGQUIT received.\n");
    exit(EXIT_SUCCESS);
  case SIGTERM:
    log_msg("Stopping ccs_servd, SIGTERM received.\n");
    exit(EXIT_SUCCESS);
  case SIGSEGV:
    log_err("Stopping ccs_servd, SIGSEGV received.\n");
    exit(EXIT_FAILURE);
  default:
    log_err("Stopping ccs_servd, unknown signal received.\n");
    exit(EXIT_FAILURE);
  }
}

static int set_signals(void){
  signal(SIGINT, &sig_handler);
  signal(SIGQUIT, &sig_handler);
  signal(SIGTERM, &sig_handler);
  signal(SIGSEGV, &sig_handler);
  return 0;
}

static void print_usage(FILE *stream){
  fprintf(stream,
	  "Usage:\n\nccs_servd [options]\n\nOptions:\n"
	  " -h          Print this help, then exit.\n"
	  " -n          No Daemon.  Run in the foreground.\n"
	  " -P <port>   Port number on which to listen for requests.\n"
	  " -p <path>   Path to cluster configuration directories.\n"
	  " -V          Print version information.\n"
	  " -v          Verbose.\n"
	  );
}


/*
 * is_valid_archive_dir: does the directory contain .cca files
 * @path:
 *
 * Returns: 1 if ok, 0 if not valid
 */
static int is_valid_archive_dir(char *path){
  DIR *dir;
  struct dirent *dirent;
  FILE *fp;
  char filename[256];
  char line[256];
  char *msg=NULL;
  int found=0;

  dir = opendir(path);
  if(!dir){
    fprintf(stderr, "Unable to open the directory, %s: %s\n",
	    ccs_path, strerror(errno));
    return 0;
  }

  while((dirent = readdir(dir))){
    if((dirent->d_name[strlen(dirent->d_name)-4] == '.') &&
       (dirent->d_name[strlen(dirent->d_name)-3] == 'c') &&
       (dirent->d_name[strlen(dirent->d_name)-2] == 'c') &&
       (dirent->d_name[strlen(dirent->d_name)-1] == 'a')){
      memset(filename, 0, sizeof(filename));
      strcpy(filename, path);
      if(filename[strlen(filename)-1] != '/'){
	filename[strlen(filename)] = '/';
      }
      strcpy(filename+strlen(filename), dirent->d_name);
      fp = fopen(filename, "r");
      if(!fp){
	fprintf(stderr, "Unable to open, %s: %s\n",
		filename, strerror(errno));
	continue;
      }

      if(!fgets(line, sizeof(line), fp)){
	fclose(fp);
	continue;
      }

      if(strncmp(line, "#CCA:", 5)){
	fclose(fp);
	continue;
      }
      found=1;
      if(!msg){
	msg = malloc(512);
	memset(msg, 0, 512);
	sprintf(msg, "Valid cluster entries:\n %s", dirent->d_name);
	msg[strlen(msg)-4] = '\n';
	msg[strlen(msg)-3] = '\0';
      } else {
	strcpy(msg+strlen(msg), dirent->d_name);
	msg[strlen(msg)-4] = '\n';
	msg[strlen(msg)-3] = '\0';
      }	
      fclose(fp);
    }
  }
  closedir(dir);
  if(!found){
    fprintf(stderr, "There are no valid CCS archives in %s.\n", path);
    return 0;
  } else {
    if(verbose){
      printf("%s\n", msg);
    }
    return 1;
  }
}


/*
 * is_valid_archive_file
 * @archive: the name of the archive file to check
 *
 * Returns: 1 if ok, 0 if not valid
 */
static int is_valid_archive_file(char *archive){
  uint32 hash=0;
  char line[256];
  FILE *arch_fp = NULL;
  char *filedata = NULL;
  struct stat stat_buf;

  if(stat(archive, &stat_buf)){
    log_sys_err("Unable to stat the CCS archive, %s\n", archive);
    return 0;
  }

  /* check if archive modified since last check */

  if((arch_fp = fopen(archive, "r")) == NULL){
    log_sys_err("Unable to open CCS archive \"%s\": ", archive);
    goto fail;
  }

  while(fgets(line, sizeof(line), arch_fp)){
    uint mtime=0;
    uint size=0;
    char filename[256];
    char filename_cmp[256];
    if(sscanf(line, "#nigeb=%s mtime=%u size=%u\n",
	      filename, &mtime, &size) == 3){
      filedata = malloc(size);
      if(!filedata)
	goto fail;
      if(fread(filedata, sizeof(char), size, arch_fp) != size){
	  log_err("Unable to read file, %s from the CCS archive, %s.\n",
		  filename, archive);
	  goto fail;
      }
      if(!fgets(line, sizeof(line), arch_fp) ||
	 !sscanf(line, "#dne=%s hash=%X\n", filename_cmp, &hash) ||
	 strcmp(filename, filename_cmp)){
	goto fail;
      } else {
	if(hash != ccs_hash(filedata, size)){
	  goto fail;
	}
      }
      free(filedata);
      filedata = NULL;
    }
  }

  /* is a valid archive file */
  return 1;

 fail:
  if(filedata)
    free(filedata);
  if(arch_fp)
    fclose(arch_fp);
  return 0;
}

static int process_request(int afd){
  int found=0;
  int alternate_request=0;
  FILE *arch_fp=NULL;
  char *tmp;
  char request[100];
  char ccs_archive[100];
  char ccs_file[100];
  char line[256];
  char sector[512];
  ccs_fth_t header;
  struct iovec iov[2];

  header.fth_magic = 0x101674;
  header.fth_error = 0;
  header.fth_last = 0;

  iov[0].iov_base = &header;
  iov[0].iov_len  = sizeof(header);
  iov[1].iov_base = sector;

  memset(request, 0, sizeof(request));
  memset(ccs_archive, 0, sizeof(ccs_archive));
  memset(ccs_file, 0, sizeof(ccs_file));
  memset(sector, 0, sizeof(sector));

  if(read(afd, request, sizeof(request)) < 0){
    log_sys_err("Unable to read from socket.\n");
    goto fail;
  }

  /* Remember, the only request that can exist in an archive are **
  ** CCS files and will have the extention .ccs.  So, anything   **
  ** without a .ccs extention can effectively be considered a    **
  ** command/option...........................................   */
  request[99] = '\0';  /* truncate any requests that are too large */
  log_dbg("Received a request for \"%s\".\n", request);

  tmp = strstr(request, "/");
  if(tmp){
    tmp[0] = '\0';
    tmp++;
  } else {
    if(request[0] != 0)
      log_err("Ill formed request, \"%s\"\n", request);
    goto fail;
  }
  strcpy(ccs_file, tmp);

  sprintf(ccs_archive, "%s/%s.cca", ccs_path, request);

  if(!strcmp(ccs_file, "list_clusters")){
    DIR *dir;
    struct dirent *dirent;
    char _name[256];

    if(verbose){
      log_msg("Received request for cluster list.\n");
    }     
    memset(_name, 0, sizeof(_name));
    dir = opendir(ccs_path);
    while((dirent = readdir(dir))){
      if((tmp = strstr(dirent->d_name, ".cca")) && (tmp[4] == '\0')){
	memcpy(_name, dirent->d_name, strlen(dirent->d_name)-4);
	if(strlen(_name)+strlen(sector) > sizeof(sector)){
	  closedir(dir);
	  goto fail;
	}
	sprintf(sector+strlen(sector), "%s\n", _name);
	memset(_name, 0, sizeof(_name));
      }
    }
    closedir(dir);
    alternate_request=1;
    goto out;
  }
  if(verbose){
    if(!strcmp(ccs_file, "list_files")){
      log_msg("Received request for file list.\n");
    } else {
      log_msg("Received request for file \"%s\" from archive \"%s.cca\".\n",
	      ccs_file, request);
    }
  }     

  if(!is_valid_archive_file(ccs_archive)){
    log_err("WARNING:\n");
    log_err("The CCS archive, %s.cca, has been corrupted.\n", request);
    log_err("This may cause cluster-wide problems.\n");
    log_err("Always use ccs_tool to create or update CCS archives.\n\n");
  }

  if((arch_fp = fopen(ccs_archive, "r")) == NULL){
    log_sys_err("Unable to open CCS archive \"%s\": ", ccs_archive);
    goto fail;
  }

  while(fgets(line, sizeof(line), arch_fp)){
    uint mtime=0;
    uint size=0;
    char filename[256];
    char filename_cmp[256];
    if(sscanf(line, "#nigeb=%s mtime=%u size=%u\n",
	      filename, &mtime, &size) == 3){
      int diff, amount;

      if(!strcmp(ccs_file, "list_files")){
	/* fill sector with file names */
	alternate_request=1;
	if(strlen(filename)+strlen(sector) > sizeof(sector)){
	  /* for now, error out.  However, should start new sector */
	  goto fail;
	}
	sprintf(sector+strlen(sector), "%s\n", filename); 
	continue;
      } else if(strcmp(ccs_file, filename)){
	/* Not the file we are looking for */
	continue;
      }

      found=1;
      diff=size;
      amount=(diff < sizeof(sector))?diff:sizeof(sector);

      while(1){
	memset(sector, 0, sizeof(sector));
	if(fread(sector, sizeof(char), amount, arch_fp) != amount){
	  log_err("Unable to read file, %s from temporary archive.\n",
		  ccs_file);
	  goto fail;
	}
	diff -= amount;

	iov[1].iov_len = header.fth_count = amount;

	if(diff){
	  if(ccs_writev(afd, iov, 2, 10)){
	    log_sys_err("Unable to complete write to socket.\n");
	    goto fail;
	  }
	} else {
	  if(!fgets(line, sizeof(line), arch_fp) ||
	     !sscanf(line, "#dne=%s\n", filename_cmp) ||
	     strcmp(filename, filename_cmp)){
	    goto fail;
	  } else {
	    header.fth_last=1;
	    
	    if(ccs_writev(afd, iov, 2, 10)){
	      log_err("Unable to complete write to socket.\n");
	      goto fail;
	    }
	    return 0;
	  }
	}
	amount=(diff < sizeof(sector))?diff:sizeof(sector);
      }
    }
  }
 out:
  if(alternate_request){
    found = 1;
    header.fth_magic = 0x101674;
    header.fth_count = strlen(sector)+1;
    header.fth_last  = 1;
    header.fth_error = 0;
    iov[1].iov_len = header.fth_count;
    /* ATTENTION -- verify strlen is correct. */

    if(ccs_writev(afd, iov, 2, 10)){
      log_err("Unable to complete write to socket.\n");
      goto fail;
    }
  }  
    
  if(!found){
    header.fth_magic = 0x101674;
    header.fth_count = 0;
    header.fth_last  = 1;
    header.fth_error = 0;

    write(afd, (char *)&header, sizeof(header));
  }  

  if(arch_fp) fclose(arch_fp);
  if(verbose){
    log_msg("Completed request for %s.\n",
	    (!strcmp(ccs_file, "list_clusters"))? "cluster list":
	    (!strcmp(ccs_file, "list_files"))? "file list": ccs_file);
  } else {
    log_dbg("Completed request for \"%s\".\n", ccs_file);
  }
  return 0;

 fail:
  log_dbg("Failed on request for \"%s\".\n", ccs_file);
  header.fth_magic = 0x101674;
  header.fth_count = 0;
  header.fth_last  = 1;
  header.fth_error = 1;

  write(afd, (char *)&header, sizeof(header));
  if(arch_fp) fclose(arch_fp);
  return -1;
}
