/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**  
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include "ccs.h"
#include "fence_log.h"

#define MAX_FENCE_ARGS 128

/* build_fence_command
 * @method_cn: the CCS node to build the fence cmd from
 * @fence_agent: where the agent will be placed
 * @fence_args:  where the agent arguments are placed
 *
 * Consider the following segment from fence.ccs:
 * ...
 * fence {
 *   fence1 {
 *     brocade {
 *       port = 5
 *     }
 *   }
 * }
 * ...
 * It would be the node describing the subsection "brocade"
 * that gets passed in as a parameter.  This function
 * will look in fence.ccs and find the brocade HW, and
 * build the command that would issue the fence.  NOTE:
 * MEMORY IS ALLOCATED FOR THE FENCE_AGENT AND FENCE_ARGS
 * AND *MUST* BE FREED WHEN IT IS NO LONGER NEEDED.
 *
 * Returns: 0 on success, -1 on error
 */
static int build_fence_command(ccs_node_t *method_cn,
			       char **fence_agent, char **fence_args){
  int i, error;
  ccs_node_t *fence_cn=NULL;
  ccs_node_t *hw_cn=NULL;
  ccs_node_t *tmp_cn=NULL;
  ccs_value_t *value;
  char request[256];
  char tmp_arg[100];

  const char *agent=NULL;
  char *args[MAX_FENCE_ARGS]; /* no more than MAX_FENCE_ARGS possible args */
  int arg_index=0;
  int arg_size=0;

  log_dbg("Entering build_fence_command()\n");

  *fence_agent = *fence_args = NULL;

  memset(args, 0, sizeof(char *)*MAX_FENCE_ARGS);

  error = open_ccs_file(&fence_cn, "fence.ccs");
  if(error == -EPARSE){
    log_err("Unable to open fence.ccs:  Format error.\n");
    log_err("Has the CCS archive been corrupted?\n\n");
    goto fail;
  } else if(error){
    char *msg = strerror(-error);
    log_err("Unable to open fence.ccs:  %s\n", msg);
    goto fail;
  }

  sprintf(request, "fence_devices/%s", method_cn->key);
  hw_cn = find_ccs_node(fence_cn, request, '/');
  if(!hw_cn){
    log_err("Cannot find hardware description for %s\n",
	    method_cn->key);
    goto fail;
  }

  sprintf(request, "%s/agent", hw_cn->key);
  agent = find_ccs_str(hw_cn, request, '/', NULL);

  if(!agent){
    log_err("No agent specified for fencing method, %s\n",
	    method_cn->key);
    goto fail;
  }
  *fence_agent = (char *)malloc(strlen(agent)+1);
  if(!*fence_agent){
    log_err("Can't allocate memory for agent string.\n");
    goto fail;
  }
  strcpy(*fence_agent, agent);

  log_dbg("Fencing agent:  %s\n", agent);

  for(tmp_cn = hw_cn->child;
      (tmp_cn && arg_index < MAX_FENCE_ARGS);
      tmp_cn = tmp_cn->sib){
    if(!strcmp(tmp_cn->key, "agent")){
      continue;
    }
    for (value = tmp_cn->v; value && arg_index < MAX_FENCE_ARGS; value = value->next){
      if(value->type == CCS_STRING){
        sprintf(tmp_arg, "%s=%s", tmp_cn->key, value->v.str);
      } else if(tmp_cn->v->type == CCS_INT){
        sprintf(tmp_arg, "%s=%d", tmp_cn->key, value->v.i);
      } else {
        log_err("Unsupported value type for %s/%s (%d).\n",
                method_cn->key, tmp_cn->key, value->type);
        goto fail;
      }
      args[arg_index] = (char *)malloc(strlen(tmp_arg)+1);
      if(!args[arg_index]){
        log_err("Cannot allocate memory for fence command.\n");
        goto fail;
      }
      strcpy(args[arg_index], tmp_arg);
      arg_index++;
    }
  }

  tmp_cn = method_cn->child;
  if(!tmp_cn){
    log_msg("Warning:  There are no node specific fencing parameters.\n");
  } else {
    for(tmp_cn = method_cn->child;
	(tmp_cn && (arg_index < MAX_FENCE_ARGS));
	tmp_cn = tmp_cn->sib){
      for (value = tmp_cn->v; value && arg_index < MAX_FENCE_ARGS; value = value->next){
	if(value->type == CCS_STRING){
	  sprintf(tmp_arg, "%s=%s", tmp_cn->key, value->v.str);
	} else if(value->type == CCS_INT){
	  sprintf(tmp_arg, "%s=%d", tmp_cn->key, value->v.i);
	} else {
	  log_err("Unsupported value type for %s/%s (%d).\n",
		  method_cn->key, tmp_cn->key, value->type);
	  goto fail;
	}
	args[arg_index] = (char *)malloc(strlen(tmp_arg)+1);
	if(!args[arg_index]){
	  log_err("Cannot allocate memory for fence command.\n");
	  goto fail;
	}
	strcpy(args[arg_index], tmp_arg);
	arg_index++;
      }
    }
  }

  if(arg_index == MAX_FENCE_ARGS){
    log_err("Too many arguments to fence agent.\n");
    goto fail;
  }

  log_dbg("Agent parameters:\n");
  for(i=0; i < arg_index; i++){
    log_dbg("  %s\n", args[i]);
  }

  for(i = 0; i < arg_index; i++){
    /* plus 1 for the extra \n character */
    arg_size += (strlen(args[i])+1);
  }
  arg_size++;  /* for the trailing \0 character */
  *fence_args = (char *)malloc(arg_size);
  if(!*fence_args){
    log_err("Can't allocate space for fence arguments.\n");
    goto fail;
  }

  memset(*fence_args, 0, arg_size);
  for(i = 0; i < arg_index; i++){
    arg_size = strlen(*fence_args);
    sprintf(&((*fence_args)[arg_size]), "%s\n", args[i]);
  }
  
  for(i = 0; i < arg_index; i++){
    free(args[i]);
  }

  log_dbg("Exiting build_fence_command() - SUCCESS\n");
  return 0;

 fail:
  for(i = 0; i < arg_index; i++){
    free(args[i]);
  }
  if(*fence_agent) free(*fence_agent);
  if(*fence_args) free(*fence_args);
  if(fence_cn) close_ccs_file(fence_cn);
  log_dbg("Exiting build_fence_command() - FAILURE\n");
  return -1;
}


/*
 * run_fence_command - exec the actual fence agent and get rtn status
 * @agent:
 * @agent_args:
 *
 * This function will exec the actual agent program.  It will grab
 * its completion status and return accordingly.
 *
 * Return: 0 if agent successful, -1 on failure
 */
static int run_fence_command(char *agent, char *agent_args){
  char agent_output[256];
  int pid, val, status;
  int pr_fd, pw_fd;  /* parent read/write file descriptors */
  int cr_fd, cw_fd;  /* child read/write file descriptors */
  int fd1[2];
  int fd2[2];

  cr_fd = cw_fd = pr_fd = pw_fd = -1;

  if(pipe(fd1)){
    goto fail;
  }
  pr_fd = fd1[0];
  cw_fd = fd1[1];

  if(pipe(fd2)){
    goto fail;
  }
  cr_fd = fd2[0];
  pw_fd = fd2[1];

  pid = fork();
  if(pid < 0){
    log_err("Unable to fork and exec fence command.\n");
    goto fail;
  } else if(pid){
    val = fcntl(pr_fd, F_GETFL, 0);
    val |= O_NONBLOCK;
    fcntl(pr_fd, F_SETFL, val);

    /* send fence arguments to the agent */
    if(write(pw_fd, agent_args, strlen(agent_args)) !=
       strlen(agent_args)){
      log_err("Unable to send arguments to the agent.\n");
      goto fail;
    }

    close(pw_fd);
    
    waitpid(pid, &status, 0);

    memset(agent_output,0, sizeof(agent_output));

    if(WIFEXITED(status)){
      if(WEXITSTATUS(status)){
	log_err("The agent (%s) reports: ", agent);
	while(read(pr_fd, agent_output, sizeof(agent_output)-1) > 0){
	  log_err("%s", agent_output);
	  memset(agent_output,0, sizeof(agent_output));
	}
	log_err("\n");
	goto fail;
      } else {
	if(read(pr_fd, agent_output, sizeof(agent_output)-1) < 0){
	  log_err("Unable to read output from the agent.\n");
	} else {
	  log_msg("The agent (%s) reports: %s\n", agent, agent_output);
	}
      }
    } else {
      log_err("The agent (%s) exited abnormally.\n", agent);
      goto fail;
    }
  } else {
    /* I guess it's safe to close the first 3 fd's, since its a new **
    ** program anyway.............................................. */
    close(1);
    dup(cw_fd);
    close(2);
    dup(cw_fd);

    close(0);
    dup(cr_fd);

    /* close(cw_fd); keep this open so parent can report all errors. */
    close(pr_fd);
    close(cr_fd); close(pw_fd);

    execlp(agent, agent, NULL);
    {
      char complaint[100];
      sprintf(complaint, "Failed to exec the agent, %s.", agent);
      write(cw_fd, complaint, strlen(complaint));
    }
    exit(EXIT_FAILURE);
  }

  close(pr_fd); close(cw_fd); close(cr_fd); close(pw_fd);
  return 0;

 fail:
  if(pr_fd > -1) close(pr_fd); /* even though in_fd will never be -1 */
  if(cw_fd > -1) close(cw_fd);
  if(cr_fd > -1) close(cr_fd);
  if(pw_fd > -1) close(pw_fd);
  return -1;
}

/*
 * process_fence_request
 * @node: the name of the node to fence
 *
 * This fuction takes the name of a node in the cluster
 * and looks up all the information require to fence it.
 * Once it obtains the information from the CCS system,
 * it performs the fence.
 *
 * Returns: 0 on success, -1 on failure/error
 */
int process_fence_request(char *node){
  int error=0;
  int success=0;
  ccs_node_t *nodes_cn;
  ccs_node_t *method_cn;
  ccs_node_t *tmp_cn;
  char request[256];
  char *fence_agent;
  char *fence_args;

  log_dbg("Entering process_fence_request()\n");
  error = open_ccs_file(&nodes_cn, "nodes.ccs");
  if(error == -EPARSE){
    log_err("Unable to open nodes.ccs:  Format error.\n");
    log_err("Has the CCS archive been corrupted?\n\n");
    goto fail;
  } else if(error){
    char *msg = strerror(-error);
    log_err("Unable to open nodes.ccs:  %s\n", msg);
    goto fail;
  }

  sprintf(request, "nodes/%s", node);

  tmp_cn = find_ccs_node(nodes_cn, request, '/');

  if(!tmp_cn){
    log_err("Cannot locate the cluster node, %s\n", node);
    goto fail;
  }

  sprintf(request, "%s/fence", node);
  tmp_cn = find_ccs_node(tmp_cn, request, '/');
  if(!tmp_cn || !tmp_cn->child){
    log_err("No fence methods specified for node, %s\n", node);
    goto fail;
  }

  for(tmp_cn = tmp_cn->child; tmp_cn; tmp_cn = tmp_cn->sib){
    log_msg("Performing fence method, %s, on %s.\n",
	   tmp_cn->key, node);
    for(method_cn = tmp_cn->child; method_cn; method_cn = method_cn->sib){
      if(build_fence_command(method_cn, &fence_agent, &fence_args)){
	success = 0;
	log_err("Unable to build fence command.\n");
	break; /* skip to next method */
      } else {
	log_dbg("%s:\n"
		"%s\n", fence_agent, fence_args);
	if(run_fence_command(fence_agent, fence_args)){
	  success = 0;
	} else {
	  success = 1;
	}
	free(fence_agent); fence_agent = NULL;
	free(fence_args); fence_args = NULL;
	if(!success) break;  /* skip to next method */
      }
    }
    if(success){
      break;
    }
  }

  if(!success){
    goto fail;
  }

  if(nodes_cn) close_ccs_file(nodes_cn);
  log_dbg("Exiting process_fence_request() - SUCCESS\n");
  return 0;
  
 fail:
  log_err("All fencing methods FAILED!\n");
  if(nodes_cn) close_ccs_file(nodes_cn);
  log_dbg("Exiting process_fence_request() - FAILURE\n");
  return -1;
}
