/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "servd.h"
#include <linux/gnbd_request.h>
#include "device.h"
#include "gserv.h"
#include "interfaced.h"

LIST_HEAD(device_list);
struct semaphore device_list_lk;

void
init_gserv_device(void)
{
	init_MUTEX(&device_list_lk);
}

int
devices_need_gulm(void)
{
	list_t *list_item;
	dev_info_t *dev_info = NULL;
	int found = 0;

	list_for_each(list_item, &device_list) {
		dev_info = list_entry(list_item, dev_info_t, list);
		if (needs_gulm(dev_info)) {
			found = 1;
			break;
		}
	}

	return found;
}

dev_info_t *
__find_device(char *name)
{
	list_t *list_item;
	dev_info_t *dev_info = NULL;

	list_for_each(list_item, &device_list) {
		dev_info = list_entry(list_item, dev_info_t, list);
		if (strcmp(name, dev_info->name) == 0)
			break;
		dev_info = NULL;
	}

	return dev_info;
}

dev_info_t *
find_and_lock_device(char *name)
{
	dev_info_t *dev;

	down(&device_list_lk);
	dev = __find_device(name);
	if (dev)
		down(&dev->lock);
	up(&device_list_lk);

	return dev;
}

int
create_device(char *name, char *path, u32 timeout, u8 readonly)
{
	int err;
	dev_info_t *tmp;
	struct inode *inode;

	down(&device_list_lk);
	if (__find_device(name)) {
		printe("gnbd name \"%s\" already in use\n", name);
		err = -EEXIST;
		goto fail_lock;
	}
	/* log into gulm master */
	tmp = kmalloc(sizeof (dev_info_t), GFP_KERNEL);
	if (!tmp) {
		printe("couldn't allocate memory for gnbd info : %s\n", name);
		err = -ENOMEM;
		goto fail_lock;
	}
	memset(tmp, 0, sizeof (dev_info_t));

	init_MUTEX(&tmp->lock);
	tmp->timeout = timeout;
	tmp->readonly = readonly;
	tmp->name = kmalloc(buflen(name), GFP_KERNEL);
	if (!tmp->name) {
		printe("couldn't allocate memory for gnbd name : %s\n", name);
		err = -ENOMEM;
		goto fail_info;
	}
	strcpy(tmp->name, name);
	tmp->path = kmalloc(buflen(path), GFP_KERNEL);
	if (!tmp->path) {
		printe("couldn't allocate memory for gnbd path : %s\n", path);
		err = -ENOMEM;
		goto fail_name;
	}
	strcpy(tmp->path, path);
	tmp->file = filp_open(path, O_RDWR | O_SYNC, 0600);
	if ((err = IS_ERR(tmp->file)) > 0) {
		printe("can't open file (%ld)\n", PTR_ERR(tmp->file));
		err = PTR_ERR(tmp->file);
		goto fail_malloc;
	}

	inode = tmp->file->f_dentry->d_inode;
	if (S_ISBLK(inode->i_mode)) {
		struct gendisk *device_info;
		tmp->rdev = inode->i_rdev;
		if (blk_size[MAJOR(tmp->rdev)])
			tmp->size =
			    blk_size[MAJOR(tmp->rdev)][MINOR(tmp->rdev)];
		else if ((device_info = get_gendisk(tmp->rdev)) != NULL)
			tmp->size = device_info->sizes[MINOR(tmp->rdev)];
		else {
			printe("using MAX_DISK_SIZE for size\n");
			tmp->size = MAX_DISK_SIZE;
		}
		tmp->size *= 1024;

		if (blksize_size[MAJOR(tmp->rdev)])
			tmp->blksize =
			    blksize_size[MAJOR(tmp->rdev)][MINOR(tmp->rdev)];
		if (!tmp->blksize)
			tmp->blksize = BLOCK_SIZE;
	} else if (S_ISREG(inode->i_mode)) {
		if (tmp->file->f_op->read == NULL ||
		    tmp->file->f_op->write == NULL) {
			printe("Can't find the read/write f_ops\n");
			err = -EIO;
			goto fail_file;
		}
		if (!cached(tmp)) {
			printe
			    ("WARNING: can't run uncached on a regular file\n");
			printe("switching gnbd device \"%s\" to cached\n",
			       name);
			tmp->timeout = 0;
		}
		tmp->blksize = inode->i_blksize;
		if (!tmp->blksize)
			tmp->blksize = BLOCK_SIZE;
		tmp->size = inode->i_size;
		if (tmp->size <= 0) {
			printe("zero file size or error reading inode\n");
			err = -EIO;
			goto fail_file;
		}
	} else {
		printe("expected either a block device or a regular file\n");
		err = -EINVAL;
		goto fail_file;
	}

	if (needs_gulm(tmp)) {
		err = start_interface();
		if (err) {
			printe
			    ("cannot export an uncached server without interfacing GULM\n");
			goto fail_file;
		}
	}

	printd("server blocksize is %d bytes\n", tmp->blksize);
	printd("server size is %" PRIu64 " bytes\n", tmp->size);

	list_add(&tmp->list, &device_list);

	MOD_INC_USE_COUNT;
	up(&device_list_lk);
	return 0;

      fail_file:
	fput(tmp->file);
      fail_malloc:
	kfree(tmp->path);
      fail_name:
	kfree(tmp->name);
      fail_info:
	kfree(tmp);
      fail_lock:
	up(&device_list_lk);
	return err;
}

int
remove_device(char *name, int force)
{
	int err = 0;
	dev_info_t *dev;

	down(&device_list_lk);
	dev = __find_device(name);
	if (!dev) {
		printe("can't find gnbd device \"%s\"\n", name);
		err = -ENODEV;
		goto fail;
	}
	down(&(dev->lock));
	if (force) {
		err = kill_gserv(0, dev);
		if (err)
			goto fail_lock;
	} else if (find_gserv_info(0, dev)) {
		printe("device %s is in use. Cannot remove\n", dev->name);
		err = -EBUSY;
		goto fail_lock;
	}
	list_del(&dev->list);
	fput(dev->file);
	kfree(dev->path);
	kfree(dev->name);
	kfree(dev);

	if (devices_need_gulm() == FALSE)
		stop_interface();

	up(&device_list_lk);
	MOD_DEC_USE_COUNT;
	return 0;

      fail_lock:
	up(&dev->lock);
      fail:
	up(&device_list_lk);
	return err;
}

int
names_list(char **buffer, u32 * list_size)
{
	int err = 0;
	list_t *list_item;
	char *ptr;
	dev_info_t *dev_info = NULL;
	u32 size = 0;

	*buffer = NULL;
	down(&device_list_lk);
	list_for_each(list_item, &device_list) {
		dev_info = list_entry(list_item, dev_info_t, list);
		size += buflen(dev_info->name);
	}
	if (size == 0) {
		*list_size = size;
		goto exit;
	}
	ptr = kmalloc(size, GFP_KERNEL);
	if (ptr == NULL) {
		printe("cannot allocate memory for names list\n");
		err = -ENOMEM;
		goto exit;
	}
	*buffer = ptr;
	*list_size = size;
	list_for_each(list_item, &device_list) {
		dev_info = list_entry(list_item, dev_info_t, list);
		memcpy(ptr, dev_info->name, buflen(dev_info->name));
		ptr += buflen(dev_info->name);
	}

      exit:
	up(&device_list_lk);
	return err;
}

int
full_list(char **buffer, u32 * list_size)
{
	int err = 0;
	list_t *list_item;
	char *ptr;
	dev_info_t *dev_info = NULL;
	full_list_t tmp;
	u32 size = 0;

	*buffer = NULL;
	down(&device_list_lk);
	list_for_each(list_item, &device_list) {
		dev_info = list_entry(list_item, dev_info_t, list);
		size += (uint32_t) ((sizeof (tmp) + buflen(dev_info->name) +
				     buflen(dev_info->path) + 7) & ~7);
	}
	if (size == 0) {
		*list_size = size;
		goto exit;
	}
	ptr = kmalloc(size, GFP_KERNEL);
	if (ptr == NULL) {
		printe("cannot allocate memory for server list\n");
		err = -ENOMEM;
		goto exit;
	}
	*buffer = ptr;
	*list_size = size;
	list_for_each(list_item, &device_list) {
		dev_info = list_entry(list_item, dev_info_t, list);
		tmp.devsize = dev_info->size;
		tmp.blksize = dev_info->blksize;
		tmp.timeout = dev_info->timeout;
		tmp.namelen = buflen(dev_info->name);
		tmp.len = (sizeof (tmp) + buflen(dev_info->name) +
			   buflen(dev_info->path) + 7) & ~7;
		tmp.readonly = dev_info->readonly;
		FULL_LIST_TO_BE(&tmp);
		memcpy(ptr, &tmp, sizeof (tmp));
		memcpy(ptr + sizeof (tmp), dev_info->name,
		       buflen(dev_info->name));
		memcpy(ptr + sizeof (tmp) + buflen(dev_info->name),
		       dev_info->path, buflen(dev_info->path));
		ptr +=
		    ((sizeof (tmp) + buflen(dev_info->name) +
		      buflen(dev_info->path) + 7) & ~7);
	}

      exit:
	up(&device_list_lk);
	return err;
}
