/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "servd.h"
#include "trans.h"
#include "fence.h"

LIST_HEAD(timestamp_list);
LIST_HEAD(banned_list);
struct semaphore timestamp_list_lk;
struct semaphore banned_list_lk;

void
init_fence(void)
{
	init_MUTEX(&timestamp_list_lk);
	init_MUTEX(&banned_list_lk);
}

int
update_timestamp_list(u32 ip, u64 timestamp)
{
	int err = 0;
	list_t *list_item;
	timestamp_t *client;

	down(&timestamp_list_lk);
	list_for_each(list_item, &timestamp_list) {
		client = list_entry(list_item, timestamp_t, list);
		if (ip == client->ip) {
			client->timestamp = timestamp;
			goto out;
		}
	}
	GNBD_RETRY(client =
		   (timestamp_t *) kmalloc(sizeof (timestamp_t), GFP_KERNEL),
		   client);
	if (!client) {
		printe("couldn't allocate memory for the client timestamp\n");
		err = -ENOMEM;
		goto out;
	}
	client->ip = ip;
	client->timestamp = timestamp;
	list_add(&client->list, &timestamp_list);

      out:
	up(&timestamp_list_lk);
	return err;
}

int
new_timestamp(u32 ip, u64 timestamp)
{
	int new = 1;
	list_t *list_item;
	timestamp_t *client;

	down(&timestamp_list_lk);
	list_for_each(list_item, &timestamp_list) {
		client = list_entry(list_item, timestamp_t, list);
		if (ip == client->ip) {
			if (timestamp == client->timestamp)
				new = 0;
			goto out;
		}
	}
      out:
	up(&timestamp_list_lk);
	return new;
}

void
clear_timestamp_list(void)
{
	list_t *list_item, *tmp;
	timestamp_t *client;

	down(&timestamp_list_lk);
	list_for_each_safe(list_item, tmp, &timestamp_list) {
		client = list_entry(list_item, timestamp_t, list);
		list_del(list_item);
		kfree(client);
	}
	up(&timestamp_list_lk);
	return;
}

int
__check_banned_list(u32 ip)
{
	list_t *list_item;
	int is_banned = FALSE;
	banned_t *banned_client;

	list_for_each(list_item, &banned_list) {
		banned_client = list_entry(list_item, banned_t, list);
		if (ip == banned_client->ip) {
			is_banned = TRUE;
			break;
		}
	}
	return is_banned;
}

int
add_to_banned_list(u32 ip)
{
	int err = 0;
	banned_t *banned_client;
	char ip_str[16];

	down(&banned_list_lk);
	if (__check_banned_list(ip)) {
		printe("client %s already banned\n", beip_to_str(ip, ip_str));
		goto out;
	}
	GNBD_RETRY(banned_client =
		   (banned_t *) kmalloc(sizeof (banned_t), GFP_KERNEL),
		   banned_client);
	if (!banned_client) {
		printe("couldn't allocate memory for the banned client\n");
		err = -ENOMEM;
		goto out;
	}
	banned_client->ip = ip;
	list_add(&banned_client->list, &banned_list);

      out:
	up(&banned_list_lk);
	return err;
}

int
remove_from_banned_list(u32 ip, u64 timestamp)
{
	int err = 0;
	list_t *list_item;
	banned_t *banned_client;

	down(&banned_list_lk);
	list_for_each(list_item, &banned_list) {
		banned_client = list_entry(list_item, banned_t, list);
		if (banned_client->ip == ip) {
			if (timestamp && !new_timestamp(ip, timestamp)) {
				err = -EPERM;
				goto out;
			}
			list_del(list_item);
			kfree(banned_client);
			goto out;
		}
	}
      out:
	up(&banned_list_lk);
	return err;
}

void
clear_banned_list(void)
{
	list_t *list_item, *tmp;
	banned_t *banned_client;

	down(&banned_list_lk);
	list_for_each_safe(list_item, tmp, &banned_list) {
		banned_client = list_entry(list_item, banned_t, list);
		list_del(list_item);
		kfree(banned_client);
	}
	up(&banned_list_lk);

	return;
}

int
list_banned(char **buffer, u32 * list_size)
{
	int err = 0;
	list_t *list_item;
	u32 *ptr;
	banned_t *banned_client;
	u32 size = 0;

	*buffer = NULL;
	down(&banned_list_lk);
	list_for_each(list_item, &banned_list)
	    size += sizeof (u32);
	if (size == 0) {
		*list_size = size;
		goto exit;
	}
	ptr = kmalloc(size, GFP_KERNEL);
	if (ptr == NULL) {
		printe("cannot allocate memory to list banned clients\n");
		err = -ENOMEM;
		goto exit;
	}
	*buffer = (char *) ptr;
	*list_size = size;
	list_for_each(list_item, &banned_list) {
		banned_client = list_entry(list_item, banned_t, list);
		*ptr = banned_client->ip;
		ptr++;
	}

      exit:
	up(&banned_list_lk);
	return err;
}
