/******************************************************************************
*******************************************************************************
**
**  Copyright (C) Sistina Software, Inc.  2002-2003  All rights reserved.
**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
**
**  This copyrighted material is made available to anyone wishing to use,
**  modify, copy, or redistribute it subject to the terms and conditions
**  of the GNU General Public License v.2.
**
*******************************************************************************
******************************************************************************/

#include "gnbd.h"
#include "trans.h"
#include "macros.h"
#include <linux/gnbd_request.h>
#include "monitor.h"

int
receiver(void *data)
{
	gnbd_t *device = (gnbd_t *) data;
	trans_hdr_t header;
	struct list_head *pos;
	int err = 0;
	struct request *req;
	u32 len;
	u64 offset;
	int blocks;
	signal_info_t sig_info;

	gnbd_daemonize("receiver");

	setup_blocking(&sig_info);

	device->receiver_tsk = current;

	device->server_alive = TRUE;
	complete(&device->receiver_wait);
	while (1) {
		if (transfer_info
		    (RECV, device->sock, &header, sizeof (trans_hdr_t)) != 0) {
			printe("receive failed for %d (%s)\n",
			       gnbd_get_minor(device),
			       gnbd_device_names[gnbd_get_minor(device)]);
			err = -1;
			goto shutdown_receive;
		}
		printd("Received request: type=0x%x, key=0x%" PRIx64
		       ", size=%u, offset=%" PRIu64 "\n",
		       (unsigned int) header.type, header.key, header.len,
		       header.offset);
		if (header.type == gnbd_shutdown) {
			printd("Got a shutdown command\n");
			err = 0;
			goto shutdown_receive;
		}
		if (header.type != gnbd_read_reply
		    && header.type != gnbd_write_reply
		    && header.type != gnbd_error) {
			printe("Unknown header type received : 0x%x\n",
			       (unsigned int) header.type);
			err = -1;
			goto shutdown_receive;
		}
		if (header.type == gnbd_read_reply
		    && header.key == GNBD_PING_KEY
		    && header.len == cpu_to_be32(0)
		    && header.offset == cpu_to_be64(0)) {
			printd("got a ping reply\n");
			if (device->timeout) {
				down(&device->trans_queue_lk);
				if (gnbd_pending(device))
					mod_timer(&device->timer,
						  jiffies +
						  (device->timeout * HZ));
				else
					del_timer(&device->timer);
				up(&device->trans_queue_lk);
			}
			continue;
		}
		req = NULL;
		list_for_each(pos, &device->queue) {
			if ((unsigned long) pos == header.key) {
				req = list_entry(pos, struct request, queue);
				break;
			}
		}
		if (req == NULL) {
			printe
			    ("couldn't match incoming header with an existing request\n");
			err = -1;
			goto shutdown_receive;
		}
		if (header.type == gnbd_error) {
			printe("received an error message from the server\n");
			if (header.err == error_invalid_request) {
				printe("And the server blames ME!\n");
				remove_request(device, req);
				req->errors = TRUE;
				gnbd_end_request(req);
			}
			err = 0;
			goto shutdown_receive;
		}
		offset = (u64) (req->sector) * GNBD_HARDSECT;
		len = req->nr_sectors * GNBD_HARDSECT;
		header.len = (size_t) (be32_to_cpu(header.len));
		header.offset = (loff_t) (be64_to_cpu(header.offset));
		if (header.len != len || header.offset != offset) {
			printe("I requested %u bytes, offset %" PRIu64
			       ". I got %u bytes, offset %" PRIu64
			       ". This is bad\n", len, offset, header.len,
			       header.offset);
			err = -1;
			goto shutdown_receive;
		}
		blocks = header.len / req->bh->b_size;
		if (header.type == gnbd_read_reply) {
			printd("Receiving %u bytes of data\n", header.len);
			err = recv_request_data(device->sock, req, blocks);
			if (err) {
				printe
				    ("couldn't receive data from the the server: %d\n",
				     err);
				err = -1;
				goto shutdown_receive;
			}
		}
		remove_request(device, req);
		gnbd_end_request(req);
	}
      shutdown_receive:
	device->sock->ops->shutdown(device->sock, 2);
	down(&device->sender_lk);
	if (device->sock != NULL)
		sock_release(device->sock);
	device->sock = NULL;
	up(&device->sender_lk);
	if (device->timeout) {
		del_timer(&device->timer);
		spin_lock_irqsave(&current->sighand->siglock, sig_info.flags);
		flush_signals(current);
		siginitsetinv(&current->blocked,
			      (sigmask(SIGKILL) | sigmask(SIGTERM)));
		recalc_sigpending();
		spin_unlock_irqrestore(&current->sighand->siglock,
				       sig_info.flags);
		if (err) {
			err = DONT_WAIT;
			if (device->server_alive)
				err = send_expire_request(device);
			if (err != DONT_WAIT) {
				printe("waiting for %s to be fenced\n",
				       device->server_name);
				if (device->server_alive)
					wait_for_completion(&device->wait);
			}
		}
		down(&device->notify_lk);
		device->server_alive = FALSE;
		up(&device->notify_lk);
		empty_trans_queue(device);
	} else
		device->server_alive = FALSE;
	device->receiver_tsk = NULL;
	complete(&device->receiver_wait);
	return err;
}

/******************************************************************************
 * resend_requests()
 *
 * For each pending request in the transfer queue, steps through and re-issues
 * that request to the server.
 * This function is only intended to be called if we lost a server
 * connection and are reestablishing a new connection.
 */
static int
resend_requests(gnbd_t * dev)
{
	struct list_head *tmp;
	struct request *req;
	trans_hdr_t header;
	int err = 0;
	signal_info_t sig_info;

	printe("resending requests\n");
	if (dev->timeout)
		mod_timer(&dev->timer,
			  jiffies + ((unsigned long) dev->timeout * HZ));
	list_for_each(tmp, &dev->queue) {
		req = list_entry(tmp, struct request, queue);
		header.type =
		    (req->cmd == READ) ? gnbd_read_req : gnbd_write_req;
		header.key = (unsigned long) (&req->queue);
		header.err = 0;
		header.len = cpu_to_be32(req->nr_sectors * GNBD_HARDSECT);
		header.offset =
		    cpu_to_be64((u64) (req->sector) * GNBD_HARDSECT);

		printd("Re-requesting key 0x%" PRIx64 "\n", header.key);

		block_signals(&sig_info);

		down(&dev->sender_lk);

		if (gnbd_connected(dev) != INFO_CONNECTED)
			err = -ENOTCONN;

		if (!err && req->cmd == READ)
			err =
			    transfer_info(SEND, dev->sock, &header,
					  sizeof (trans_hdr_t));
		if (!err && req->cmd == WRITE)
			err = send_request_data(dev->sock, req, &header,
						dev->iov);

		if (err)
			send_sig(SIGINT, dev->receiver_tsk, 0);

		up(&dev->sender_lk);

		unblock_signals(&sig_info);

		if (err) {
			printe("failed trying to resend request (%d)\n", err);
			goto exit;
		}
	}

      exit:
	return err;
}

static int
gnbd_login(int num, gnbd_t * dev)
{
	int err = 0;
	login_req_t login_request;
	login_reply_t login_reply;
	signal_info_t sig_info;
	char *name = NULL;
	u16 len = 0;
	u32 cmd;

	printd("sending login info\n");
	strcpy(login_request.devname, gnbd_device_names[num]);
	login_request.timestamp = cpu_to_be64(insmod_time);
	login_request.version = cpu_to_be16(PROTOCOL_VERSION);

	block_signals(&sig_info);
	down(&dev->sender_lk);

	cmd = GNBD_SERV_LOGIN_REQ;
	err = transfer_be32(SEND, dev->sock, &cmd);
	if (!err)
		err = transfer_info(SEND, dev->sock, &login_request,
				    sizeof (login_request));

	up(&dev->sender_lk);

	if (err)
		goto end_transfer;

	err = timed_transfer_info(RECV, dev->sock, &login_reply,
				  sizeof (login_reply));
	if (err || login_reply.err)
		goto end_transfer;

	len = be16_to_cpu(login_reply.name_size);
	name = kmalloc(len, GFP_KERNEL);
	if (!name) {
		printe("cannot allocate memory for server name\n");
		err = -ENOMEM;
		goto end_transfer;
	}
	err = timed_transfer_info(RECV, dev->sock, name, len);

      end_transfer:
	unblock_signals(&sig_info);

	if (err) {
		if (name != NULL)
			kfree(name);
		printe("login to server failed (%d)\n", err);
		goto exit;
	}

	if (login_reply.err) {
		printe("server refused login (%d)\n", -login_reply.err);
		err = -ECONNREFUSED;
		goto exit;
	}

	if (dev->server_name)
		kfree(dev->server_name);
	dev->server_name = name;

	dev->size = be64_to_cpu(login_reply.devsize);
	dev->readonly = login_reply.readonly;
	dev->timeout = be32_to_cpu(login_reply.timeout);

      exit:
	return err;
}

int
start_receiver(int num, gnbd_t * dev)
{
	int err = 0;
	unsigned long flags;
	request_queue_t *q = &blk_dev[gnbd_major].request_queue;

	down(&dev->receiver_lk);
	if (dev->receiver_tsk)
		goto fail;

	err = connect_to_server(dev->servip, dev->servport, &dev->sock);
	if (err)
		goto fail;

	err = gnbd_login(num, dev);
	if (err)
		goto fail_socket;

	init_completion(&dev->wait);
	init_completion(&dev->receiver_wait);
	err = kernel_thread(receiver, dev, 0);
	if (err < 0) {
		printe("Couldn't start receiver process (%d)\n", err);
		goto fail_socket;
	}

	/* wait for thread to get itself set up before continuing. */
	wait_for_completion(&dev->receiver_wait);

	/* before or after the thread exists? */
	if (gnbd_pending(dev)) {
		err = resend_requests(dev);
		if (err)
			goto fail;
	}
	spin_lock_irqsave(&io_request_lock, flags);
	q->plugged = 1;
	spin_unlock_irqrestore(&io_request_lock, flags);
	generic_unplug_device((void *) q);
	up(&dev->receiver_lk);
	return 0;

      fail_socket:
	sock_release(dev->sock);
	dev->sock = NULL;
      fail:
	up(&dev->receiver_lk);
	return err;
}

void
stop_receiver(gnbd_t * dev)
{
	trans_hdr_t shutdown_req;
	signal_info_t sig_info;

	down(&dev->receiver_lk);

	shutdown_req.type = gnbd_shutdown;
	shutdown_req.key = 0;
	shutdown_req.len = 0;
	shutdown_req.offset = 0;

	block_signals(&sig_info);

	down(&dev->sender_lk);

	if (gnbd_connected(dev) == INFO_CONNECTED)
		transfer_info(SEND, dev->sock, &shutdown_req,
			      sizeof (shutdown_req));

	up(&dev->sender_lk);

	unblock_signals(&sig_info);

	if (dev->receiver_tsk)
		wait_for_completion(&dev->receiver_wait);
	up(&dev->receiver_lk);

}
